/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* About sending LSU packets: LSU can be sent 1) for a retransmition, when
 * the rmx timer fires. See ospf_rmx_timer(). 2) when flooding a received
 * LSU. See ospf_flood(). 3) When flooding a self originated LSU. See
 * ospf_flood_selforiginate(). 4) When receiving a LS request packet. See
 * ospf_send_lsudirect() and ospf_lsr_recv().
 * In case 2 and 3, LSU are put on the neigbour queue, and sent after
 * all processing in done. See ospf_ngbrmx_flush(), ospf_flush_rmx().
 * In case 1) ospf_send_rmx() doe sthe job of retransmiting. In case
 * 4) ospf_send_lsudirect() send the LSU which are not queued.
 */
#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"

/* Build the Database summary list for a given neighbour.
 */
void
ospf_buildDataSum __PF1(ngb, ospf_ngb_t *)
{
	vertex_t *vtx;
	ospf_area_t *area;
	int type;

	/* Lookup all areas' DB, by type
	 */
	OSPF_ALL_AREA(area) {
		for(type = LST_RTR; type < LST_MAXTYPE; type++) {
			if (!isflooded(ngb, type, area))
				continue;

			PT_WALK_TYPE(&area->area_tree, type, vtx) {
				/* Ignore dead LS.
		 	 	 */
				if (BIT_TEST(vtx->vtx_flags,
						LF_HIDDEN|LF_REQUEST|LF_DELETE))
					continue;
			
				/* Move the vertex to the neighbour list. If
				 * the age is ospf_maxAge them we add it to
				 * the LS request list. (RFC #10.3)
		 		 */
				if (VTX_AGE_MAX(vtx))
					NGB_LIST_ADD_RMX(ngb, vtx->vtx_type,
						&vtx->vtx_key, vtx);
				else
					NGB_LIST_ADD_DS(ngb, vtx->vtx_type,
						&vtx->vtx_key, vtx);
			} PT_WALK_TYPE_END(&ospf_tree, type, vtx);
		}
	} OSPF_ALL_AREA_END(area);
}

/* Look for an LS in *all* the area databases. It's used to answer LSR
 * request for stuff which can be in any area (C.f. AEX and AIP). Also
 * used for self originated LSA with could have been announced by a neigbour,
 * and put into the wrong area. If 'newtree' is no null, and the LS is not in
 * this tree, we move it there!
 */
vertex_t *
ospf_global_find __PF5(tree, db_tree_t *, type, int, key, lskey_t *,
		searchdata, searchdata_t *, newtree, db_tree_t *)
{
	vertex_t *vtx = NULL;
	ospf_area_t *area;
	searchdata_t searchsave;
	 
	searchdata->sd_last = NULL;

	switch(type) {
	case LST_RTR:
	case LST_NTW:
		/* Those have to be looked up in one area only
		 */
		return(db_find(tree, type, key, searchdata));

	case LST_ASIP:
	case LST_ASBR:
	case LST_ASEX:
		/* We look everywhere!
	 	*/
		OSPF_ALL_AREA(area) {
			if ((vtx = db_find(&area->area_tree, type, key,
					&searchsave)))
				break;
			/* If the tree currently searched is the prefered one,
			 * save the searchdata value (so in case of a
			 * ls_create on that tree, the data will be ok).
			 */
			if (tree == &area->area_tree)
				searchdata->sd_last = searchsave.sd_last;

		} OSPF_ALL_AREA_END(area);

		if (!vtx)
			return(NULL);
		break;
	}

	/* If newtree is non null, and the LS is not in this tree, we move it
	 * there!
	 */
	if (newtree && newtree != &area->area_tree) {
		tree_node_del(&area->area_tree.pt_root[vtx->vtx_type],
			(u_int8 *)&vtx->vtx_key, ospf_key_len[vtx->vtx_type],
			(tree_node_t *)vtx);
		ASSERT(!BIT_TEST(vtx->vtx_flags, TNF_EXTERN));
		tree_add((u_int8 *)&vtx->vtx_key, ospf_key_len[type],
			&newtree->pt_root[type], (tree_node_t *)vtx,
			searchdata->sd_last, NULL);
	}
	return(vtx);
}

/* We just got an acknowlegment of the last DD packet we sent. We
 * reset the current buffer (array ....).
 */
void
ospf_ddack __PF1(ngb, ospf_ngb_t *)
{
	/* Just empty the array. The nodes have already been removed from
	 * the list (tree) when been put in the array.
	 */
	ngb->ngb_lastDDsize = 0;
}

/* Check if the copy in our database is less recent. If so, create a new
 * vertex (empty one, marked as REQUEST) and add the vertex on the list of LS
 * requests.
 */
void
ospf_ls_needupdate __PF2(lsa, ospf_lsa_t *, ngb, ospf_ngb_t *)
{
	vertex_t *vtx;
	searchdata_t searchdata;

	/* Loopup the DB to see if it's already there
         */
        if ((vtx = ospf_global_find(&ngb->ngb_tree, lsa->lsa_type,
			&lsa->lsa_key, &searchdata, NULL))) {
		/* If it's more recent, nothing to do.
		 */
		if (ls_cmpage(VTX_AGE(vtx), vtx->vtx_seq, vtx->vtx_cksum,
                       	lsa->lsa_age, lsa->lsa_seq, lsa->lsa_cksum) >= 0)
                	return;
	}
	else {
		/* We create a fake vertex which will be marked LF_REQUEST.
		 * The size is fixed to the minimum.
		 */
		vtx = ls_vtx_create(&ngb->ngb_tree, &lsa->lsa_key,
			lsa->lsa_options, 0, lsa->lsa_age, lsa->lsa_seq,
			0, lsa->lsa_type, sizeof(ospf_lsa_hdr_t), &searchdata);
		BIT_SET(vtx->vtx_flags, LF_REQUEST);
		ospf_trace_vtx(vtx, "REQUEST");
	}

	/* ****** lpj's addition. XXXXXX ********
	 * If it's a sef originated LSA, get rid of it now!
	 *
	if (ospf_self_originate(ngb, lsa, vtx, &searchdata))
		return;
	 */

	/* Add this guy to the list of LS requests.
	 */
	NGB_LIST_ADD_LSR(ngb, vtx->vtx_type, &vtx->vtx_key, vtx);

	return;
}

/* Neighbour list manipulation routines.
 * Empty a list (tree).
 */
void
db_tree_kill __PF2(tree, db_tree_t *, flag, flag_t)
{
	int type;

	for(type = 0; type < LST_MAXTYPE; type++)
		if (tree->pt_root[type]) {
			tree_delete_rec(tree->pt_root[type], flag);
			tree->pt_root[type] = NULL;
		}

	tree->pt_null_nodes = 0;
	tree->pt_used_nodes = 0;
}

/* Get the next element on the list/tree. Free the tree structure.
 * This is used to empty the LS request list. Take one element from the
 * list and send it. The trick is that we take a (pseudo) random element,
 * so if the router establish an adjacency with RT1 and RT2, it will
 * ask for diferent LS from RT1 and RT2, and therefore fill up the database
 * faster. Well, doesnot make sense yet, but it will eventuallly.
 */
vertex_t *
ngb_list_next __PF1(tree, db_tree_t *)
{
	tree_node_t *node, *prev;
	ospf_ptwalk_t ptw;
	int Counter = 1, type;
	static int Random = (int)ngb_list_next;
	int junk;
	vertex_t *vtx;

	for(type = 0; type < LST_MAXTYPE; type++)
		if ((node = tree->pt_root[type]))
			break;

	/* Is it empty?
	 */
	if (!node)
		return((vertex_t *)NULL);

	ptw.tw_stkind = 0;
	Random ^= (task_time.gt_current.ut_usec ^ junk);

	/* Go down the tree until we get to a leaf.
	 */
	while(1) {
		/* Are we at a leaf?
		 */
		if (!TREENODE_DOWN(node, node->tn_left) &&
			!TREENODE_DOWN(node, node->tn_right))
			break;

		PTW_PUSH(&ptw, node);

		/* Go down one stage.
		 */
		if (Random & (Counter <<= 1))
			node = TREENODE_DOWN(node, node->tn_left)?
					node->tn_left: node->tn_right;
		else
			node = TREENODE_DOWN(node, node->tn_right)?
					node->tn_right: node->tn_left;
        }

	/* We's found a node pointed by 'node'. We can remove that one,
	 * and all the previous ones in the stack if they have no leaf.
	 */
	vtx = node->tn_data;
	node->tn_data = NULL;

	while(node) {
                if (node->tn_data || TREENODE_DOWN(node, node->tn_left) ||
				TREENODE_DOWN(node, node->tn_right))
                        break;

                /* Ok, we delete that one. Get the previous node first
                 * to update its links.
                 */
                prev = PTW_POP(&ptw);
                if (prev) {
                        if (prev->tn_left == node)
				prev->tn_left = (node->tn_left != node)?
					node->tn_left:
						(node->tn_right != node)?
							node->tn_right:
							prev;
                        else
                                prev->tn_right = (node->tn_left != node)?
					node->tn_left:
						(node->tn_right != node)?
							node->tn_right:
							prev;
			if (prev->tn_left == prev && prev->tn_right == prev)
				BITINDEX_RESET(prev->tn_bitindex);
		}

                TREENODE_FREE(node);
                node = prev;
        }

        /* If no node left, reset the root.
         */
        if (!node)
                tree->pt_root[type] = NULL;

	tree->pt_used_nodes--;
	return(vtx);
}

/* Send all LS update in the array lsuQueue[]. Those LSU are sent after
 * a LS request is recieved. LSU are not retransmitted, not put in the RMX
 * queue.
 */
void
ospf_send_lsudirect __PF3(ngb, ospf_ngb_t *, lsuQueue[], vertex_t *, nbVtx, int)
{
	ospf_lsu_pck_t *pck;
	vertex_t *vtx;
	int ind, pcklen, nbAdvt, maxlen;
	ospf_lsa_t *lsa;

	/* Get a packet, type LSU.
	 */
        OSPF_PACKET(pck, OSPF_PCK_LSU, maxlen, ngb->ngb_ifap);

	lsa = pck->pck_lsu_lsa;
	vtx = lsuQueue[0];
	ind = 0;
	nbAdvt = 0;
	pcklen = sizeof(ospf_lsu_pck_t) - sizeof(ospf_lsa_t);

	/* We send as much LSA in a LSU packet as posible.
	 */
	do {
		/* If the next LSA gonna fit in the packet?
		 */
		if (pcklen + vtx->vtx_len > maxlen) {
			/* Well, send the current packet, reset eveything.
			 */
			pck->pck_lsu_nbadv = htonl(nbAdvt);
			ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
				ospf_send_addr(ngb->ngb_intf, SEND_LSU_DIRECT,
					ngb));

			nbAdvt = 0;
			OSPF_PACKET(pck, OSPF_PCK_LSU, maxlen, ngb->ngb_ifap);
			pcklen = sizeof(ospf_lsu_pck_t) - sizeof(ospf_lsa_t);
			lsa = pck->pck_lsu_lsa;
		}

		/* Fill the lsa with the values from the vertex.
		 */
		ospf_lsa_set(lsa, vtx);
		pcklen += vtx->vtx_len;

		/* Get the next one...
		 */
		(int)lsa += vtx->vtx_len;
		vtx = lsuQueue[++ind];
		nbAdvt++;

	} while(ind < nbVtx);

	/* Send the last packet.
	 */
	pck->pck_lsu_nbadv = htonl(nbAdvt);
	ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
		ospf_send_addr(ngb->ngb_intf, SEND_LSU_DIRECT, ngb));
}

/* Convert a vertex into an LSA.
 */
void
ospf_lsa_set __PF2(lsa, ospf_lsa_t *, vtx, vertex_t *)
{
	ospf_lsa_rtlink_t *lsrt;
	ospf_vtxLink_t *vl;
	int len, ind;

	ASSERT(!BIT_TEST(vtx->vtx_flags, LF_REQUEST|LF_HIDDEN));

	len = sizeof(ospf_lsa_hdr_t);

	switch(vtx->vtx_type) {
	case LST_RTR:
		len = sizeof(ospf_lsa_rtr_t) - sizeof(ospf_lsa_rtlink_t);
		lsrt = lsa->lsa_rtr_links;

		for(vl = vtx->vtx_links, ind = 0; vl; vl = vl->vl_next) {
			if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
                                continue;
			LSA_RL_SET(lsrt, vl->vl_id, vl->vl_data, vl->vl_type,
				0, vl->vl_cost);
			(int)lsrt += sizeof(ospf_lsa_rtlink_t) -
				sizeof(ospf_tosmetric_t);
			len += sizeof(ospf_lsa_rtlink_t) -
				sizeof(ospf_tosmetric_t);
			ind++;
		}
		LSA_RT_SET(&lsa->lsa_rt, vtx->vtx_rtr_bits, ind);
		break;

	case LST_NTW:
		/* A network LS, mask and router IDs...
		 */
		LSA_NT_SET(&lsa->lsa_nt, vtx->vtx_mask);
		len += sizeof(u_int32);

		for(vl = vtx->vtx_links, ind = 0; vl; vl = vl->vl_next) {
			if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
				continue;
			LSA_NL_SET(&lsa->lsa_ntw_rtid[ind++], vl->vl_id);
			len += sizeof(u_int32);
		}
		break;

	case LST_ASBR:
		lsa->lsa_sla_mask = 0; /* dixit the RFC, p111!!!! */
		lsa->lsa_sla_tosmet32 = htonl(vtx->vtx_tosmet);

		len = sizeof(ospf_lsa_sla_t);
		break;

	case LST_ASIP:
		lsa->lsa_sla_mask = vtx->vtx_mask;
		lsa->lsa_sla_tosmet32 = htonl(vtx->vtx_tosmet);

		len = sizeof(ospf_lsa_sla_t);
		break;

	case LST_ASEX:
		lsa->lsa_ext_mask = vtx->vtx_mask;
		lsa->lsa_ext_tosmet32 = htonl(vtx->vtx_tosmet);
		lsa->lsa_ext_tag = htonl(vtx->vtx_tag);
		lsa->lsa_ext_forw = vtx->vtx_forward;

		len = sizeof(ospf_lsa_ext_t);
		break;

	default:
		ASSERT(0);
	}

	/* Then the header... vtx->vtx_cksum should be correct when set,
	 * or must be computed when unset.
	 */
	LSA_SET(lsa, VTX_AGE(vtx), vtx->vtx_options, vtx->vtx_type,
		vtx->vtx_id, vtx->vtx_advrt, vtx->vtx_seq, vtx->vtx_cksum,
		vtx->vtx_len);

	if (BIT_TEST(vtx->vtx_flags, LF_REQUEST))
		/* For vertices marked as LF_REQUEST, the data in the DB is not
		 * correct yet, so dont't worry about the lengh or checksum.
		 */
		 return;

	if (!vtx->vtx_cksum)
		vtx->vtx_cksum = lsa->lsa_cksum;

	/* That's a good one!!!!
	 */
	ASSERT(!vtx->vtx_len || len == vtx->vtx_len);
	ASSERT((vtx->vtx_type == LST_ASEX && vtx->vtx_advrt == ospf_rtid) ||
		!vtx->vtx_cksum || !lsa->lsa_cksum ||
		vtx->vtx_cksum == lsa->lsa_cksum);
	if (ospf_check_lsacksum(lsa)) {
		u_int16 cksum;
		u_int16 cksum2;

		cksum = lsa->lsa_cksum;
		lsa->lsa_cksum = 0;
		(void)iso_cksum((void_t)&lsa->lsa_options,
			(size_t)(vtx->vtx_len - sizeof(u_int16)),
			(byte *)&lsa->lsa_cksum);
		cksum2 = iso_cksum((void_t)&lsa->lsa_options,
			(size_t)(vtx->vtx_len - sizeof(u_int16)),
			NULL);
		trace_only_tf(ospf_trace_options, 0,
			("OSPF ospf_lsa_set: Wrong checksum %x wanted %x (%x)",
				cksum, lsa->lsa_cksum, cksum2));
		ASSERT(0);
	}
}

/* Receive a LS Update packet. RFC1583 s13, p76.
 */
int
ospf_lsu_recv __PF4(pck, ospf_lsu_pck_t *, len, int, ospf_intf, ospf_intf_t *,
        src, sockaddr_un *)
{
	ospf_ngb_t *ngb, *allngb;
	int nblsa, cmp, ack_flag;
	ospf_lsa_t *lsa;
	vertex_t *vtx, *req;
	searchdata_t searchdata;

        /* find out about the identity of the neighbour
         */
        if (!(ngb = ospf_find_ngb(ospf_intf, pck->pck_lsu_rtid)))
                return(OSPF_ERR_INVNGB);

	/* Check the neighbour state... Ignore the packet if < EXCHANGE.
	 */
	if (ngb->ngb_state < OSPF_NGBS_EXCHANGE)
		return(OSPF_NOERR);

	/* go through all LSA in the packet.
	 */
	for(nblsa = NTOHL(pck->pck_lsu_nbadv), lsa = pck->pck_lsu_lsa; nblsa;
			nblsa--, (int)lsa += lsa->lsa_len) {
		/* Validate the checksum... (#1)
		 */
		if (ospf_check_lsacksum(lsa))  {
			trace_log_tf(ospf_trace_options, 0, LOG_ERR,
			("OSPF RECV: bad checksum LSA %s (%d) %A Advrt %A",
				trace_value(ospf_lsa_types, lsa->lsa_type),
				lsa->lsa_type,
				sockbuild_in(0, lsa->lsa_id),
				sockbuild_in(0, lsa->lsa_advrt)));
			NTOHS(lsa->lsa_len); /* Needed to compute the next LSA*/
			continue;
		}

		/* From now we use host byte order.
		 */
		NTOH_LSH(&lsa->lsa_hdr);

		/* Check the type and stub announcements. (#2 & #3)
		 */
		if (!LS_BOUND(lsa->lsa_type) || (lsa->lsa_type == LST_ASEX &&
			!BIT_TEST(ospf_intf->ospf_if_options,OSPF_PCK_OPT_E))){

			trace_log_tf(ospf_trace_options, 0, LOG_ERR,
			("OSPF Recv: invalid type LSA %s (%d) %A Advrt %A",
				trace_value(ospf_lsa_types, lsa->lsa_type),
				lsa->lsa_type,
                                sockbuild_in(0, lsa->lsa_id),
                                sockbuild_in(0, lsa->lsa_advrt)));
			continue;
		}

		/* Get our copy from the DB.
		 */
		vtx = ospf_global_find(&ngb->ngb_tree, lsa->lsa_type,
			&lsa->lsa_key, &searchdata, NULL);

		if (vtx)
			ospf_trace_vtx(vtx, "LSU RECV");

		/* the LS is MaxAge (#4) and not in the DB.
		 */
		if (lsa->lsa_age >= ospf_maxAge && LS_NOT_INDB(vtx)) {
			/* Ack the packet
			 */
			ospf_send_ack(ngb, lsa, vtx, ACKF_NOW);

			/* Purge the LS Request list. If vtx exists.
			 */
			if (vtx) {
				req = vtx;
				NGB_LIST_DEL_LSR(ngb, vtx);
				ASSERT(!NGB_LIST_FIND_LSR(req, ngb, req));
			}

			/* If the neighbour state is < FULL, install the LS
			 * into the DB.	
			 */
			if (ngb->ngb_state < OSPF_NGBS_FULL) {
				if (!vtx)
					vtx = ospf_ls_create(&ngb->ngb_tree,
						lsa, NULL, &searchdata);
				else
					ospf_ls_update(VTX_TREE(vtx), lsa, vtx);
				vtx->vtx_nextRecv = time_sec + ospf_minLSInt;
			}
			continue;
		}

		/* The new copy is better than what we have. (#5)
		 */
		if (LS_NOT_INDB(vtx) ||
				(cmp = ls_cmpage(VTX_AGE(vtx), vtx->vtx_seq,
				vtx->vtx_cksum, lsa->lsa_age, lsa->lsa_seq,
				lsa->lsa_cksum)) < 0) {
			/* If the database copy was installed less than
			 * minLSInt seconds ago, discard the packet. (The DB
			 * copy mult be a 'real' copy)
			 */
			if (vtx &&
				!BIT_TEST(vtx->vtx_flags, LF_REQUEST|LF_HIDDEN) &&
				vtx->vtx_nextRecv > time_sec)
				continue;

			/* Replace the current vtx by the new one.
			 * If it's a self originated LSA some action has
			 * to be taken (RFC#13.4). The RFC say that (d) install
			 * the new copy in the DB, and *then* (f) process the
			 * case of self originated LSA. ????? wierd. I won't
			 * do it that way!
			 */
			ack_flag = 0;
			if (!ospf_self_originate(ngb, lsa, vtx, &searchdata)) {
				if (vtx)
					ospf_ls_update(VTX_TREE(vtx), lsa,vtx);
				else
					vtx = ospf_ls_create(&ngb->ngb_tree,
						lsa, NULL, &searchdata);
				vtx->vtx_nextRecv = time_sec + ospf_minLSInt;

				/* Some SPF incremental computation maybe...?
	 			 */
				ospf_spf_inc(ngb->ngb_area, vtx);

				/* Flood the LSA to *some* subset of interfaces.
			 	 */
				ack_flag = ospf_flood(ngb, vtx);
			}
			else
				ack_flag = vtx? ACKF_NEW: ACKF_NOW;

			/* Purge the LS from the retransmission list of
			 * *all* nieghbours.
			 */
			if (vtx) {
				NGB_ALL(allngb) {
					(void)ngb_list_del_rmx(allngb, vtx);
				} NGB_ALL_END(allngb);
			}

			/* Ack the packet.
			 */
			ospf_send_ack(ngb, lsa, vtx, ack_flag);

			/* Handling of self originated LSA should be here????
			 */

			continue;
		}

		/* If there is a request for that LSA in the LS request list,
		 * this is an error. (#6)
		 */
		if (NGB_LIST_FIND_LSR(req, ngb, vtx)) {
			ospf_ngb_event(EV_NBG_BADLSREQ, ngb);
			return(OSPF_ERR_BADLSR);
		}

		/* The LS is the same as the one in the DB. (#7)
		 */
		if (!cmp) {
			/* Delete the vtx from the retransmisison list.
			 * Acknoledge the stuff... 
			 */
			if (ngb_list_del_rmx(ngb, vtx))
				ospf_send_ack(ngb, lsa, vtx, ACKF_IMPLIED);
			else
				ospf_send_ack(ngb, lsa, NULL, ACKF_NOW);

			continue;
		}

		/* The database copy is more recent. Send a warning.
		 */
		trace_log_tf(ospf_trace_options, 0, LOG_WARNING,
			("OSPF LSU: Database copy more recent: LSA discarded"));
	}

	/* If the LS Request list is empty, we can go to the FULL
	 * state.
	 */
	if (ngb->ngb_state == OSPF_NGBS_LOADING && NGB_LIST_EMPTY_LSR(ngb))
		ospf_ngb_event(EV_NBG_LOADDONE, ngb);

	return(OSPF_NOERR);
}

/* Trigger a delayed acknoledgement send.
 */
void
ospf_ngb_acktimer __PF2(tip, task_timer *, interval, time_t)
{
        ospf_ngb_t *ngb;

        ngb = (ospf_ngb_t *)tip->task_timer_data;
	ospf_send_ack(ngb, NULL, NULL, ACKF_DELAY);
}

/* Send an acknowledgment to the neighbour.
 */
void
ospf_send_ack __PF4(ngb, ospf_ngb_t *, lsa, ospf_lsa_t *, vtx, vertex_t *,
		flags, flag_t)
{
	int delayed = 0;
	ospf_lsa_hdr_t *lsh;
	int maxlen, pcklen;
	ospf_lsa_pck_t *pck;
	int type = SEND_LSA_DIRECT;

	/* Figure out if we need to ACK now or later. RFC #13.5
	 */
	switch(flags) {
	case ACKF_DELAY:
		/* This call has been triggered by the timer. There must be
		 * something to send.
		 */
		ASSERT(ngb->ngb_lastAckIndex);

		delayed = 0;
		type = SEND_LSA_DELAYED;
		break;

	case ACKF_FBACK:
		/* We flooded back the LSA.
		 */
		return;

	case ACKF_NEW:
	 	/* If we have a more recent LSA...
	 	 */
		if (INTF_ISBACKUP(ngb->ngb_intf) && !NGB_ISDR(ngb))
			return;
		delayed = 1;
		break;

	case ACKF_IMPLIED:
		/* Implied duplicate LSA received.
		 */
		if (!INTF_ISBACKUP(ngb->ngb_intf) || !NGB_ISDR(ngb))
			return;
		delayed = 1;
		break;

	case ACKF_NOW:
		break;

	default:
		ASSERT(0);
	}

	/* If the delay buffer is full we don't delay.
	 */
	if (delayed && ngb->ngb_lastAckIndex == ngb->ngb_maxACKinPacket)
		delayed = 0;

	/* Acknoledgement to be sent are put in an array for (later)
	 * processing. If the array is full, the ACK is sent now.
	 * vtx are put in the array only if ACKF_NOW is not set.
	 */
	if (delayed) {
		ASSERT(vtx);
		ngb->ngb_ackList[ngb->ngb_lastAckIndex++] = vtx;

		/* We set up a timer for delayed ACK if none is scheduled yet.
		 */
		if (!BIT_TEST(ngb->ngb_ackTimer->task_timer_flags,
				TIMERF_PROCESSING))
			task_timer_set(ngb->ngb_ackTimer, 0, ngb->ngb_ackDelay);
		return;
	}

	/* Now we send a ACK packet. First we put whatever is in
	 * the ngb_ackList[] array, then the lsa if there is one.
	 */
	OSPF_PACKET(pck, OSPF_PCK_LSA, maxlen, ngb->ngb_ifap);
	lsh = pck->pck_lsa_lsh;
	pcklen = sizeof(ospf_pck_header_t);

	for(;ngb->ngb_lastAckIndex; ngb->ngb_lastAckIndex--) {
		vtx = ngb->ngb_ackList[ngb->ngb_lastAckIndex - 1];
		LSA_SET((ospf_lsa_t *)lsh, VTX_AGE(vtx), vtx->vtx_options,
			vtx->vtx_type, vtx->vtx_id, vtx->vtx_advrt,
			vtx->vtx_seq, vtx->vtx_cksum, vtx->vtx_len);
		lsh++;
		pcklen += sizeof(ospf_lsa_hdr_t);
	}

	if (lsa) {
		*lsh = *(ospf_lsa_hdr_t *)lsa;
		pcklen += sizeof(ospf_lsa_hdr_t);
	}

	/* Now send the packet!
	 */
	ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf, 
		ospf_send_addr(ngb->ngb_intf, type, ngb));

	/* reset the ack timer, since we just sent one.
	 */
	task_timer_reset(ngb->ngb_ackTimer);
}

/* Flood the LSA to all other neighbours. Return ACKF_FBACK if the advertisement
 * has been flooded back to the receiving interface (used by ospf_send_ack)
 * 'fromngb' is the sender of the LSA. For self-originated LSA, fromngb is NULL.
 */
int
ospf_flood __PF2(fromngb, ospf_ngb_t *, vtx, vertex_t *)
{
	ospf_intf_t *ospf_intf;
	int send_lsu, cmp, floodback = ACKF_NEW;
	ospf_ngb_t *ngb;
	vertex_t *req, *junkvtx;

	/* RFC p79. Go through all interfaces.
	 */
	OSPF_IF_ADDR(ospf_intf) {
		/* Check the neighbours to find out if we send the LSU on that
		 * interface.
		 */
		send_lsu = 0;
		OSPF_INTF_NGB(ngb, ospf_intf) {
			/* If state < EXCHANGE, nothing.
			 */
			if (ngb->ngb_state < OSPF_NGBS_EXCHANGE)
				continue;

			/* The adjacency is not full. Check the LS request list.
			 */
			if (ngb->ngb_state < OSPF_NGBS_FULL &&
					(NGB_LIST_FIND_LSR(req, ngb, vtx))) {
				cmp = ls_cmpage(VTX_AGE(req), req->vtx_seq,
					req->vtx_cksum, VTX_AGE(vtx),
					vtx->vtx_seq, vtx->vtx_cksum);
				/* It's less recent?
				 */
				if (cmp > 0)
					continue;

				/* The same? Remove it from the list.
				 */
				if (!cmp) {
					req = vtx;
					NGB_LIST_DEL_LSR(ngb, vtx);
					ASSERT(!NGB_LIST_FIND_LSR(req, ngb,
						req));
					continue;
				}

				/* The new LSA is more recent.
				 */
				req = vtx;
				NGB_LIST_DEL_LSR(ngb, vtx);
				ASSERT(!NGB_LIST_FIND_LSR(junkvtx, ngb, req));
			}

			/* Is the LSA from this neighbour?
			 */
			if (ngb == fromngb)
				continue;

			/* Add the LSA to the retransmission list.
			 */
			NGB_LIST_ADD_RMX(ngb, vtx->vtx_type, &vtx->vtx_key,vtx);
			NGB_FLUSH_ADD(ngb);
			send_lsu = 1;
		}

		/* If the LSA was not added to any neighbour RMX list, nothing
		 * to do.
		 */
		if (!send_lsu)
			continue;

		/* If we got this LSA from this interface, from the DR or
		 * Backup, nothing to do.
		 */
		if (fromngb && fromngb->ngb_intf == ospf_intf &&
				(NGB_ISDR(fromngb) || NGB_ISBACKUP(fromngb)))
			continue;

		/* Are we backup?
		 */
		if (fromngb && fromngb->ngb_intf == ospf_intf &&
				INTF_ISBACKUP(ospf_intf))
			continue;

		/* Put the LSU on the neigbour's queue for transmission.
		 */
		NGB_LIST_ADD_RMX(fromngb, vtx->vtx_type, &vtx->vtx_key,vtx);
		NGB_FLUSH_ADD(fromngb);

		/* Are we flooding back the LSA on the receiving interface?
		 */
		if (fromngb && fromngb->ngb_intf == ospf_intf)
			floodback = ACKF_FBACK;

	} OSPF_IF_ADDR_END(ospf_intf);

	return(floodback);
}

/* Send all LS waiting to be sent. The LS are in the RMX queue. We send
 * as much as we can in 1 packet, them move them to the retransmit list.
 * This function may be called when a LSA is put in the queue, which could
 * happend when: receiving a packet (ospf_recv()), interface change,
 * a route change (ospf_flash()), the wait timer fires, spf computation, and?
 */
void
ospf_flush_rmx __PF1(ngb, ospf_ngb_t *)
{
	ospf_lsu_pck_t *pck;
	int pcklen, maxlen;
	ospf_lsa_t *lsa;
	vertex_t *vtx;
	int nbAdvt;

	/* Get a packet, type LSU.
         */
        OSPF_PACKET(pck, OSPF_PCK_LSU, maxlen, ngb->ngb_ifap);

        lsa = pck->pck_lsu_lsa;
	pcklen = sizeof(ospf_lsu_pck_t) - sizeof(ospf_lsa_t);
	nbAdvt = 0;

	/* All LSA sent will be put into the buffer for later retransmition.
	 * So we need some space in the buffer.
	 */
	do {
		NGB_LIST_NEXT_RMX(ngb, vtx);

		/* Are we done?
                 */
                if (!vtx)
			break;

		if (ngb->ngb_RMXbuffer_len + vtx->vtx_len >
				maxlen - sizeof(ospf_pck_header_t)) {
			/* Hoop, put it back...
			 */
			NGB_LIST_ADD_RMX(ngb, vtx->vtx_type, &vtx->vtx_key,vtx);
                        break;
		}
 
                /* Put it in the buffer and into the packet to be sent.
		 */
		NGB_LIST_ADD_RMX_BUFF(ngb, vtx->vtx_type, &vtx->vtx_key, vtx);
		ospf_lsa_set(lsa, vtx);
                pcklen += vtx->vtx_len;
		(int)lsa += (int)ntohs(lsa->lsa_len);
		nbAdvt++;
 	} while(1);

	/* Send the packet.
         */
	pck->pck_lsu_nbadv = htonl(nbAdvt);
	ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf, 
		ospf_send_addr(ngb->ngb_intf, SEND_LSU_FLOODING, ngb));
}

/* Handle self originated LSA. Return 1 if it is self-originated and action
 * has been taken.
 */
int
ospf_self_originate __PF4(ngb, ospf_ngb_t *, lsa, ospf_lsa_t *, vtx, vertex_t *,
		searchdata, searchdata_t *)
{
	/* Is it a self-originated LSA?
	 */
	if (lsa->lsa_advrt != ospf_rtid &&
			(lsa->lsa_type != LST_NTW ||
			!if_withlcladdr(sockbuild_in(0, lsa->lsa_id), 0)))
		return(0);

	trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
		("OSPF DB: Got self originated LSA on area %A LS %s: %A",
				sockbuild_in(0, ngb->ngb_area->area_id),
				trace_state(lsa->lsa_type, ospf_lsa_types),
				sockbuild_in(0, lsa->lsa_id)));

	/* vtx is our current version. If it is no more in the DB, then
	 * we have to flush it fron the routing domain.
	 */
	if (LS_NOT_INDB(vtx)) {
		/* If the vertex does not exist, create one, else update it.
		 */
		if (!vtx)
			vtx = ls_vtx_create(&ngb->ngb_tree, &lsa->lsa_key,
				lsa->lsa_options, 0, 0, lsa->lsa_seq, 0,
				lsa->lsa_type, 0, searchdata);
		else
			ospf_ls_update(VTX_TREE(vtx), lsa, vtx);

		/* Set its age to MaxAge, mark it for deletion, and flush it.
		 */
		BIT_SET(vtx->vtx_flags, LF_DELETE);
		VTX_AGE_SET(vtx, ospf_maxAge);
		ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
		return(1);
	}

	/* Increment our own seq number past the vtx's one, and flush it.
	 */
	vtx->vtx_seq = lsa->lsa_seq + 1;
	vtx->vtx_cksum = 0;
	ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
	return(1);
}

/* Process incoming Link State Acknoledgement packet
 */
int
ospf_lsa_recv __PF4(pck, ospf_lsa_pck_t *, len, int, ospf_intf, ospf_intf_t *,
        src, sockaddr_un *)
{
	int nblsh;
	ospf_lsa_hdr_t *lsh;
	ospf_ngb_t *ngb;
	vertex_t buffer, *vtx;

	/* find out about the identity of the neighbour
         */
        if (!(ngb = ospf_find_ngb(ospf_intf, pck->pck_lsa_rtid)))
                return(OSPF_ERR_INVNGB);

	/* Check the state of the neighbour
	 */
	if (ngb->ngb_state < OSPF_NGBS_EXCHANGE)
		/* Ignore it.
		 */
		return(OSPF_NOERR);

	/* Go though all LS headers.
	 */
	nblsh = PCK_LSA_NBLSH(pck->pck_lsa_len);
	lsh = ((ospf_lsa_pck_t *)pck)->pck_lsa_lsh;
	for(; nblsh; nblsh--, lsh++) {

		NTOH_LSH(lsh);

		vtx = &buffer;
		vtx->vtx_type = lsh->lsh_type;
		vtx->vtx_key = lsh->lsh_key;
		
		/* Delete it from the retransmission queue.
		 */
		if (!(vtx = ngb_list_del_rmx_find(ngb, lsh->lsh_type,
				&lsh->lsh_key))) {
			/* The LSU was not in the list. weird...
			 */
			ospf_check_tree(&ngb->ngb_lsRmx, 1);
			ospf_check_tree(&ngb->ngb_RMXbuffer, 1);

			trace_log_tf(ospf_trace_options, 0, LOG_WARNING,
                        ("OSPF LSA: Unsolicited LSA: type %s ID %A",
				trace_state(lsh->lsh_type, ospf_lsa_types),
				sockbuild_in(0, lsh->lsh_id)));
			continue;
		}

		/* Self originated processing. If the LS has been flooded
		 * with a seq number of MAX_LS_SEQ and MaxAge, we may want
		 * to create a new LS starting at MIN_LS_SEQ. The area is
		 * the one of the neighbour.
		 */
		if (BIT_TEST(vtx->vtx_flags, LF_SEQWRAP)) {
			BIT_RESET(vtx->vtx_flags, LF_SEQWRAP);
			vtx->vtx_seq = MIN_LS_SEQ;
			vtx->vtx_cksum = 0;
			VTX_AGE_SET(vtx, 0);
			ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
		}

		/* If the vertex was scheduled for deletion, get rid of it!
		 */
		if (BIT_TEST(vtx->vtx_flags, LF_DELETE))
			if (!--vtx->vtx_qref && !ospf_nbNgbNotReady)
				ospf_ls_delete(vtx, LSDEL_NORM);
	}

	/* If the RMX queue is not empty, move some of it into the
	 * RMX buffer and transmit some of it. Well, it will be done by
	 * ospf_flush_rmx() at the end of the OSPF processing.
	 */

	return(OSPF_NOERR);
}

/* Remove a vtertex from a neighbour retransmission list. Return the vertex if
 * the vertex was there.
 */
vertex_t *
ngb_list_del_rmx __PF2(ngb, ospf_ngb_t *, vtx, vertex_t *)
{
	vertex_t *deadvtx;

	if ((deadvtx = db_del_fast(&ngb->ngb_lsRmx, vtx))){
                TRACE_NGB_LIST(ngb, vtx->vtx_type, &vtx->vtx_key,
			"DEL RMX queue", VTX_AGE(vtx));
                TRACE_NGB_LIST_END((ngb)->ngb_lsRmx.pt_used_nodes);
		return(deadvtx);
	}

	if ((deadvtx = db_del_fast(&ngb->ngb_RMXbuffer, vtx))) {
		TRACE_NGB_LIST(ngb,vtx->vtx_type, &vtx->vtx_key,
			"DEL RMX buffer", VTX_AGE(vtx));
                TRACE_NGB_LIST_END((ngb)->ngb_RMXbuffer.pt_used_nodes);
		ngb->ngb_RMXbuffer_len -= vtx->vtx_len;
		return(deadvtx);
        }
	return(NULL);
}

/* Same as ngb_list_del_rmx() execpt we don't have a pointer to the veretex,
 * just the type and the key. XXXX Yuk. Could be optimized.
 */
vertex_t *
ngb_list_del_rmx_find __PF3(ngb, ospf_ngb_t *, type, int, key, lskey_t *)
{
	vertex_t *deadvtx;
	tree_node_t *node, *barfoo;

	if ((node = tree_find((u_int8 *)key, ospf_key_len[type],
			ngb->ngb_lsRmx.pt_root[type], &barfoo)) &&
			node->tn_data) {
		deadvtx = db_del_fast(&ngb->ngb_lsRmx,
			(vertex_t *)node->tn_data);
		ASSERT(deadvtx);
		return(deadvtx);
	}
	else if ((node = tree_find((u_int8 *)key, ospf_key_len[type], 
			ngb->ngb_RMXbuffer.pt_root[type], &barfoo)) &&
			node->tn_data) {
		deadvtx = db_del_fast(&ngb->ngb_RMXbuffer,
			(vertex_t *)node->tn_data);
		ASSERT(deadvtx);
		return(deadvtx);
	}
	return(NULL);
}

/* Return 1 if the LS of type 'type' from area 'area' in flooded to
 * to neighbour 'ngb'.
 */
int
isflooded __PF3(ngb, ospf_ngb_t *, type, int, area, ospf_area_t *)
{
	switch(type) {
	case LST_RTR:
	case LST_NTW:
		/* Those are flooded only within the same area.
		 */
		return(area == ngb->ngb_area);

	case LST_ASIP:
	case LST_ASBR:
		/* Those are flooded only outside the area.
		 */
		return(area != ngb->ngb_area);

	case LST_ASEX:
		/* Those are flooded to non stub areas, and non virtual links.
		 */
		return(BIT_TEST(ngb->ngb_area->area_options, OSPF_PCK_OPT_E) &&
			!BIT_TEST(ngb->ngb_intf->ospf_if_flags, IFS_OSPFVLINK));

	default:
		/* ???? Bug.
		 */
		ASSERT(0);
		return(0);
	}
}

/* Flash method for the OSPF task.
 */
void
ospf_flash __PF2(tp, task *, change_list, rt_list *)
{
	rt_head *rth;
	rt_entry *rtactive;
	vertex_t *vtx;
	adv_results result;
	ospf_lsa_t lsa;
	searchdata_t searchdata;

	/* Go through all the routes, and see what we can do.
	 */
	RT_LIST(rth, change_list, rt_head) {

		ASEXTREE_BUG("ospf_flash 1");
		/* We don't care about non IP routes (is it really usefull?)
		 */
		if (socktype(rth->rth_dest) != AF_INET)
			continue;

		/* Get rid of some routes, cause we should not deal with those.
		 * Included are OSPF routes and Direct routes.
		 */
		rtactive = rth->rth_active;
		if (rtactive && (rtactive->rt_gwp->gw_proto == RTPROTO_OSPF ||
				rtactive->rt_gwp->gw_proto == RTPROTO_DIRECT ||
				BIT_TEST(rtactive->rt_state, RTS_NOADVISE)))
			continue;

		/* Get our personnal parameter associated with this route.
		 */
		OSPF_TSI_GET(rth, ospf_task->task_rtbit, vtx);

		/* If no vtx exist, we need to create one, and this is a new
		 * route too.
		 */
		if (!vtx) {
			/* No active route? Bye bye...
			 */
			if (!rtactive)
				continue;

			/* Do we want to export this route to the OSPF domain?
			 */
			if (!export(rtactive, 0, ospf_export_list, NULL,
					NULL, &result))
				continue;

			/* Generate a LS External for that route, and flood it.
			 */
			LSA_EXT_SET((ospf_lsa_ext_t *)&lsa,
				SOCK32(rth->rth_dest_mask),
				BIT_TEST(result.res_flag, OSPF_EXPORT_TYPE2)?
					1: 0,
				0, result.res_metric,
				SOCK32(RT_ROUTER(rtactive)),
				result.res_metric2);
			LSA_SET_H(&lsa, 0, ospf_area_list->area_options,
				LST_ASEX, rfc1582apFid(rth), ospf_rtid,
				MIN_LS_SEQ, 0, sizeof(ospf_lsa_ext_t));

			/* Add the self originated LSA to the DB.
			 */
			vtx = NULL;
			ospf_self_add(&vtx, &lsa, &ospf_area_list->area_tree,
				rtactive);

			/* Set the OSPF parameter for that route.
			 */
			OSPF_TSI_SET(rth, ospf_task->task_rtbit, vtx);

			ASEXTREE_BUG("ospf_flash 1.1");
			continue;
		}
		
		/* See what's changed. We care about next hop, metric, tag?
		 * Well, for now, delete the route is the is no more active
		 * route or we don't want it, else update.
		 */
		if (!rtactive || !export(rtactive, 0, ospf_export_list, NULL,
				NULL, &result)) {
			/* Flood it with the max Age.
			 */
			BIT_SET(vtx->vtx_flags, LF_DELETE);
			VTX_AGE_SET(vtx, ospf_maxAge);
                        ospf_flood_selforiginate(ospf_area_list,
                                ospf_area_list->area_rtrls);
			ASEXTREE_BUG("ospf_flash 2");
			continue;
		}
		else {
			/* Do a small update, no call to ospf_ls_update()?
			 */
			VTX_AGE_SET(vtx, 0);
			vtx->vtx_rt = rtactive;
			vtx->vtx_cksum = 0;

			/* Update the self originated LS
			 */
			ospf_self_add(&vtx, NULL, &ospf_area_list->area_tree,
				rtactive);

			ASEXTREE_BUG("ospf_flash 5");
		}	

		ASEXTREE_BUG("ospf_flash 6");
	} RT_LIST_END(rth, change_list, rt_head);
}

/* Algorithm to assign a LS ID to an external route. It is supposed to
 * be appendix F of RFC 1583, but i'll try something simplier.
 */
u_int32
rfc1582apFid __PF1(rth, rt_head *)
{
	/* If this is not the default route, set to 1 all null bits
	 * off the mask. Kind of like appendix F, but we don't care about
	 * RFC1247 compliance when possible.
	 */
	return(SOCK32(rth->rth_dest_mask)?
		SOCK32(rth->rth_dest) | ~SOCK32(rth->rth_dest_mask):
		0);
}

/* New policy for the OSPF task.
 */
void
ospf_newpolicy __PF2(tp, task *, change_list, rt_list *)
{
	/* Basicly do like ospf_flash()?
	 */
	ASEXTREE_BUG("ospf_newpolicy 1");
	ospf_flash(tp, change_list);
	ASEXTREE_BUG("ospf_newpolicy 2");
}




