/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* All database, link state advertissements, etc... stuff.
 * The database is a fully linked (by pointers) set of vertex.
 */
#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"

#define DEBUG_CHECK_VTX(vtx) { \
	ospf_lsa_t __lsa; \
	u_int16 __cksum = (vtx)->vtx_cksum; \
	ospf_lsa_set(&__lsa, vtx); \
	(vtx)->vtx_cksum = __cksum; \
}

#define DEBUG_EMPTYLINK(vtx) { \
	ospf_vtxLink_t *__vl; \
	for(__vl = (vtx)->vtx_links; __vl; __vl = __vl->vl_next) { \
		ASSERT(__vl->vl_vtx); \
	} \
}

ospf_ptwalk_t 	*ospf_ptwalk = NULL;
lsage_t		ospf_maxAge = OSPF_MAXAGE;
lsage_t		ospf_maxAgeDiff = OSPF_MAXAGEDIFF;
lsage_t		ospf_minLSInt = OSPF_MINLSINT;
gw_entry	*ospf_defaultGateway;
time_t		ospf_refresh_start = 0;
task_job	*ospf_refresh_job = NULL;
ospf_ptwalk_t	*ospf_refresh_walk = NULL;
ospf_area_t	*ospf_refresh_area = NULL;
int		ospf_refreshChunkSize = OSPF_REFRESH_CHUNK;

/* A fake route used for the router's RTR LSA.
 */
rt_entry	*ospf_nullrt = NULL;

/* Lengh of the key fo rdatabase lookups. For router and networks LS, only
 * the ID is used.
 */
int ospf_key_len[LST_MAXTYPE] = { 0, 4, 4, 8, 8, 8 };

const bits ospf_vtx_flags[] = {
	{ LF_FOUND,	"Found" },
	{ LF_HIDDEN, 	"Hidden" },
	{ LF_FLUSH,	"Flush" },
	{ LF_REQUEST,	"Req" },
	{ LF_ASYM,	"Asym" },
	{ TNF_EXTERN,	"Extern" },
	{ LF_SEQWRAP,	"SeqWrap" },
	{ LF_DELETE,	"Delete" },
	{ 0, NULL }
};

const bits ospf_lsa_types[] = {
	{ LST_NULL,	"NULL" },
	{ LST_RTR,	"RTR" },
	{ LST_NTW,	"NTW" },
	{ LST_ASIP,	"AIP" },
	{ LST_ASBR,	"ABR" },
	{ LST_ASEX,	"AEX" },
	{ 0, NULL }
};
const bits ospf_lslinks_types[] = {
	{ 0,		"NULL!!" },
	{ LSRLT_P2P,	"P2P" },
	{ LSRLT_TRANS,	"TRANS" },
	{ LSRLT_STUB,	"STUB" },
	{ LSRLT_VLINK, 	"VLINK" },
	{ LSRLT_RTR,	"RTR" },
	{ 0, NULL }
};

/* Add a LSA to the database.
 */
vertex_t *
ospf_ls_add __PF2(tree, db_tree_t *, lsa, ospf_lsa_t *)
{
	vertex_t *vtx;
	searchdata_t searchdata;

	/* Loopup the DB to see if it's already there
	 */
	if ((vtx = db_find(tree, lsa->lsa_type, &lsa->lsa_key, &searchdata))) {
		/* If our data is better, get out!
	 	 */
		if (ls_cmpage(VTX_AGE(vtx), vtx->vtx_seq, vtx->vtx_cksum,
			lsa->lsa_age, lsa->lsa_seq, lsa->lsa_cksum) >= 0)
			return(vtx);

		DEBUG_EMPTYLINK(vtx);

		/* else update the data
	 	 */
		ospf_ls_update(tree, lsa, vtx);
		OSPF_TREE_CHECK(tree);
	}
	else
		vtx = ospf_ls_create(tree, lsa, NULL, &searchdata);

	return(vtx);
}

/* Create an LS. The announcing router and seq number must be there!
 * It's created from a real LSA and includes the links. The LSA header is in
 * the host byte order, all the rest is in network byte order.
 */
vertex_t *
ospf_ls_create __PF4(tree, db_tree_t *, lsa, ospf_lsa_t *, rt, rt_entry *,
		searchdata, searchdata_t *)
{
	vertex_t *vtx;
	int nbRtrs, ind, nbLinks;
	ospf_lsa_rtlink_t *link;
	ospf_vtxLink_t **vlp;
	metric_t metric2 = 0;
	ipaddr_t mask;
	gw_entry *gw = ospf_defaultGateway;
	u_int32 tag = 0;
	int routeit = 1;

	/* General parameters...
	 */
	vtx = ls_vtx_create(tree, &lsa->lsa_key, lsa->lsa_options, 0,
		lsa->lsa_age, lsa->lsa_seq, lsa->lsa_cksum,
		lsa->lsa_type, lsa->lsa_len, searchdata);

	/* Type dependant parameters... The links. They have to be kept in
	 * the same order than in the LSA.
	 */
	switch(lsa->lsa_type) {
	case LST_RTR:
		vtx->vtx_rtr_bits = lsa->lsa_rtr_bits;
		mask = RTMASK;

		/* Add all the links at the end of the chain.
		 */
		nbLinks = NTOHS(lsa->lsa_rtr_nblinks);
		link = lsa->lsa_rtr_links;
		vlp = &vtx->vtx_links;
		while(nbLinks) {
			/* If the RTR is connected to a transit of stub network,
			 * we don't need to create a route to it.
			 */
			if (link->lsrl_type == LSRLT_TRANS ||
					link->lsrl_type == LSRLT_STUB)
				routeit = 0;

			/* Create the link...
			 */
			*vlp = ls_vtxlink_create(NULL, NULL,
					link->lsrl_type,
					link->lsrl_data, 0,
					NTOHS(link->lsrl_metric));

			/* And make sure the reverse link exist.
			 */
			(*vlp)->vl_vtx = ls_reverse_vtx(tree, link->lsrl_id,
				link->lsrl_type, link->lsrl_data,
				link->lsrl_metric, vtx, *vlp);

			(int)link += sizeof(ospf_lsa_rtlink_t)
				+ (link->lsrl_nbtos - 1) *
					sizeof(ospf_tosmetric_t);
			nbLinks--;
			vlp = &(*vlp)->vl_next;
		}
		/* Set the LF_ROUTE bit
		 */
		if (routeit || BIT_TEST(vtx->vtx_flags, LF_HIDDEN))
			BIT_SET(vtx->vtx_flags, LF_ROUTE);
		else
			BIT_RESET(vtx->vtx_flags, LF_ROUTE);
		break;

	case LST_NTW:
		mask = lsa->lsa_ntw_mask;

		/* Add all the routers.
		 */
		nbRtrs = LSA_NTW_NBRTR(lsa, lsa->lsa_len);
		vlp = &vtx->vtx_links;
		for(ind = 0; ind < nbRtrs; ind++) {
			/* Create the link...
                         */
			*vlp = ls_vtxlink_create(NULL, NULL,
				LSRLT_RTR, 0, 0, 0);

			/* And make sure the reverse link exist.
                         */
                        (*vlp)->vl_vtx = ls_reverse_vtx(tree,
				lsa->lsa_ntw_rtid[ind], LSRLT_RTR,
				0, 0, vtx, *vlp);
			vlp = &(*vlp)->vl_next;
		}
		break;

	case LST_ASBR:
		/* We are a gateway. We need to create a gateway entry which
		 * will own some extwernal routes, and store it somewhere....
		 */
		mask = HOSTMASK;
		vtx->vtx_tosmet = NTOHL(lsa->lsa_sla_tosmet32);

		/* Get the gateway which is suposed to own that route.
		 */
		gw = gw_locate(&tree->pt_area->area_gwlist,
				RTPROTO_OSPF, ospf_task, 0, 0,
				sockbuild_in(0, vtx->vtx_advrt), 0);
		break;

	case LST_ASIP:
		mask = lsa->lsa_sla_mask;
		vtx->vtx_tosmet = NTOHL(lsa->lsa_sla_tosmet32);

		/* Get the gateway which is suposed to own that route.
		 */
		gw = gw_locate(&tree->pt_area->area_gwlist,
				RTPROTO_OSPF, ospf_task, 0, 0,
				sockbuild_in(0, vtx->vtx_advrt), 0);
		break;

	case LST_ASEX:
		mask = lsa->lsa_ext_mask;
		vtx->vtx_tosmet = NTOHL(lsa->lsa_ext_tosmet32);
		vtx->vtx_forward = lsa->lsa_ext_forw;
		tag = NTOHL(lsa->lsa_ext_tag);

		/* We do not support TOS. Yuk XXXXXX
		 */
		ASSERT(lsa->lsa_len == sizeof(ospf_lsa_ext_t));

		/* Make the gateway except for self originated LSA.
		 */
		if (lsa->lsa_advrt != ospf_rtid)
			gw = gw_locate(&tree->pt_area->area_gwlist,
				RTPROTO_OSPF, ospf_task, 0, 0,
				sockbuild_in(0, lsa->lsa_ext_forw?
					lsa->lsa_ext_forw: lsa->lsa_advrt), 0);
		break;

	default:
		ASSERT(0);
	}

	/* Not done if rt is already defined.
	 */
	if (!rt)
		vtx->vtx_rt = vtx_get_route(lsa->lsa_id, mask, gw, metric2,
                		tag, vtx);
	else
		vtx->vtx_rt = rt;

	ospf_trace_vtx(vtx, "CREATE");
        OSPF_TREE_CHECK(tree);

	DEBUG_CHECK_VTX(vtx);
	return(vtx);
}

/* Find out what gateway will be associated with the vertex.
 */
gw_entry *
vtx_get_gw __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	switch(vtx->vtx_type) {
	case LST_RTR:
		/* If the bit E or B is set, we want our own gateway which
		 * will contain the summary route sent by that router.
		 */
		if (BIT_TEST(vtx->vtx_rtr_bits, LSRTR_BITS_B|LSRTR_BITS_E))
			return(gw_locate(&area->area_gwlist, RTPROTO_OSPF,
				ospf_task, 0, 0, sockbuild_in(0, vtx->vtx_id),
				0));
		else
			/* We just use the default gateway for that area.
			 */
			return(area->area_defaultgw);
	case LST_NTW:
		/* Use the default one.
		 */
		return(area->area_defaultgw);
	case LST_ASBR:
		/* We are a gateway. We need to create a gateway entry which
                 * will own some extwernal routes, and store it somewhere....
                 */
		return(gw_locate(&area->area_gwlist, RTPROTO_OSPF, ospf_task,
			0, 0, sockbuild_in(0, vtx->vtx_id), 0));
	case LST_ASIP:
		/* Use default.
		 */
		return(area->area_defaultgw);
	case LST_ASEX:
		/* Use default.
                 */
                return(area->area_defaultgw);
	default:
		ASSERT(0);
	}
	return(NULL);
}

/* Create a rt_entry structure to store the (future) route.
 */
rt_entry *
vtx_get_route __PF6(addr, ipaddr_t, mask, ipaddr_t, gw, gw_entry *, metric2,
		metric_t, tag, u_int32, vtx, vertex_t *)
{
	rt_parms rtp;
	rt_entry *rt;

	/* The self originated Router LSA for our router will have a fake
	 * associated route! 'cause when getting deleted, gated looses its
	 * interface! (well, at least looks like that to me).
	 * This route will never be deleted!
	 */
	if (vtx->vtx_type == LST_RTR && vtx->vtx_id == ospf_rtid)
		return(ospf_nullrt);

	rtp.rtp_dest = sockbuild_in(0, addr & mask);
        rtp.rtp_dest_mask = inet_mask_locate(mask);
        rtp.rtp_n_gw = 0;
        rtp.rtp_gwp = gw;
        rtp.rtp_metric = ospf_lsinfinity;
        rtp.rtp_metric2 = metric2;
        rtp.rtp_tag = tag;
        rtp.rtp_state = RTS_INTERIOR;
		/*RTS_HIDDEN|RTS_NOTINSTALL|RTS_NOADVISE|RTS_INTERIOR;*/
        rtp.rtp_preference = -1;	/* Won't be installed/used */
        rtp.rtp_preference2 = 0;
        rtp.rtp_rtd = (void_t)vtx;
#ifdef  PROTO_ASPATHS
        rtp.rtp_asp = NULL;
#endif
	rt = rt_add(&rtp);
	OSPF_RT_VTX(rt) = vtx;

	ASSERT(rt);
	ASSERT(!BIT_TEST(rt->rt_state, RTS_ACTIVE));
        return(rt);
}

/* Create a vertex. No link yet. Put it in the DB.
 */
vertex_t *
ls_vtx_create __PF10(tree, db_tree_t *, key, lskey_t *, options, options_t,
		flags, int, age, lsage_t, seq, lsseq_t, cksum, lscksum_t,
		type, int, len, lslenght_t,
		searchdata, searchdata_t *)
{
	vertex_t *vtx;

        vtx = DB_VTX_ALLOC();
        vtx->vtx_key = *key;
        vtx->vtx_options = options;
        vtx->vtx_flags = flags;
        vtx->vtx_seq = seq;
        vtx->vtx_cksum = cksum;
        vtx->vtx_links = NULL;
        vtx->vtx_type = type;
	vtx->vtx_len = len;
	vtx->vtx_spfcount = tree->pt_area->area_spfcount;
	vtx->vtx_gw = vtx_get_gw(tree->pt_area, vtx);
	VTX_AREA(vtx) = tree->pt_area;
 
	vtx->vtx_nextRecv = 0;
	VTX_AGE_SET(vtx, age);

	/* Add the new LS to the tree. We add it now, so the reverse
         * lookup for the link will find it!
         */
        db_add(tree, type, &vtx->vtx_key, searchdata, vtx, 0);

	tree->pt_area->area_graph.gr_nbVtx[type]++;

	return(vtx);
}


/* Update an already existing LS. From an LSA. The LSA header should be in host
 * byte order, the rest is in network byte order.
 */
void
ospf_ls_update __PF3(tree, db_tree_t *, lsa, ospf_lsa_t *, vtx, vertex_t *)
{
	int nbLinks, nbRtrs, ind;
	ospf_lsa_rtlink_t *link;
	ospf_vtxLink_t **prev_vlp, **vlp, *old_list;
	ospf_vtxLink_t *backLink;
	ipaddr_t mask;
	gw_entry *gw = ospf_defaultGateway;
	metric_t metric2 = 0;
	u_int32 tag = 0;
	int routeit = 1;

	vtx->vtx_cksum = lsa->lsa_cksum;
	vtx->vtx_seq = lsa->lsa_seq;
	vtx->vtx_len = lsa->lsa_len;
	vtx->vtx_options = lsa->lsa_options;
	vtx->vtx_advrt = lsa->lsa_advrt;
	vtx->vtx_spfcount = tree->pt_area->area_spfcount;

	VTX_AGE_SET(vtx, lsa->lsa_age);		/* Change installation time. */

	old_list = vtx->vtx_links;
	vlp = &vtx->vtx_links;
	*vlp = NULL;

	/* If the LSA is real (i.e. advrt exist), clear the DEAD flags,
	 * and the LF_REQUEST.
	 */
	if (lsa->lsa_advrt)
		BIT_RESET(vtx->vtx_flags, LF_REQUEST|LF_HIDDEN|LF_DELETE);

	switch(lsa->lsa_type) {
        case LST_RTR:
		/* Go though all the links, add them if not here yet.
		 * Update the link type, it may have changed.
		 */
		vtx->vtx_rtr_bits = lsa->lsa_rtr_bits;
		mask = RTMASK;

		nbLinks = NTOHS(lsa->lsa_rtr_nblinks);
		link = lsa->lsa_rtr_links;
		while(nbLinks) {
			NTOH_RTLINK(link);

			/* If the RTR is connected to a transit of stub network,
			 * we don't need to create a route to it.
			 */
			if (link->lsrl_type == LSRLT_TRANS ||
					  link->lsrl_type == LSRLT_STUB)
				routeit = 0;

			if ((*vlp = vtxlink_findwithid_prev(&prev_vlp,
					link->lsrl_type,
					link->lsrl_id, &old_list))) {

				/* first we remove the link from the old
				 * list. Then put it in the new one. Then
				 * update the parameters.
				 */
				*prev_vlp = (*vlp)->vl_next;
				(*vlp)->vl_next = NULL;
				(*vlp)->vl_type = link->lsrl_type;
				(*vlp)->vl_data = link->lsrl_data;
				(*vlp)->vl_cost = link->lsrl_metric;
				BIT_RESET((*vlp)->vl_flags, LF_HIDDEN);

				/* update the LT_ASYM status of the link
                                 * and reverse link.
                                 */
				backLink = vtxlink_find((*vlp)->vl_vtx, vtx);
				ASSERT(backLink);
				if (BIT_TEST(backLink->vl_flags, LF_HIDDEN))
					(*vlp)->vl_flags |= LF_ASYM;
				else {
					LINK_MATCH(backLink, *vlp);

					/* Update data for NTW to RTR links.
					 */
					if (backLink->vl_type == LSRLT_RTR)
						backLink->vl_data =
							(*vlp)->vl_data;
				}
			}
			else {
				/* Create a new link, and make sure the
				 * peer and reverse link exist.
				 */
				*vlp = ls_vtxlink_create(NULL, NULL,
						link->lsrl_type,
						link->lsrl_data, 0,
						link->lsrl_metric);
				(*vlp)->vl_vtx = ls_reverse_vtx(tree,
					link->lsrl_id, link->lsrl_type,
					link->lsrl_data, link->lsrl_metric,
					vtx, *vlp);
			}
			(int)link += sizeof(ospf_lsa_rtlink_t)
                                + (link->lsrl_nbtos - 1) *
                                        sizeof(ospf_tosmetric_t);
			vlp = &(*vlp)->vl_next;
			nbLinks--;
		}

		/* Delete all link left in the old list.
		 */
		prev_vlp = vlp;
		for(vlp = &old_list; *vlp;)
			vlp = ls_vtxlink_delete(vlp, vtx);

		/* old_rllist may still contain 'DEAD' link which need
		 * to be added to the vertex's link list.
		 */
		*prev_vlp = old_list;

		/* Set the LF_ROUTE bit
                 */
                if (routeit || BIT_TEST(vtx->vtx_flags, LF_HIDDEN))
                        BIT_SET(vtx->vtx_flags, LF_ROUTE);
                else
                        BIT_RESET(vtx->vtx_flags, LF_ROUTE);
                break;

		/* Do we need to update the vtx_gw? [XXXX]
		 */
		break;

	case LST_NTW:
		mask = lsa->lsa_ntw_mask;

		/* Go though all the routers in the LSA, add them if not there
		 * yet.
		 */
		nbRtrs = LSA_NTW_NBRTR(lsa, lsa->lsa_len);
		for(ind = 0; ind < nbRtrs; ind++) {
			if ((*vlp = vtxlink_findwithid_prev(&prev_vlp,
					LSRLT_RTR, lsa->lsa_ntw_rtid[ind], 
					&old_list))) {

				/* first we remove the link from the old
                                 * list. Then put it in the new one. Then
                                 * update the parameters.
                                 */
				*prev_vlp = (*vlp)->vl_next;
				(*vlp)->vl_next = NULL;
				BIT_RESET((*vlp)->vl_flags, LF_HIDDEN);

				/* update the LT_ASYM status of the link
				 * and reverse link.
				 */
                                backLink = vtxlink_find((*vlp)->vl_vtx, vtx);
                                ASSERT(backLink);
                                if (BIT_TEST(backLink->vl_flags, LF_HIDDEN))
                                        (*vlp)->vl_flags |= LF_ASYM;
                                else {
					LINK_MATCH(*vlp, backLink);
					(*vlp)->vl_data = backLink->vl_data;
				}
			}
			else {
				/* Create a new link, and make sure the
                                 * peer and reverse link exist.
                                 */
				*vlp = ls_vtxlink_create(NULL, NULL,
						LSRLT_RTR, 0, 0, 0);
				(*vlp)->vl_vtx = ls_reverse_vtx(tree,
					lsa->lsa_ntw_rtid[ind],
					LSRLT_RTR, 0, 0, vtx, *vlp);
			}
			vlp = &(*vlp)->vl_next;
		}
		/* Delete all links left in the old list.
		 */
		prev_vlp = vlp;
		for(vlp = &old_list; *vlp;)
			vlp = ls_vtxlink_delete(vlp, vtx);

		/* old_nllist may still contain 'DEAD' link which need
                 * to be added to the vertex's link list.
                 */
                *prev_vlp = old_list;
		break;

	case LST_ASBR:
		/* Update the gateway entry. ??? XXXX
		 */
		mask = HOSTMASK;
		vtx->vtx_tosmet = NTOHL(lsa->lsa_sla_tosmet32);

		/* Make the gateway which is suposed to own that route.
                 */
		gw = gw_locate(&tree->pt_area->area_gwlist, RTPROTO_OSPF,
                        ospf_task, 0, 0, sockbuild_in(0, vtx->vtx_advrt), 0);
		break;

	case LST_ASIP:
		mask = lsa->lsa_sla_mask;
		vtx->vtx_tosmet = NTOHL(lsa->lsa_sla_tosmet32);

		/* Make the gateway which is suposed to own that route.
                 */
		gw = gw_locate(&tree->pt_area->area_gwlist, RTPROTO_OSPF,
                        ospf_task, 0, 0, sockbuild_in(0, vtx->vtx_advrt), 0);
		break;

	case LST_ASEX:
		mask = lsa->lsa_ext_mask;
                vtx->vtx_tosmet = NTOHL(lsa->lsa_ext_tosmet32);
		vtx->vtx_forward = lsa->lsa_ext_forw;
		tag = NTOHL(lsa->lsa_ext_tag);

		if (!vtx->vtx_rt || vtx->vtx_tag != lsa->lsa_ext_tag)
			BIT_SET(vtx->vtx_changes, VC_TAG);

                /* We do not support TOS. Yuk XXXXXX
                 */
                ASSERT(lsa->lsa_len == sizeof(ospf_lsa_ext_t));

                /* Make the gateway except for self originated LSA.
                 */
                if (lsa->lsa_advrt != ospf_rtid)
                        gw = gw_locate(&tree->pt_area->area_gwlist,
                                RTPROTO_OSPF, ospf_task, 0, 0,
                                sockbuild_in(0, lsa->lsa_ext_forw?
                                        lsa->lsa_ext_forw: lsa->lsa_advrt), 0);
                break;

	default:
		ASSERT(0);
	}

	/* If the mask has changed, we delete the old route, create a new one?
	 */
	if (vtx->vtx_rt && mask != vtx->vtx_mask) {
		/* XXXX */
	}

	/* Create a new rt_entry if none exist yet. Else update the parameters.
	 */
	if (!vtx->vtx_rt)
		vtx->vtx_rt = vtx_get_route(lsa->lsa_id, mask, gw, metric2,
                                tag, vtx);
	else {
		vtx->vtx_mask = mask;
		vtx->vtx_tag = tag;
		vtx->vtx_rt_gw = gw;
	}

	ospf_trace_vtx(vtx, "UPDATE");
	DEBUG_CHECK_VTX(vtx);
}

/* Create a link between a router and a net/router.
 */
ospf_vtxLink_t *
ls_vtxlink_create __PF6(nextlinks, ospf_vtxLink_t *, vtx, vertex_t *,
		type, linktype_t, data, u_int32, flags, flag_t, cost, cost_t)
{
	ospf_vtxLink_t *newlink;

	newlink = DB_VTXLINK_ALLOC();
	newlink->vl_next = nextlinks;
	newlink->vl_type = type;
	newlink->vl_data = data;
	newlink->vl_flags = flags;
	newlink->vl_cost = cost;
	newlink->vl_vtx = vtx;

	return(newlink);
}

/* Find the vertex at the other end of the link described by rtid/type/data/...
 * If this vertex does not exist, we create one as well, as the reverse link.
 * The type of the vertex depends on the type of the link.
 * Also updat ethe flags (pflags) of the link.
 * Only the first 4 bytes (ID) of the key are valid.
 */
vertex_t *
ls_reverse_vtx __PF7(tree, db_tree_t *, id, u_int32, type, linktype_t,
		data, u_int32, cost, cost_t, backvtx, vertex_t *,
		backlink, ospf_vtxLink_t *)
{
	ospf_vtxLink_t *vl;
	vertex_t *vtx = NULL;
	searchdata_t searchdata;
	lskey_t key;
	vertex_t *oldvtx;		/* XXX Debug */

	key.lsk_id = id;
	key.lsk_advrt = backvtx->vtx_advrt; /* Does matter when creating */

	switch(type) {
	case LSRLT_RTR:
		/* This is a fake link type, which is supposed to
		 * refers to a link from a network to a router.
		 * So we look for a router.
		 */
		oldvtx = vtx = db_find(tree, LST_RTR, &key, &searchdata);
		if (!vtx) {
			/* We create a vtx, and them a link back to backvtx.
			 */
			vtx = ls_vtx_create(tree, &key, 0, LF_HIDDEN, 0,
				RES_LS_SEQ, 0, LST_RTR, 0, &searchdata);
			vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
					backvtx, LSRLT_TRANS, 0, LF_HIDDEN, 0);
			backlink->vl_flags |= LF_ASYM;
			break;
		}

		/* Make sure that the reverse link exist from vtx to backvtx.
		 * This link can be either LSRLT_TRANS or LSRLT_STUB.
		 */
		if (!(vl = vtxlink_find(backvtx, vtx))) {
			vtx->vtx_links = ls_vtxlink_create( vtx->vtx_links,
				backvtx, LSRLT_TRANS, 0, LF_HIDDEN, 0);
			backlink->vl_flags |= LF_ASYM;
		}
		else if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
			/* Update the reverse link data
		 	 */
			backlink->vl_flags |= LF_ASYM;
		else {
			LINK_MATCH(backlink, vl);
			backlink->vl_data = vl->vl_data;
		}
		/* The hidden vtx are always routed.
		 */
		if (BIT_TEST(vtx->vtx_flags, LF_HIDDEN))
			BIT_SET(vtx->vtx_flags, LF_ROUTE);

		break;

	case LSRLT_P2P:
	case LSRLT_VLINK:
		/* We have a link from a router to a router. We look for a
		 * router. Only the type of the link difers between P2P and
		 * VLINK.
		 */
		oldvtx = vtx = db_find(tree, LST_RTR, &key, &searchdata);
                if (!vtx) {
                        /* We create a vtx, and them a link back to backvtx.
                         */
                        vtx = ls_vtx_create(tree, &key, 0, LF_HIDDEN, 0,
				RES_LS_SEQ, 0, LST_RTR, 0, &searchdata);
                        vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
                                backvtx, type, 0, LF_HIDDEN, 0);
			backlink->vl_flags |= LF_ASYM;
                        break;
                }

                /* Make sure that the reverse link exist from vtx to backvtx.
                 * This link is LSRLT_P2P/LSRLT_VLINK.
                 */
                if (!(vl = vtxlink_find(backvtx, vtx))) {
                        vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
                                backvtx, type, 0, LF_HIDDEN, 0);
			backlink->vl_flags |= LF_ASYM;
                }
		else if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
			backlink->vl_flags |= LF_ASYM;
                break;

	case LSRLT_STUB:
		/* This is a link between a router and a network, the link
		 * data is the mask. We are looking for a network. Stub
		 * network won't be announced with an LSA, so we need
		 * to create a vertex and the associated route (rt_entry)
		 */
		oldvtx = vtx = db_find(tree, LST_NTW, &key, &searchdata);
		if (!vtx) {
                        /* We create a vtx, and them a link back to backvtx.
                         */
                        vtx = ls_vtx_create(tree, &key, 0, LF_HIDDEN, 0,
				RES_LS_SEQ, 0, LST_NTW, 0, &searchdata);
			vtx->vtx_rt = vtx_get_route(key.lsk_id, data,
					ospf_defaultGateway, 0, 0, vtx);
			vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
					backvtx, LSRLT_RTR,  backlink->vl_data,
					LF_HIDDEN, backlink->vl_cost);
			backlink->vl_flags |= LF_ASYM;
			break;
		}

		/* Make sure that the reverse link exist from vtx to backvtx.
                 */
                if (!(vl = vtxlink_find(backvtx, vtx))) {
			vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
				backvtx, LSRLT_RTR, backlink->vl_data,
				LF_HIDDEN, backlink->vl_cost);
			backlink->vl_flags |= LF_ASYM;
		}
		else if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
			backlink->vl_flags |= LF_ASYM;
		break;

	case LSRLT_TRANS:
		/* This is a link between a router and a network.
                 */
		oldvtx = vtx = db_find(tree, LST_NTW, &key, &searchdata);
                if (!vtx) {
                        /* We create a vtx, and them a link back to backvtx.
                         */
                        vtx = ls_vtx_create(tree, &key, 0, LF_HIDDEN, 0,
				RES_LS_SEQ, 0, LST_NTW, 0, &searchdata);
                        vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
				backvtx, LSRLT_RTR, backlink->vl_data,
                                LF_HIDDEN, backlink->vl_cost);
			backlink->vl_flags |= LF_ASYM;
                        break;
                }

                /* Make sure that the reverse link exist from vtx to backvtx.
                 */
                if (!(vl = vtxlink_find(backvtx, vtx))) {
                        vtx->vtx_links = ls_vtxlink_create(vtx->vtx_links,
				backvtx, LSRLT_RTR, backlink->vl_data,
                                LF_HIDDEN, backlink->vl_cost);
			backlink->vl_flags |= LF_ASYM;
                }
		else if (BIT_TEST(vl->vl_flags, LF_HIDDEN))
			backlink->vl_flags |= LF_ASYM;
                break;
	default:
		ASSERT(0);
		break;
	}

	if (!oldvtx)
		ospf_trace_vtx(vtx, "REVCREAT");
	else
		ospf_trace_vtx(vtx, "REVUPD");
	OSPF_TREE_CHECK(tree);

	return(vtx);
}

/* For a network we stop the search at the sortest mask match.
 * If flag is PTF_SHORTER we stop at the shorter mask match.
 */
vertex_t *
db_find __PF4(tree, db_tree_t *, type, int, key, lskey_t *,
		searchdata, searchdata_t *)
{
	tree_node_t *node;

	node = tree_find((u_int8 *)key, ospf_key_len[type], tree->pt_root[type],
		&searchdata->sd_last);

#ifdef OSPF_DEBUG_FIND
	trace_only_tf(ospf_trace_options, 0,
		("OSPF db_find looking for %A/%A (%d): %s %A",
			sockbuild_in(0, key->lsk_id),
			sockbuild_in(0, key->lsk_advrt),
			ospf_key_len[type],
			(node && TREENODE_VTX(node))?
				"Found":
				"Not found at",
			(node && TREENODE_VTX(node))?
				sockbuild_str(""):
			    	(!searchdata->sd_last)?
					sockbuild_str("ROOT"):
					sockbuild_in(0,
					    searchdata->sd_last->tn_key[0])));
#endif
	return(node? TREENODE_VTX(node): NULL);
}

/* Add a new node to the tree
 */
void
db_add __PF6(tree, db_tree_t *, type, int, key, lskey_t *,
		searchdata, searchdata_t *, vtx, vertex_t *, flags, flag_t)
{
	if (BIT_TEST(flags, TNF_EXTERN))
		tree_add((u_int8 *)key, ospf_key_len[type],&tree->pt_root[type],
			NULL, searchdata->sd_last, (void_t)vtx);
	else
		tree_add((u_int8 *)key, ospf_key_len[type],&tree->pt_root[type],
			(tree_node_t *)vtx, searchdata->sd_last, NULL);

	tree->pt_used_nodes++;
}

/* Same as db_add(), except we did not do a search before, and no
 * searchdata is available. So we do a serach first, and a add.
 * This function is only for TNF_EXTERN nodes.
 */
void
db_add_direct __PF4(tree, db_tree_t *, type, int, key, lskey_t *,
		vtx, vertex_t *)
{
	searchdata_t searchdata;
	tree_node_t *node;

        OSPF_TREE_CHECK(tree);

	if ((node = tree_find((u_int8 *)key, ospf_key_len[type],
			tree->pt_root[type], &searchdata.sd_last))) {
		/* If there is no associated vertex, add it.
		 */
		ASSERT(BIT_TEST(node->tn_flags, TNF_EXTERN));

		if (!node->tn_data)
			node->tn_data = (void_t)vtx;
	}
	else {
#ifdef OSPF_DEBUG_FIND
		trace_only_tf(ospf_trace_options, 0,
                ("OSPF db_add_direct looking for %A/%A (%d): %s %A",
                        sockbuild_in(0, key->lsk_id),
                        sockbuild_in(0, key->lsk_advrt),
                        ospf_key_len[type],
                        node? "Found": "Not found at",
                        node? sockbuild_str(""):
                            (!searchdata.sd_last)?
                                sockbuild_str("ROOT"):
                                sockbuild_in(0,
                                        searchdata.sd_last->tn_key[0])));
#endif

		tree_add((u_int8 *)key, ospf_key_len[type],&tree->pt_root[type],
			NULL, searchdata.sd_last, (void_t)vtx);
	}

	tree->pt_used_nodes++;

	OSPF_TREE_CHECK(tree);
}

/* Delete a node in a tree. Used the fast algorithm for external nodes only.
 */
vertex_t *
db_del_fast __PF2(tree, db_tree_t *, vtx, vertex_t *)
{
	vtx = (vertex_t *)tree_del_fast(&tree->pt_root[vtx->vtx_type],
		(u_int8 *)&vtx->vtx_key, ospf_key_len[vtx->vtx_type],
		(void_t)vtx);
	if (vtx)
		tree->pt_used_nodes--;
	return(vtx);
}

/* Delete a node in the tree. Used complex algorithm for internal nodes
 * only (i.e. node where vertex = node).
 */
void
db_delete __PF1(vtx, vertex_t *)
{
	trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
		("OSPF DB: Deleting: %s", ls_vtx_name(vtx)));

	OSPF_TREE_CHECK(VTX_TREE(vtx));

	/* first delete the node...
	 */
	tree_node_del(&VTX_TREE(vtx)->pt_root[vtx->vtx_type],
		(u_int8 *)&vtx->vtx_key, ospf_key_len[vtx->vtx_type],
		(tree_node_t *)vtx);

	/* Update all ospf_ptwalk_t which may point to it.
	 */
	ospf_ptwalk_delnode(VTX_AREA(vtx), (tree_node_t *)vtx);

	/* And free the vertex...
	 */
	VTX_TREE(vtx)->pt_used_nodes--;

        OSPF_TREE_CHECK(VTX_TREE(vtx));

	DB_VTX_FREE(vtx);
}

/* Delete a network/router link from the link list of a vertex.
 * *deadvlp is set to the next link.
 */
ospf_vtxLink_t **
ls_vtxlink_delete __PF2(deadvlp, ospf_vtxLink_t **, vtx, vertex_t *)
{
	ospf_vtxLink_t **vlp;
	ospf_vtxLink_t *lost;

	/* first we ned to find the reverse link from the link's vertex to
	 * the vertex.
	 */
	for(vlp = &(*deadvlp)->vl_vtx->vtx_links; *vlp; vlp = &(*vlp)->vl_next){
		if ((*vlp)->vl_vtx != vtx)
			continue;

		/* We go it. If it is DEAD, then we remove both links.
		 */
		if (BIT_TEST((*vlp)->vl_flags, LF_HIDDEN)) {
			/* First the peer's link...
			 */
			lost = *vlp;
			*vlp = (*vlp)->vl_next;
			DB_VTXLINK_FREE(lost);

			/* If the peer (rotue) has no more links, and
			 * it is marked as DEAD, remove it from the DB.
			 * Note: ospf_ls_delete() will try to remove
			 * all links, even if we know that there're all gone!
			 */
			if (!(*deadvlp)->vl_vtx->vtx_links &&
				BIT_TEST((*deadvlp)->vl_vtx->vtx_flags,LF_HIDDEN))
				ospf_ls_delete((*deadvlp)->vl_vtx, LSDEL_NORM);

			/* Then the link.
			 */
			lost = *deadvlp;
			*deadvlp = (*deadvlp)->vl_next;
			DB_VTXLINK_FREE(lost);

			return(deadvlp);
		}
		else {
			/* Just mark it as DEAD.
			 */
			BIT_SET((*deadvlp)->vl_flags, LF_HIDDEN);
			return(&(*deadvlp)->vl_next);
		}
	}

	/* If we reach that part, the data is corrupted.
	 */
	ASSERT(0);
	return(NULL);
}

/* Find a generic vertex link (ospf_vtxLink_t) from vtx_src to vtx_dst.
 */
ospf_vtxLink_t *
vtxlink_find __PF2(vtx_src, vertex_t *, vtx_dst, vertex_t *)
{
	ospf_vtxLink_t *vl;

	for(vl = vtx_src->vtx_links; vl; vl = vl->vl_next)
		if (vl->vl_vtx == vtx_dst)
			return(vl);
	return(NULL);
}

/* Find an vtx link from the vertex links list, given its ID and type.
/* As for the type, LSRLT_STUB and
 * LSRLT_TRANS will be considered the same, and LSRLT_VLINK and LSRLT_P2P too.
 */
ospf_vtxLink_t *
vtxlink_findwithid __PF3(type, linktype_t, id, lsid_t, vtx, vertex_t *)
{
	ospf_vtxLink_t *vl;

	for(vl = vtx->vtx_links; vl; vl = vl->vl_next)
		if (vl->vl_id == id &&
			((vl->vl_type == type) ||
			 (vl->vl_type == LSRLT_STUB && type == LSRLT_TRANS) ||
			 (vl->vl_type == LSRLT_TRANS && type == LSRLT_STUB) ||
			 (vl->vl_type == LSRLT_VLINK && type == LSRLT_P2P) ||
			 (vl->vl_type == LSRLT_P2P && type == LSRLT_VLINK)))
			return(vl);
	return((ospf_vtxLink_t *)NULL);
}

/* The same but we also return the previous pointer in the list.
 */
ospf_vtxLink_t *
vtxlink_findwithid_prev __PF4(prev, ospf_vtxLink_t ***, type, linktype_t,
		id, lsid_t, linkList, ospf_vtxLink_t **)
{
	for(*prev = linkList; **prev; *prev = &(**prev)->vl_next)
		if ((**prev)->vl_id == id &&
                     (((**prev)->vl_type == type) ||
                      ((**prev)->vl_type == LSRLT_STUB && type == LSRLT_TRANS)||
                      ((**prev)->vl_type == LSRLT_TRANS && type == LSRLT_STUB)||
                      ((**prev)->vl_type == LSRLT_VLINK && type == LSRLT_P2P)||
                      ((**prev)->vl_type == LSRLT_P2P && type == LSRLT_VLINK)))
                      return(**prev);
        return((ospf_vtxLink_t *)NULL);
}

/* check the database. Look for assymetric links between 2 vertices A and B
 * (i.e. A points to B, but B do not points to A). Return the error.
 * It's an atomic operation, at the end all vertices and links must have
 * been marked as FOUND. We also check for 'detached' parts of the graph (i.e
 * all vertices not marked FOUND, but found on the tree.
 */
int
ospf_check_db __PF1(area, ospf_area_t *)
{
	vertex_t *vtx;

	/* first check the links...
	 */
	(void)rec_check_vtx(area->area_graph.gr_root);

	/* them go through the tree and check whatever is cool (and reset
	 * the FOUND flag).
	 */
	PT_WALK(&area->area_tree, vtx) {
		/* We want the FOUND flag on. Else it's a dead part of the
		 * graph.
		 */
		if (BIT_TEST(vtx->vtx_flags, LF_FOUND))
			BIT_RESET(vtx->vtx_flags, LF_FOUND);
		else
			trace_only_tf(ospf_trace_options, 0,
				("OSPF DB: Lost graph node %s",
				ls_vtx_name(vtx)));

	} PT_WALK_END(area->area_tree, vtx);

	return(0);
}

/* recursive function to check the graph.
 */
int
rec_check_vtx __PF1(vtx, vertex_t *)
{
        ospf_vtxLink_t *vl;

	if (vtx->vtx_type != LST_RTR && vtx->vtx_type != LST_NTW)
		return(0);

	/* go though all the links...
	 */
	for(vl = vtx->vtx_links; vl; vl = vl->vl_next) {
		if (BIT_TEST(vl->vl_flags, LF_FOUND))
			continue;

		BIT_SET(vl->vl_flags, LF_FOUND);
		switch(check_link(vl->vl_vtx, vtx, vl->vl_flags)) {
		case CKL_MISSING:
			/* No back link. ho ho!!!!
			 */
			trace_only_tf(ospf_trace_options, 0,
			("OSPF DB: No back link from %s to %s",
				ls_vtx_name(vtx),
				ls_vtx_name(vl->vl_vtx)));
			break;
		case CKL_DEAD:
			/* The back link is dead.
			 */
			trace_only_tf(ospf_trace_options, 0,
			("OSPF DB: Inconsistant link from %s to %s",
				ls_vtx_name(vtx),
                                       ls_vtx_name(vl->vl_vtx)));
                               break;
		default:
			break;
		}

		/* now we check the vertex itself
		 */
		(void)rec_check_vtx(vl->vl_vtx);
	}
	return(0);
}

/* check one link to a vertex. toVtx is the remote vertex, fromVtx is the
 * local (the origine of the link)
 */
int
check_link __PF3(toVtx, vertex_t *, fromVtx, vertex_t *, flags, flag_t)
{
	ospf_vtxLink_t *vl;

	/* Just go through all the links of toVtx, and look for fromVtx.
	 */
	for(vl = toVtx->vtx_links; vl; vl = vl->vl_next)
		if (vl->vl_vtx == fromVtx)
			if (BIT_TEST(vl->vl_flags, LF_HIDDEN) !=
					BIT_TEST(flags, LF_HIDDEN))
				return(CKL_DEAD);
			else
				return(CKL_OK);
	return(CKL_MISSING);
}

/* Generate an LSA for our own router.
 */
void
ospf_gen_rtrlsa __PF1(area, ospf_area_t *)
{
	ospf_lsa_t lsa;
	ospf_lsa_rtlink_t *lsartlink;
	ospf_ngb_t *ngb;
	ospf_intf_t *ospf_intf;
	lsseq_t seq;
	int nbLinks;
	searchdata_t searchdata;

	trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
		("OSPF DB: Generating LS for area %A router %A",
		sockbuild_in(0, area->area_id), sockbuild_in(0, ospf_rtid)));

	nbLinks = 0;
	lsartlink = &lsa.lsa_rtr_links[0];

	/* We go through all the interfaces in this area, as in RFC1583p68.
	 */
	OSPF_IF_ADDR_AREA(area, ospf_intf) {

		switch(ospf_intf->ospf_if_state) {
		case OSPF_IFS_UNDEF:
		case OSPF_IFS_NOOSPF:
		case OSPF_IFS_DOWN:
			/* Skip it.
		 	 */
			break;

		case OSPF_IFS_P2P:
			/* point 2 point interfaces (includes virtual)...
		  	 */
			ngb = ospf_intf->ospf_if_ngb;
			if (ngb && ngb->ngb_state == OSPF_NGBS_FULL) {
				/* ... with fully adjacent neighbour
				 */
				/* XXXX won't work with Virtual intf */
				LSA_RL_SET(lsartlink, ngb->ngb_rtid,
					(ospf_intf->ospf_if_socklocaddr)?
						ospf_intf->ospf_if_locaddr:
				ospf_intf->ospf_if_ifap->ifa_link->ifl_index,
					BIT_TEST(ospf_intf->ospf_if_flags,
						IFS_OSPFVLINK)?
						LSRLT_P2P: LSRLT_VLINK,
					0, ospf_intf->ospf_if_cost);
				nbLinks++;
				(int)lsartlink += sizeof(ospf_lsa_rtlink_t) -
					sizeof(ospf_tosmetric_t);
			}
			else if (ngb) {
				/* no neighbour yet. Stub network.
				 */
				LSA_RL_SET(lsartlink, ngb->ngb_addr,
					HOSTMASK, LSRLT_STUB, 0,
					ospf_intf->ospf_if_cost);
				nbLinks++;
				(int)lsartlink += sizeof(ospf_lsa_rtlink_t) -
					sizeof(ospf_tosmetric_t);
			}
			break;

		case OSPF_IFS_LOOPBACK:
			/* Don't care about loopback.
		 	 */
			break;

		case OSPF_IFS_DROTHER:
		case OSPF_IFS_BACKUPDR:
		case OSPF_IFS_DR:
			/* We look for a fully adjacent neighbour...
			 */
			OSPF_INTF_NGB(ngb, ospf_intf)
				if (ngb->ngb_state == OSPF_NGBS_FULL)
					break;
			if (ngb) {
				/* This is a transit network...
				 */
				LSA_RL_SET(lsartlink,ospf_intf->ospf_if_drAddr,
					ospf_intf->ospf_if_addr, LSRLT_TRANS,
					0, ospf_intf->ospf_if_cost);
				nbLinks++;
				(int)lsartlink += sizeof(ospf_lsa_rtlink_t) -
					sizeof(ospf_tosmetric_t);
				break;
			}

			/* Fall through....
			 */
			
		case OSPF_IFS_WAITING:
			/* If waiting, add a stub network.
		 	 */
			LSA_RL_SET(lsartlink, IFNET(ospf_intf),
			    ospf_intf->ospf_if_mask, LSRLT_STUB, 0,
			    ospf_intf->ospf_if_cost);
			nbLinks++;
			(int)lsartlink += sizeof(ospf_lsa_rtlink_t) -
				sizeof(ospf_tosmetric_t);
			break;

		default:
			ASSERT(0);
		}
	} OSPF_IF_ADDR_AREA_END(area, ospf_intf);

	/* Update/create the self originated LSA.
	 */
	LSA_RT_SET(&lsa.lsa_rt, area->area_rtrbits, nbLinks);
	LSA_SET_H(&lsa, 0, 0, LST_RTR, ospf_rtid, ospf_rtid, seq, 0,
		(int)lsartlink - (int)&lsa);

	ospf_self_add(&area->area_rtrls, &lsa, &area->area_tree, NULL);

	/* this vertex is the root of the LS graph.
	 */
	area->area_graph.gr_root = area->area_rtrls;

	/* Schedule a SPF algorithm computation for the graph.
	 */
	if (!ospf_nbNgbNotReady)
		ospf_spf_schedule();
}

/* Generate a network LSA. For a network that is connected to us, and
 * we are the DR, and there is at least 1 full neighbour.
 */
void
ospf_gen_ntwlsa __PF1(ospf_intf, ospf_intf_t *)
{
	ospf_lsa_t lsa;
	u_int32 *ntlist;
	ospf_ngb_t *ngb;
	lsseq_t seq;
	searchdata_t searchdata;
	
	/* Go through all the neighbourg in FULL state, add them to the
	 * LSA. Don't forget ourself.
	 */
	ntlist = lsa.lsa_ntw_rtid;
	LSA_NL_SET(ntlist, ospf_rtid);
	ntlist++;

	OSPF_INTF_NGB(ngb, ospf_intf) {
                if (ngb->ngb_state != OSPF_NGBS_FULL)
			continue;
		LSA_NL_SET(ntlist, ngb->ngb_rtid)	
		ntlist++;
	}

	/* Ok, if ntlist is only 1, that means thare is no fully adjacent
	 * neighbour, so do not announce the net.
	 */
	if (ntlist - lsa.lsa_ntw_rtid <= 1)
		return;

	trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
		("OSPF DB: Generating LS for area %A network %A",
		sockbuild_in(0, ospf_intf->ospf_if_area->area_id),
		sockbuild_in(0, ospf_intf->ospf_if_addr)));

	/* Update/Create the self originated LSA.
	 */
	LSA_NT_SET(&lsa.lsa_nt, ospf_intf->ospf_if_mask);
	LSA_SET_H(&lsa, 0, 0, LST_NTW, ospf_intf->ospf_if_addr, ospf_rtid, seq,
		0, (int)ntlist - (int)&lsa);

	ospf_self_add(&ospf_intf->ospf_if_ntls, &lsa, &ospf_intf->ospf_if_tree,
		NULL);

	if (!ospf_nbNgbNotReady)
		ospf_spf_schedule();
}

/* flush a network LS when we are not the DR anymore.
 */

/* return 0 if there are the same, 1 if the first one is more recent, -1
 * if the first one is older.
 */
int
ls_cmpage __PF6(age1, lsage_t, seq1, lsseq_t, cksum1, lscksum_t,
		age2, lsage_t, seq2, lsseq_t, cksum2, lscksum_t)
{
	if (seq1 > seq2)
		return(1);
	if (seq1 < seq2)
		return(-1);

	/* now seq1 == seq2
	 */
	if (cksum1 > cksum2)
		return(1);
	if (cksum1 < cksum2)
		return(-1);

	/* now cksum1 == cksum2
	 */
	if (age1 >= ospf_maxAge) {
		if (age2 < ospf_maxAge)
			return(1);
		return(0);
	}
	if (age2 >= ospf_maxAge) {
		if (age1 < ospf_maxAge)
			return(-1);
		return(0);
	}
	if (AGEDIFF(age1, age2) > ospf_maxAgeDiff)
		if (age1 < age2)
			return(1);
		else
			return(-1);

	return(0);
}

/* Check the tree. Print it if flag is non null.
 */
void
ospf_check_tree __PF2(tree, db_tree_t *, flag, int)
{
	tree_node_t *node;
	int corrupted = 0;
	searchdata_t searchdata;

	if (flag)
		trace_only_tf(ospf_trace_options, 0,
("OSPF TREE: Dumping the whole tree RTR %p (%p) NTW %p (%p)  AIP %p (%p) ASBR %p (%p) ASEX %p (%p)",
			tree->pt_root[LST_RTR], &tree->pt_root[LST_RTR],
			tree->pt_root[LST_NTW], &tree->pt_root[LST_NTW],
			tree->pt_root[LST_ASIP], &tree->pt_root[LST_ASIP],
			tree->pt_root[LST_ASBR], &tree->pt_root[LST_ASBR],
			tree->pt_root[LST_ASEX], &tree->pt_root[LST_ASEX]));

	PT_WALK_NODE(tree, node) {
		vertex_t *vtx = (vertex_t *)node;

		ASSERT(BIT_TEST(node->tn_flags, TNF_EXTERN) ||
			!vtx->vtx_rt || vtx->vtx_rt->rt_head);

		if (!flag)
			continue;

		if (BIT_TEST(node->tn_flags, TNF_EXTERN)) {
			trace_only_tf(ospf_trace_options, 0,
		("OSPF TREE: node %s (%d) bit [%d:%X] %A:%A <%s> %p %p %p",
					trace_state(ospf_lsa_types, __type),
					__type,
					node->tn_bitindex.bi_byte,
					node->tn_bitindex.bi_bits,
					sockbuild_in(0, node->tn_key[0]),
					sockbuild_in(0, node->tn_key[1]),
					trace_bits(ospf_vtx_flags,
						node->tn_flags),
					node, node->tn_left, node->tn_right));
			if (node->tn_data)
				trace_only_tf(ospf_trace_options, 0,
					("OSPF TREE: \t=> VTX %s (%p)",
					ls_vtx_name((vertex_t *)node->tn_data),
					node->tn_data));
			else
				trace_only_tf(ospf_trace_options, 0,
					("OSPF TREE: \t=> *** Empty ***"));
		}
		else
			ospf_trace_vtx(vtx, "TREE");

		/* Consistancy check of the tree. We should be able to find
		 * all nodes with associated vtx.
		 */
		if (TREENODE_VTX(node) && !db_find(tree, __type,
				(lskey_t *)&node->tn_key, &searchdata))
			corrupted++;

	} PT_WALK_NODE_END(tree, node);

	if (flag)
		trace_only_tf(ospf_trace_options, 0,
                	("OSPF TREE: end of dump"));

	ASSERT(!corrupted);
}

/* Some test on the LSA import into the DB. Create a small graph,
 * R1 <- Transit-> N1 <- Transit -> R2 <- P2P -> R1 (loop)
 */
void
ospf_test_db()
{
	ospf_lsa_rtlink_t *link;
	ospf_lsa_t lsa;
	vertex_t *r1, *r2, *r3, *n1;

#define NETDOT(a,b,c,d)	(htonl(((a) << 24) + ((b) << 16) + ((c) << 8) + (d)))

	/* R1
	 */
	trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: RTR 1.2.3.4"));

	rt_open(ospf_task);

	link = lsa.lsa_rtr_links;
	LSA_RL_SET(link, NETDOT(1,2,3,0), NETDOT(1,2,3,4),
			LSRLT_TRANS, 0, 1);
	(int)link += sizeof(ospf_lsa_rtlink_t) - sizeof(ospf_tosmetric_t);
	LSA_RL_SET(link, NETDOT(1,2,3,7), 0, LSRLT_P2P, 0, 1);
	LSA_RT_SET(&lsa.lsa_rt, 0, 2);
	LSA_SET_H(&lsa, 0, 0, LST_RTR, NETDOT(1,2,3,4), NETDOT(1,2,3,4),
		MIN_LS_SEQ, 0, 48);
	r1 = ospf_ls_add(&ospf_area_list->area_tree, &lsa);

        /* N1
         */
        trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: NTW 1.2.3/24"));

        lsa.lsa_ntw_rtid[0] = NETDOT(1,2,3,4);
        lsa.lsa_ntw_rtid[1] = NETDOT(1,2,3,7);
        lsa.lsa_ntw_mask = NETDOT(255,255,255,0);
        LSA_SET_H(&lsa, 0, 0, LST_NTW, NETDOT(1,2,3,0), NETDOT(1,2,3,0),
                MIN_LS_SEQ, 0, 32);
        n1 = ospf_ls_add(&ospf_area_list->area_tree, &lsa);

	/* R2
	 */
	trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: RTR 1.2.3.7"));

	link = lsa.lsa_rtr_links;
	LSA_RL_SET(link, NETDOT(1,2,3,0), NETDOT(1,2,3,7),
			LSRLT_TRANS, 0, 1);
	(int)link += sizeof(ospf_lsa_rtlink_t) - sizeof(ospf_tosmetric_t);
	LSA_RL_SET(link, NETDOT(1,2,3,4), 0, LSRLT_P2P, 0, 1);
	LSA_RT_SET(&lsa.lsa_rt, 0, 2);
	LSA_SET_H(&lsa, 0, 0, LST_RTR, NETDOT(1,2,3,7), NETDOT(1,2,3,7),
		MIN_LS_SEQ, 0, 48);
	r2 = ospf_ls_add(&ospf_area_list->area_tree, &lsa);

	/* R3
	 */
	LSA_RT_SET(&lsa.lsa_rt, 0, 0);
        LSA_SET_H(&lsa, 0, 0, LST_RTR, NETDOT(1,2,3,10), NETDOT(1,2,3,10),
                MIN_LS_SEQ+1, 0, 24);
        r3 = ospf_ls_add(&ospf_area_list->area_tree, &lsa);


	/* Now we delete R2
	 */
	trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: Deleting 1.2.3.7"));
        /* R1
         */
        trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: RTR 1.2.3.4"));

        link = lsa.lsa_rtr_links;
        LSA_RL_SET(link, NETDOT(1,2,3,0), NETDOT(255,255,255,0),
                        LSRLT_STUB, 0, 1);
        LSA_RT_SET(&lsa.lsa_rt, 0, 1);
        LSA_SET_H(&lsa, 0, 0, LST_RTR, NETDOT(1,2,3,4), NETDOT(1,2,3,4),
                MIN_LS_SEQ+1, 0, 36);
        (void)ospf_ls_add(&ospf_area_list->area_tree, &lsa);

        /* N1
         */
        trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: NTW 1.2.3/24"));

        lsa.lsa_ntw_rtid[0] = NETDOT(1,2,3,4);
        lsa.lsa_ntw_mask = NETDOT(255,255,255,0);
        LSA_SET_H(&lsa, 0, 0, LST_NTW, NETDOT(1,2,3,0), NETDOT(1,2,3,0),
                MIN_LS_SEQ+1, 0, 28);
        ospf_ls_add(&ospf_area_list->area_tree, &lsa);

        /* R2
         */
        trace_only_tf(ospf_trace_options, 0, ("OSPF DEBUG: RTR 1.2.3.7"));
	ospf_ls_delete(r2, LSDEL_NORM);
	ospf_ls_delete(r3, LSDEL_NORM);
	ospf_ls_delete(r1, LSDEL_NORM);
	ospf_ls_delete(n1, LSDEL_NORM);

        /*LSA_RT_SET(&lsa.lsa_rt, 0, 0);
        LSA_SET_H(&lsa, 0, 0, LST_RTR, NETDOT(1,2,3,7), NETDOT(1,2,3,7),
                MIN_LS_SEQ+1, 0, 24);
        ospf_ls_add(&ospf_area_list->area_tree, &lsa);*/

	rt_close(ospf_task, NULL, 0, "OSPF DB test");
}

/* Schedule a SPF computation for all areas, full graph.
 */
void
ospf_spf_schedule __PF0(void)
{
	tracef("OSPF SPF: Scheduling a SPF comptutaion: ");

	/* If a SPF is already scheduled, give up.
	 */
	if (BIT_TEST(ospf_spf_timer->task_timer_flags, TIMERF_PROCESSING)) {
		trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
			("Already scheduled"));
		return;
	}

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
		("Set in %d seconds", ospf_spfInt));
	task_timer_set(ospf_spf_timer, 0, ospf_spfInt);
}

/* The SPF timer job.
 */
void
ospf_spf_compute __PF2(tip, task_timer *, interval, time_t)
{
	ospf_area_t *area;

	ASEXTREE_BUG("ospf_spf_compute 1");

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
		("OSPF SPF: Full SPF Computation Stating"));

	rt_open(ospf_task);

	/* dijkstra algorithm on all areas
	 */
	OSPF_ALL_AREA(area) {
		spf_root(area);
	} OSPF_ALL_AREA_END(area);

	/* Inter area routes
	 */
	psf_all_lssum_route();

	/* External routes
	 */
	spf_all_lsext_route();

	ospf_ngbrmx_flush();
	rt_close(ospf_task, NULL, 0, "OSPF SPF Computation");

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
		("OSPF SPF: Full SPF Computation Completed"));

	ASEXTREE_BUG("ospf_spf_compute 2");
}


/* Flush all LSA from the neigbour queues.
 */
void
ospf_ngbrmx_flush __PF0(void)
{
	ospf_ngb_t *ngb;

	NGB_FLUSH_ALL(ngb) {
		/* Is there somethingto send? i.e. somethiong in the queue,
		 * but nothing already sent or and waiting for ACK.
		 */
		if (ngb->ngb_lsRmx.pt_used_nodes &&
				!ngb->ngb_RMXbuffer.pt_used_nodes)
			ospf_flush_rmx(ngb);

		NGB_FLUSH_DEL(ngb);
	}
}

/* It's time to do a refresh of all the LS originated from our router.
 */
void
ospf_refresh_schedule  __PF2(tip, task_timer *, interval, time_t)
{
	/* We create a background job with a low priority to
 	 * do the job of checking the whole DB.
	 * If the last job did not finish, well, that bad!
	 */
	if (ospf_refresh_job) {
		 trace_log_tf(ospf_trace_options, 0, LOG_ERR,
		("OSPF DB: a refresh job already exist. Started at %T",
			ospf_refresh_start));
		return;
	}

	/* Initialize the job and walk structure.
	 */
	ospf_refresh_area = ospf_area_list;
	ospf_refresh_walk = ospf_pt_walkinit(ospf_refresh_area);
	ospf_refresh_start = time_sec;
	ospf_refresh_job = task_job_create(ospf_task, TASK_JOB_PRIO_WORST,
		"OSPF Refresh", ospf_refresh, NULL);
}

/* Go through the LS database, and update whatever is needed. Kind of
 * sections #14, #14.1 and #12.4
 */
void
ospf_refresh __PF1(tjp, task_job *)
{
	int i;
	vertex_t *vtx;

	rt_open(ospf_task);

	/* Go through a schunk of the DB.
	 */
	for(i = 0; i < ospf_refreshChunkSize; i++) {

		/* If no more data in the DB, we try the next area. If there is
		 * none, then that's the end old friend!
		 */
		if (!ospf_refresh_walk->tw_node) {
			(void)ospf_pt_walkdel(ospf_refresh_walk);
			if (!(ospf_refresh_area=ospf_refresh_area->area_next)) {
				task_job_delete(ospf_refresh_job);
				ospf_refresh_job = NULL;
				break;
			}
        		ospf_refresh_walk =
				ospf_pt_walkinit(ospf_refresh_area);
			continue;
		}

		vtx = TREENODE_VTX(ospf_refresh_walk->tw_node);
		ospf_trace_vtx(vtx, "REFRESH");

		/* Is the age greater than MaxAge, and the vertex not scheduled
		 * for deletetion, or seq number wrap? Also the vtx must not
		 * be dead.
		 */
		if (VTX_AGE_MAX(vtx) && !BIT_TEST(vtx->vtx_flags,
				LF_REQUEST|LF_SEQWRAP|LF_DELETE|LF_HIDDEN)) {

			/* Well, get rid of it. XXXX
			 */
			BIT_SET(vtx->vtx_flags, LF_DELETE);
			ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
			ospf_pt_walk(ospf_refresh_walk);
			continue;
		}

		/* If the vertex is marked for deletion, and no reference
		 * count set, and not in the flooding queue, get rid of it!
		 */
		if (BIT_TEST(vtx->vtx_flags, LF_DELETE) &&
			!vtx->vtx_qref && !ospf_nbNgbNotReady) {
			ospf_pt_walk(ospf_refresh_walk);
			ospf_ls_delete(vtx, LSDEL_NORM);
			continue;
		}

		/* If this is a self originated LS and its age is more
		 * than ospf_lsRefreshTime, then we originate a new LSA.
		 * Carefull, the LS should not be a dead one, or be scheduled
		 * for deletion of seq wrapping.
		 */
		if (VTX_AGE(vtx) > ospf_lsRefreshTime &&
				!BIT_TEST(vtx->vtx_flags,
					LF_REQUEST|LF_DELETE|LF_HIDDEN) &&
				vtx->vtx_advrt == ospf_rtid) {

			trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
				("OSPF DB: Refreshing %s on area %A",
				    ls_vtx_name(vtx),
				    sockbuild_in(0, VTX_AREA(vtx)->area_id)));

			/* Increment seq number.
			 */
			if (vtx->vtx_seq == MAX_LS_SEQ) {
				VTX_AGE_SET(vtx, ospf_maxAge);
				BIT_SET(vtx->vtx_flags, LF_SEQWRAP);
                                ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
				ospf_pt_walk(ospf_refresh_walk);
				continue;
			}
			vtx->vtx_seq++;
			vtx->vtx_cksum = 0;
			VTX_AGE_SET(vtx, 0);
			ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
		}
		ospf_pt_walk(ospf_refresh_walk);
	}

	rt_close(ospf_task, NULL, 0, "OSPF Refresh");
}

/* Delete a vtx from the DB. Just get rid of it and any associated
 * memory structure. If flag is set to LSDEL_KILL, then the vertex
 * is deleted without trying to remove it from the tree (it's used when
 * destroying recursively the whole tree).
 * LSDEL_KILL is used when destroying the whole tree.
 * LSDEL_NORM is used when removing a too hold/dead vertex.
 */
void
ospf_ls_delete __PF2(vtx, vertex_t *, flag, flag_t)
{
	ospf_vtxLink_t **vlp;
	ospf_area_t *area = VTX_AREA(vtx);

	ospf_trace_vtx(vtx, "DELETE");
	OSPF_TREE_CHECK(&area->area_tree);

	/* what to do if the vertex is in the flooding queue? Arf!
	 */
	ASSERT(!BIT_TEST(vtx->vtx_flags, LF_FLUSH));

	/* We delete the TSI structure in the rt_head.
	 */
	if (vtx->vtx_rt)
		OSPF_TSI_CLEAR(vtx->vtx_rt->rt_head, ospf_task->task_rtbit,vtx);

	/* Delete the route if it not a seft-originated external LS.
	 */
	if (vtx->vtx_rt && vtx->vtx_rt != ospf_nullrt &&
		(vtx->vtx_advrt != ospf_rtid || vtx->vtx_type != LST_ASEX)) {
		ASEXTREE_BUG("ospf_ls_delete 2");
		OSPF_RT_VTX(vtx->vtx_rt) = NULL; /* Just in case... */
		rt_delete(vtx->vtx_rt);
		ASEXTREE_BUG("ospf_ls_delete 3");
		vtx->vtx_rt = NULL;
	}
	else if (vtx->vtx_rt) {
		vtx->vtx_rt = NULL;
	}
	else {
		ASSERT(!vtx->vtx_rt || !vtx->vtx_rt->rt_active ||
			!vtx->vtx_rt->rt_active->rt_gwp ||
			vtx->vtx_rt->rt_active->rt_gwp->gw_proto !=
				RTPROTO_OSPF);
		vtx->vtx_rt = NULL;
	}

	/* Delete the gateway
	 */
	/* gw_delete() XXX does not exist in gated!!!! */

	/* Any type-dependent parameters.
	 */
	switch(vtx->vtx_type) {
	case LST_RTR:
	case LST_NTW:
		/* Delete all links.
		 */
		for(vlp = &vtx->vtx_links; *vlp;)
			vlp = ls_vtxlink_delete(vlp, vtx);
		break;

	case LST_ASBR:
	case LST_ASIP:
	case LST_ASEX:
		break;
	default:
		ASSERT(0);
	}

	/* Delete the thing if no link exist. Else just mark it dead.
	 */
	if ((vtx->vtx_type == LST_RTR || vtx->vtx_type == LST_NTW) &&
			vtx->vtx_links) {
		trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
			("OSPF DB: ospf_ls_delete: %s marked dead",
			ls_vtx_name(vtx)));
		BIT_SET(vtx->vtx_flags, LF_HIDDEN);
	}
	else if (flag != LSDEL_KILL)
		db_delete(vtx);
	else {
		trace_tf(ospf_trace_options, TR_OSPF_DB, 0,
			("OSPF DB: ospf_ls_delete: Freeing %s",
				ls_vtx_name(vtx)));
		TREENODE_FREE((tree_node_t *)vtx);
	}

	OSPF_TREE_CHECK(&area->area_tree);
}

/* Update/create a self originated LSA.
 */
void
ospf_self_add __PF4(oldvtx, vertex_t **, lsa, ospf_lsa_t *, tree, db_tree_t *,
		rt, rt_entry *)
{
	vertex_t *vtx;
	lsseq_t seq;
	searchdata_t searchdata;

	/* If the oldvtx does not exist, look for one in the DB
	 */
	if (!*oldvtx) {
		vtx = ospf_global_find(tree, lsa->lsa_type, &lsa->lsa_key,
			&searchdata, tree);
		/* If there is none, create one with a sequence number of
		 * RES_LS_SEQ (forbiden one!).
		 */
		if (!vtx) {
			seq = RES_LS_SEQ;
			vtx = ospf_ls_create(tree, lsa, rt, &searchdata);
		}
		else {
			seq = vtx->vtx_seq;
			ospf_ls_update(tree, lsa, vtx);
		}
		*oldvtx = vtx;
	}
	else {
		vtx = *oldvtx;
		seq = vtx->vtx_seq;
		if (lsa)
			ospf_ls_update(tree, lsa, vtx);
	}

	/* Now we have the LSA in the DB. We need to get the sequence number
	 * to use.
	 */
	if (seq == MAX_LS_SEQ) {
		/* We need to do wrapping.
		 */
		BIT_SET(vtx->vtx_flags, LF_SEQWRAP);
		VTX_AGE_SET(vtx, ospf_maxAge);
	}
	else
		vtx->vtx_seq++;

	ASSERT(!vtx->vtx_cksum);
	ospf_flood_selforiginate(VTX_AREA(vtx), vtx);
}



