/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* Patricia tree management. Just follow the code in bgp_rt.c and pray!
 */
#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"


/* Add a new node to the tree. The node contain the key and data.
 * foundnode is where we failled when searching.
 */
void
tree_add __PF6(key, u_int8 *, keylen, int, root, tree_node_t **,
		newnode, tree_node_t *, foundnode, tree_node_t *, data, void_t)
{
	tree_node_t *node, *prevnode;
	bitindex_t bitindex, diffbit;

	/* If newnode is null then we need to allocate a new node and
	 * set its data pointer.
	 */
	if (!newnode)
		TREENODE_ALLOC(newnode, data, key, keylen);

	/* Case where the tree is empty!
	 */
	if (!foundnode) {
		*root = newnode;
		newnode->tn_left = newnode->tn_right = newnode;
		BITINDEX_RESET(newnode->tn_bitindex);
		return;
	}

	/* Find which bit node and foundnode differ
	 */
	bitdiff(&diffbit, key, keylen, (u_int8 *)foundnode->tn_key);
	ASSERT(diffbit.bi_bits || diffbit.bi_byte == 8);

	/* Go down the tree until we hit the place where to insert
	 */
	prevnode = NULL;
	node = *root;
	BITINDEX_INIT(bitindex);

	while(BITINDEX_STRICTLOWER(bitindex, node->tn_bitindex)) {
                BITINDEX_SET(bitindex, node->tn_bitindex);

		/* Stop just before the nodes differ.
		 */
		if (BITINDEX_LOWER(diffbit, bitindex))
			break;

		prevnode = node;
                if (TREEBIT_TEST(bitindex, key, keylen))
                        node = node->tn_right;
                else
                        node = node->tn_left;
        }

	ASSERT(BITINDEX_CMP(diffbit, bitindex));

	/* Insert the node
	 */
	if (TREEBIT_TEST(diffbit, key, keylen)) {
		newnode->tn_right = newnode;
		newnode->tn_left = node;
	}
	else {
		newnode->tn_right = node;
		newnode->tn_left = newnode;
	}
	BITINDEX_SET(newnode->tn_bitindex, diffbit);

	if (!prevnode)
		*root = newnode;
	else if (prevnode->tn_right == node)
		prevnode->tn_right = newnode;
	else
		prevnode->tn_left = newnode;

	ASSERT(newnode->tn_right != newnode || newnode->tn_left != newnode);
	ASSERT((*root)->tn_right != *root || (*root)->tn_left != *root);
}

/* find a node in the tree. If not found, nextnode is the point of failure.
 */
tree_node_t *
tree_find __PF4(key, u_int8 *, keylen, int, root, tree_node_t *,
		nextnode, tree_node_t **)
{
	tree_node_t *node;
	bitindex_t bitindex;

	/* Case where we have no node!
	 */
	if (!root) {
		*nextnode = NULL;
		return(NULL);
	}

	node = root;
	BITINDEX_INIT(bitindex);

	while(BITINDEX_STRICTLOWER(bitindex, node->tn_bitindex)) {

		BITINDEX_SET(bitindex, node->tn_bitindex);
		if (TREEBIT_TEST(bitindex, key, keylen))
			node = node->tn_right;
		else
			node = node->tn_left;	
	}

	*nextnode = node;
	return(TREEKEY_CMP(key, keylen, node->tn_key)? node: NULL);
}

/* Recursively destroy the whole tree. Do not free the data pointed by the
 * node. If flag is set to TK_REF the reference count for the vertex is
 * descreased by 1. If set to TK_REMOVE the vertex is removed.
 * TK_REF is used when removing a vertex from a ngb retransmit queue.
 * TK_REMOVE is used when destroying the whole tree (delete an area)
 * TK_NOREF is used when removing a vertex from a ngb DD or LSrequest queue
 */
void
tree_delete_rec __PF2(node, tree_node_t *, flag, flag_t)
{
	vertex_t *vtx;

	if (TREENODE_DOWN(node, node->tn_left))
		tree_delete_rec(node->tn_left, flag);

	if (TREENODE_DOWN(node, node->tn_right))
		tree_delete_rec(node->tn_right, flag);

	if (flag == TK_REF) {
		ASSERT(BIT_TEST(node->tn_flags, TNF_EXTERN));

		/* Do some reference cleaning...
		 */
		vtx = TREENODE_VTX(node);
		if (vtx && BIT_TEST(vtx->vtx_flags, LF_DELETE) &&
			!--vtx->vtx_qref && !ospf_nbNgbNotReady)
			ospf_ls_delete(vtx, LSDEL_NORM);
		TREENODE_FREE(node);
	}
	else if (flag == TK_REMOVE) {
		ASSERT(!BIT_TEST(node->tn_flags, TNF_EXTERN));
		vtx = TREENODE_VTX(node);
		ospf_ls_delete(vtx, LSDEL_KILL);
	}
	else
		TREENODE_FREE(node);
}

/* Delete a ospf_ptwalk_t structure.
 */
ospf_ptwalk_t *
ospf_pt_walkdel __PF1(ptw, ospf_ptwalk_t *)
{
        ospf_ptwalk_t **pp;

	/* If ptw->tw_area is null that means that the area has been destryed
	 * (during reconfig for instance!)
	 */
	if (ptw->tw_area)
        	for(pp = &ptw->tw_area->area_ptwalk; *pp; pp = &(*pp)->tw_next)
                	if (*pp == ptw) {
                        	*pp = ptw->tw_next;
                        	break;
                	}

        PT_WALK_FREE(ptw);
        return((ospf_ptwalk_t *)NULL);
}

/* Initialize a ospf_ptwalk_t structure.
 */
ospf_ptwalk_t *
ospf_pt_walkinit __PF1(area, ospf_area_t *)
{
        ospf_ptwalk_t *ptw;

        ptw = PT_WALK_ALLOC();
        ptw->tw_stkind = 0;
        ptw->tw_next = ospf_ptwalk;
	ptw->tw_area = area;
        ptw->tw_tree = &area->area_tree;

        /* Find the first non null tree.
         */
        ptw->tw_type = 0;
        while(ptw->tw_type < LST_MAXTYPE &&
			!area->area_tree.pt_root[ptw->tw_type])
                ptw->tw_type++;

        if (ptw->tw_type < LST_MAXTYPE) {
                ptw->tw_node = area->area_tree.pt_root[ptw->tw_type];
                ptw->tw_start = &area->area_tree.pt_root[ptw->tw_type];
        }
	ptw->tw_next = area->area_ptwalk;
        area->area_ptwalk = ptw;

        return(ptw);
}

/* Walk the tree. tw_node will point to the current node.
 */
void
ospf_pt_walk __PF1(ptw, ospf_ptwalk_t *)
{
	tree_node_t *node, *next;

	/* If the current node is null, get out
         */
        if (!ptw->tw_node)
                return;

        /* Try the next node until we have a non null vertex
         */
        do {
                /* First try to go down left.
                 */
                if (TREENODE_DOWN(ptw->tw_node, ptw->tw_node->tn_left)) {
			PTW_PUSH(ptw, ptw->tw_node);
			ptw->tw_node = ptw->tw_node->tn_left;
			if (TREENODE_VTX(ptw->tw_node))
				return;
			continue;
		}

		/* Then try to go right
		 */
		if (TREENODE_DOWN(ptw->tw_node, ptw->tw_node->tn_right)) {
			PTW_PUSH(ptw, ptw->tw_node);
			ptw->tw_node = ptw->tw_node->tn_right;
			if (TREENODE_VTX(ptw->tw_node))
                                return;
                        continue;
		}

		/* Then go back up, until we find a right branch (from where
                 * we are not coming).
                 */
                node = ptw->tw_node;
                while((next = PTW_POP(ptw))) {
                        if (next->tn_right != node &&
					TREENODE_DOWN(next, next->tn_right)) {
				PTW_PUSH(ptw, next);
				ptw->tw_node = next->tn_right;
				if (TREENODE_VTX(ptw->tw_node))
					return;
				break;	/* will continue inner loop */
			}
			node = next;
		}
		if (next)
			break;

		/* Nothing, that's the end of the tree. Find the next tree.
                 */
                while(++ptw->tw_type < LST_MAXTYPE &&
                        !ptw->tw_tree->pt_root[ptw->tw_type]);

                if (ptw->tw_type == LST_MAXTYPE) {
                        ptw->tw_node = NULL;
                        return;
                }

		ptw->tw_node = ptw->tw_tree->pt_root[ptw->tw_type];
                ptw->tw_start = &ptw->tw_tree->pt_root[ptw->tw_type];
		return;

	} while(1);

	return;
}

/* When deleting a node, change all ospf_ptwalk_t structure that may
 * point to that node.
 */
void
ospf_ptwalk_delnode __PF2(area, ospf_area_t *, node, tree_node_t *)
{
	ospf_ptwalk_t *ptw;

	for(ptw = area->area_ptwalk; ptw; ptw = ptw->tw_next)
		if (ptw->tw_node == node)
			ospf_pt_walk(ptw);
}

/* Delete a node from the patricia tree. This is a fast delete where the
 * nodes have external data and node are freed only if no downward link
 * exists. Return the delete data.
 */
void_t
tree_del_fast __PF4(root, tree_node_t **, key, u_int8 *, keylen, int,
		data, void_t)
{
	tree_node_t *node, *prev;
	ospf_ptwalk_t ptw;
	bitindex_t bitindex;

	/* First look for a node pointing to data.
	 * Esay empty case first.
	 */
	if (!*root)
		return(NULL);

	/* Go down the tree, store all nodes we go through. We have to
	 * catch the node when going *down*, not at the end of an upward link.
	 */
	ptw.tw_stkind = 0;
        node = *root;
        BITINDEX_INIT(bitindex);

	do {
		/* Do we have it?
		 */
		if (node->tn_data == data)
			break;

		/* Get the next one.
		 */
                BITINDEX_SET(bitindex, node->tn_bitindex);
		PTW_PUSH(&ptw, node);

                if (TREEBIT_TEST(bitindex, key, keylen))
                        node = node->tn_right;
                else
                        node = node->tn_left;

		/* Check if we missed it (i.e. the node is not here)
		 */
		if (!BITINDEX_STRICTLOWER(bitindex, node->tn_bitindex))
			node = NULL;
        } while(node);

	/* If no node, we are lost, nothing to delete.
	 */
	if (!node) {
		tree_node_t *toto = NULL;

		/* ASSERT(!(node = tree_find(key, keylen, *root, &toto)) ||
				!node->tn_data); */
		return(NULL);
	}
	node->tn_data = NULL;

	/* Delete each node if no downward link exist.
	 */
	while(node) {
		if (node->tn_data || TREENODE_DOWN(node, node->tn_left) ||
				TREENODE_DOWN(node, node->tn_right))
			break;

		/* Ok, we delete that one. Get the previous node first
		 * to update its links.
		 */
		prev = PTW_POP(&ptw);
		if (prev) {
			if (prev->tn_left == node)
				prev->tn_left = (node->tn_left != node)?
					node->tn_left:
						(node->tn_right != node)?
							node->tn_right:
							prev;
			else
				prev->tn_right = (node->tn_left != node)?
					node->tn_left:
						(node->tn_right != node)?
							node->tn_right:
							prev;
			if (prev->tn_left == prev && prev->tn_right == prev)
				BITINDEX_RESET(prev->tn_bitindex);
		}

		TREENODE_FREE(node);
		node = prev;
	}

	/* If no node left, reset the root.
	 */
	if (!node)
		*root = NULL;

	ASSERT(data);
	return(data);
}

static u_int8 bits2bit[256] = {
0x00,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x20,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x40,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x20,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x80,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x20,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x40,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x20,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
0x10,0x01,0x02,0x01,0x04,0x01,0x02,0x01,0x08,0x01,0x02,0x01,0x04,0x01,0x02,0x01,
};

/* Find out where key1 and k2 differ. The LSB bit is
 * number 1:1 (byte = 1, bit = 0x01), the MSB bit is number 4:0x80
 * (byte = 4, bit = 0x80). The BYTE_ORDER stuff is not really usefull,
 * One can use whatever bit ordering one wants when comparing keys.
 */
void
bitdiff __PF4(diffbit, bitindex_t *, key1, u_int8 *, keylen, int,
		key2, u_int8 *)
{
	register int i;

	ASSERT(keylen);

	diffbit->bi_byte = 1;
#if BYTE_ORDER == BIG_ENDIAN
	for(i = --keylen; i >= 0; i--)  {
		if ((diffbit->bi_bits = bits2bit[key1[i] ^ key2[i]]))
			break;
		diffbit->bi_byte++;
	}
#else
	for(i = 0; i < keylen; i++) {
		if ((diffbit->bi_bits = bits2bit[key1[i] ^ key2[i]]))
			break;
		diffbit->bi_byte++;
	}
#endif
}

/* Remove a node from a tree. The node is not freed. Uses the tough algorithm
 * found in bgp_rt.c. node point to the node we want to kill, so we do not
 * need to compare the keys!
 */
void
tree_node_del __PF4(root, tree_node_t **, key, u_int8 *, keylen, int,
                deadnode, tree_node_t *)
{
	tree_node_t *node, *tmp, *node_prev, *node_int_prev;
	bitindex_t bitindex;

	node = *root;
	ASSERT(node);

	/* Case of one node in the tree...
	 */
	if (BITINDEX_ISRESET(node->tn_bitindex)) {
		ASSERT(node == deadnode);
		*root = NULL;
		return;
	}
	ASSERT(node->tn_right != node || node->tn_left != node);

	/* find all the pointer we need to move out the dead node.
	 */
	BITINDEX_SET(bitindex, node->tn_bitindex);
	node_prev = node_int_prev = NULL;
	for(;;) {
		if (TREEBIT_TEST(bitindex, key, keylen))
			tmp = node->tn_right;
		else
			tmp = node->tn_left;

		if (tmp == deadnode) {
			/* Are we coming from a upward link?
			 */
			if (BITINDEX_LOWER(tmp->tn_bitindex, bitindex))
				break;

			/* We are coming down. Set node_int_prev.
			 */
			node_int_prev = node;
		}
		node_prev = node;
		node = tmp;
		BITINDEX_SET(bitindex, node->tn_bitindex);
	}

	/* Ok, we have all we need. 'node' point to the node before 'deadnode',
	 * such as the link 'node' -> 'deadnode' is an upward link.
	 * 'node_prev' is the node before 'node' (downward link).
	 * 'node_int_prev' is the node before 'deadnode',
	 * such as the link 'node_int_prev' -> 'deadnode' is an downward link.
	 * The goal now is to replace 'deadnode' by 'node', update the
	 * link of the following nodes: 'node_int_prev' to point to 'node';
	 * 'node' to point to whatever 'deadnode' was pointing; 'node_prev' to
	 * point to whatever 'node' was pointing before.
	 */

	/* Remember where was 'node' pointing to (beside deadnode).
	 */
	if (node->tn_right == deadnode)
		tmp = node->tn_left;
	else
		tmp = node->tn_right;

	/* update node_prev
	 */
	if (!node_prev)
		*root = tmp;
	else if (node_prev->tn_right == node)
		node_prev->tn_right = tmp;
	else
		node_prev->tn_left = tmp;

	/* Update 'node'
	 */
	if (node != deadnode) {
		BITINDEX_SET(node->tn_bitindex, deadnode->tn_bitindex);
		node->tn_left = deadnode->tn_left;
		node->tn_right = deadnode->tn_right;

		/* Update node_int_prev. It seems to me that if node == deadnode
		 * then node_int_prev == node_prev.
		 */
		if (!node_int_prev)
			*root = node;
		else if (node_int_prev->tn_right == deadnode)
			node_int_prev->tn_right = node;
		else
			node_int_prev->tn_left = node;
	}
	/* Catch the case where we are deleting the lowest node.
	 */
	else if (tmp == deadnode) {	/* node == deadnode */
		/* The lowest node is the only node with both branch pointing
		 * to itself. So tmp will be deadnode.
		 */
		ASSERT(node_int_prev);
		ASSERT(BITINDEX_ISRESET(deadnode->tn_bitindex));
		node_int_prev->tn_right =
			node_int_prev->tn_left = node_int_prev;
		BITINDEX_RESET(node_int_prev->tn_bitindex);
	}
}



