/*
 * DRET-IPv6
 *
 * Contacts:	
 *              INRIA      <Christophe.Diot@sophia.inria.fr>
 *              MASI       <Eric.Horlait@masi.ibp.fr>
 *              This software has been developped with the financial support
 *              of DRET (French Military Research Agency).
 *
 * Version:    $Id: ripng.c,v 1.30 1996/12/30 16:45:01 eh Exp $ 
 *	
 * Authors:	
 *	        Frank Zago
 *
 * Description:
 *
 *              ripng.c is derived from rip.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *              This notice must be copied with the distributed package.
 *
 */				

/*
 * Routing Information Protocol Next Generation (aka RIPng)
 * This version should comply with draft-ietf-rip-ripng-04 working doc.
 */

#define	INCLUDE_UDP

#include "include.h"
#include "inet6.h"
#include "targets.h"
#include "ripng.h"
#include "krt.h"

static u_short  ripng_port;
flag_t          ripng_flags = 0;	/* Options */
trace          *ripng_trace_options = {0};	/* Trace flags */
metric_t        ripng_default_metric = 0;	/* Default metric to use when
					 * propogating */
pref_t          ripng_preference = 0;	/* Preference for RIP routes */

static task_timer *ripng_timer_update;	/* To send updates */
static task_timer *ripng_timer_flash;	/* To send flash updates */
static task_timer *ripng_timer_age;	/* To age routes */

static task_job *ripng_target_list_job;	/* To rebuild target list after
					 * interface changes */
static target   ripng_targets = {&ripng_targets, &ripng_targets};	/* Target list */

static if_addr *ripng_multicast_ifap = (if_addr *) 0; 

static sockaddr_un *ripng_addr_mc=NULL;    /* MC address of ripng routers */
struct in6_addr INET6_RIPNG_ALLROUTERS={{{0xff,02,0,0,0,0,0,0,0,0,0,0,0,0,0,9}}};

struct {
	struct ripng_hdr header;
	struct ripng_rte rte;
} packet_request;		/* Packet request for other's routing table */

int ripng_n_trusted = 0;	  /* Number of trusted gateways */
int ripng_n_source = 0;		  /* Number of source gateways */
gw_entry *ripng_gw_list = 0;	  /* List of RIPng gateways */
adv_entry *ripng_import_list = 0; /* List of nets to import from RIPng */
adv_entry *ripng_export_list = 0; /* List of sources to export routes to RIPng */ 
adv_entry *ripng_int_policy = 0;  /* List of interface policy */

#if	defined(PROTO_SNMP) && defined(MIB_RIPNG)
u_int           ripng_global_changes;
u_int           ripng_global_responses;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */

static const bits ripng_flag_bits[] = {
	{RIPNGF_ON, "ON"},
	{RIPNGF_SOURCE, "Source"},
	{RIPNGF_FLASHDUE, "FlashDue"},
	{RIPNGF_NOFLASH, "NoFlash"},
	{RIPNGF_RECONFIG, "ReConfig"},
	{RIPNGF_TERMINATE, "Terminate"},
	{RIPNGF_MULTI, "Multicast"},
	{0}
};

static const bits ripng_cmd_bits[] = {
    { 0,		"Invalid" },
    { RIPNGCMD_REQUEST,	"Request" } ,
    { RIPNGCMD_RESPONSE,"Response" },
    { 0 }
};    

static const bits ripng_target_bits[] = {
	{RIPNGTF_POLL, "Poll"},
	{0}
};

static const bits ripng_if_bits[] = {
	{0}
};


static const flag_t ripng_trace_masks[RIPNGCMD_MAX] = {
	TR_ALL,			  /* 0 - Invalid */
	TR_RIPNG_DETAIL_REQUEST,  /* 1 - REQUEST */
	TR_RIPNG_DETAIL_RESPONSE, /* 2 - RESPONSE */
};

const bits ripng_trace_types[] = {
	{TR_DETAIL, "detail packets"},
	{TR_DETAIL_SEND, "detail send packets"},
	{TR_DETAIL_RECV, "detail recv packets"},
	{TR_PACKET, "packets"},
	{TR_PACKET_SEND, "send packets"},
	{TR_PACKET_RECV, "recv packets"},
	{TR_DETAIL_1, "detail request"},
	{TR_DETAIL_SEND_1, "detail send request"},
	{TR_DETAIL_RECV_1, "detail recv request"},
	{TR_PACKET_1, "request"},
	{TR_PACKET_SEND_1, "send request"},
	{TR_PACKET_RECV_1, "recv request"},
	{TR_DETAIL_2, "detail response"},
	{TR_DETAIL_SEND_2, "detail send response"},
	{TR_DETAIL_RECV_2, "detail recv response"},
	{TR_PACKET_2, "response"},
	{TR_PACKET_SEND_2, "send response"},
	{TR_PACKET_RECV_2, "recv response"},
	{TR_DETAIL_3, "detail other"},
	{TR_DETAIL_SEND_3, "detail send other"},
	{TR_DETAIL_RECV_3, "detail recv other"},
	{TR_PACKET_3, "other"},
	{TR_PACKET_SEND_3, "send other"},
	{TR_PACKET_RECV_3, "recv other"},
	{0, NULL}
};

/*
 * Trace RIP packets
 */
static void
ripng_trace 
__PF7(tp, task *,
      dir, int,
      ifap, if_addr *,
      who, sockaddr_un *,
      rpmsg, register struct ripng_packet *,
      size, register size_t,
      detail, int)
{
	register struct ripng_hdr *header = &rpmsg->header;
	trace *trp = tp->task_trace;
	const char *cmd = trace_state(ripng_cmd_bits, 
				      header->ripng_cmd < RIPNGCMD_MAX ? 
				      header->ripng_cmd : 0);

	switch (header->ripng_vers) {
		
	case RIPNG_VERSION_1:  /* Only one version for ipv6 */
		break;
		
	default:
		tracef("Unknown version %d\n", header->ripng_vers);
		return;
	}
	
	if (dir) {
		/* Trace packet transmission */
		tracef("RIPng %sSENT %s(%A) -> %A ",
		       dir > 0 ? "" : "*NOT* ",
		       ifap->ifa_link->ifl_name,
		       tp->task_addr,
		       who);
	} else {
		/* Trace packet reception */
		tracef("RIPng RECV %#A ->", who, task_recv_dstaddr);
	}
	tracef("vers %d, cmd %s, length %d",
	       header->ripng_vers,
	       cmd,
	       size);

	if (rpmsg->header.ripng_zero) {
		tracef(" header's must-be-zero field not zero ");
	}
	
	switch (header->ripng_cmd) {
	case RIPNGCMD_REQUEST:
	case RIPNGCMD_RESPONSE:
		trace_only_tf(trp,
			      0,
			      (NULL));
		if (detail) {
			register struct ripng_rte *n = rpmsg->rte;
	
			for (size -= (sizeof(struct ripng_hdr));
			     size >= sizeof(struct ripng_rte);
			     n++, size -= sizeof(struct ripng_rte)) {
				
				if (n->metric<=RIPNG_METRIC_UNREACHABLE) 
					tracef("\tDest %A/%02d Metric %02d",
					       sockbuild_in6(0, &(n->prefix), 128),
					       n->preflen,
					       n->metric);
				else
					tracef("\tGateway %A",
					       sockbuild_in6(0, &(n->prefix), 128));
				trace_only_tf(trp, TRC_NOSTAMP, (NULL));
			}
			
			tracef("RIPng %s end of packet", dir ? "SENT" : "RECV");
			if (size) 
				tracef(" %d residual bytes", size);
		
			trace_only_tf(trp, TRC_NOSTAMP, (NULL));
		}
		break;

	default:
		trace_only_tf(trp,
			      0,
			      (NULL));
		break;
	}
}

/* Remove ourselves from the MC group on this interface if necessary */
static void
ripng_mc_reset 
__PF2(tp, task *,
      tlp, target *)
{
	int *count = (int *) &tlp->target_ifap->ifa_ripng_mccount;

	if (BIT_TEST(tlp->target_flags, RIPNGTF_MCSET) && !--(*count)) {
		(void) task_set_option(tp,
				       TASKOPTION_GROUP_DROP_IPV6,
				       tlp->target_ifap,
				       ripng_addr_mc);
		BIT_RESET(tlp->target_flags, RIPNGTF_MCSET);
	}
}


/* Add ourselves to the MC group on this interface */
static int
ripng_mc_set 
__PF2(tp, task *,
      tlp, target *)
{
	flag_t ifps_state = tlp->target_ifap->ifa_ps[tp->task_rtproto].ips_state;
	int *count = (int *) &tlp->target_ifap->ifa_ripng_mccount;
	int rc = FALSE;
	
	if (BIT_TEST(tlp->target_flags, RIPNGTF_MCSET))
		return TRUE;	/* Indicate MC is already set */
		
	if (!BIT_TEST(ifps_state, RIPNG_IFPS_NOMC) && !(*count)++) {
		/* Try to join the MC group on this interface */
		
		if (!BIT_TEST(tlp->target_ifap->ifa_state, IFS_MULTICAST)) {
			trace_log_tp(tp,
				     0,
				     LOG_ERR,
				     ("ripng_mc_set: Multicast not possible on %s",
				      tlp->target_ifap->ifa_link->ifl_name));
			return FALSE;
		} 

		if ((task_set_option(tp,
				     TASKOPTION_GROUP_ADD_IPV6,
				     tlp->target_ifap,
				     ripng_addr_mc) < 0)
		    && (errno != EADDRNOTAVAIL)
		    && (errno != EADDRINUSE)) {
			/* Strange, this interface is not capable of MC */
			BIT_SET(ifps_state, RIPNG_IFPS_NOMC);
			
			trace_log_tp(tp,
				     0,
				     LOG_ERR,
				     ("ripng_mc_set: Multicast not available on %s",
				      tlp->target_ifap->ifa_link->ifl_name));
			
		} else {
			/* Indicate that we successfully enabled the MC
			 * address */
			BIT_SET(tlp->target_flags, RIPNGTF_MCSET);
			tlp->target_reset = ripng_mc_reset;
			rc = TRUE;
		}
	}
	return rc;
}

/* * */
static void
ripng_send 
__PF6(tp, task *,
      ifap, if_addr *,
      flags, flag_t,
      addr, sockaddr_un *,
      msg, struct ripng_packet *,
      size, size_t)
{
	u_short port = sock2port6(addr);
	int rc;

	if (!port)            /* FIXME: how can that happen ? */
		sock2port6(addr) = ripng_port;
	
	if (inet6_get_addr_type(&sock2in6(addr), 8) == INET6_ADDR_MCAST) {
		if_addr *ifap2;
		
		/* We choose the interface and the outgoing adress, 
		   so we must set MSG_DONTROUTE */
		BIT_SET(flags, MSG_DONTROUTE);

		/* Search for the first ifap for the same device with
		 * a Link-Local adress */
		IF_ADDR(ifap2) {
			if ((ifap->ifa_link == ifap2->ifa_link) &&
			    inet6_get_addr_type(&sock2in6(ifap2->ifa_addr), 10) == 
			    INET6_ADDR_LINKLOCAL)
				break;
		} IF_ADDR_END(ifap) ;

		if (!ifap2->ifa_addr) {
			trace_tp(tp, 
				 TR_DETAIL,
				 0,
				 ("ripng_send: can't find link local adress for interface %s.",
				  ifap->ifa_link->ifl_name));
			ifap2 = ifap;
		}
		
		if (ripng_multicast_ifap != ifap2) {
			(void) task_set_option(tp,
					       TASKOPTION_MULTI_IF_IPV6,
					       ripng_multicast_ifap = ifap2);
		}
		rc = task_send_packet_source(tp, 
					     ifap2, 
					     addr, 
					     (void_t) msg, 
					     size, 
					     flags);
	} else {
		/* The packet must be routed */
		BIT_RESET(flags, MSG_DONTROUTE);

		rc = task_send_packet(tp, (void_t) msg, size, flags, addr);
	}

	if (TRACE_PACKET_SEND_TP(tp,
				 msg->header.ripng_cmd,
				 RIPNGCMD_MAX,
				 ripng_trace_masks)) {
		ripng_trace(tp,
			    rc,
			    ifap,
			    addr,
			    msg,
			    size,
			    TRACE_DETAIL_SEND_TP(tp,
						 msg->header.ripng_cmd,
						 RIPNGCMD_MAX,
						 ripng_trace_masks));
	}
	sock2port6(addr) = port; /* FIXME: can be removed if port of addr
				    is set before this function */
}


/* Supply routing table to all targets on the list */
static int
ripng_supply 
__PF5(tlp, target *,
      dest, sockaddr_un *,
      flags, flag_t,
      send_flags, flag_t,
      flash_update, int)
{
	int count = 0;
	int changes = 0;
	size_t size, max_size;
	struct ripng_packet *packet = task_get_send_buffer(struct ripng_packet *);
	struct ripng_rte *fillp = &packet->rte[0];
	sockaddr_un *nexthop = inet6_addr_any; /* Default router is ourself */
	int is_us = TRUE;	/* nexthop is us */
	register td_entry *tdp;

	/* Initialize header */
	packet->header.ripng_cmd = RIPNGCMD_RESPONSE;
	packet->header.ripng_vers = RIPNG_VERSION_1;
	packet->header.ripng_zero = 0;                 /* Always for v1 */

	/* Calculate max size rounded out to size of route entry */
	max_size = RIPNG_MAXSIZE(tlp->target_ifap);
	max_size -= max_size % sizeof(struct ripng_rte);

	/* Open the routing table in case a holddown is over */
	rt_open(tlp->target_task);

	TD_LIST(tdp, &tlp->target_td) {
		int cleanup = 0;

		if (flash_update) {
			if (!BIT_TEST(tdp->td_flags, TDF_CHANGED))
				break; /* End of changes for this target */
		} else {
			/* Check for termination of holddown */
			if (!BIT_TEST(ripng_flags, RIPNGF_TERMINATE)
			    && BIT_TEST(tdp->td_flags, TDF_HOLDDOWN|TDF_POISON)
			    && !--tdp->td_metric) {
				/* Holddown is over - queue it to be released*/
				cleanup++;
			}
		}
		
		if (BIT_TEST(tdp->td_flags, TDF_CHANGED)) {
			/* Reset the changed field */
			BIT_RESET(tdp->td_flags, TDF_CHANGED);
		}

		size = (byte *) fillp - (byte *) packet;

		/* send the packet if 
		     - it is full 
		     - we need to put a rte with a next hop, with room for 
		       only one rte */
		if ((size >= max_size)
		    || ((!nexthop || (RT_ROUTER(tdp->td_rt) != nexthop)) &&
			(((byte *)(fillp+2)-(byte *)packet) >= max_size))) {
			/* Send packet */
			
			ripng_send(tlp->target_task,
				   tlp->target_ifap,
				   send_flags,
				   dest,
				   packet,
				   size);
			count++;
			
			/* Reset the fill pointer for the next time */
			fillp = &packet->rte[0];
			nexthop = inet6_addr_any; 
		}
	{		
		/* Put this entry in the packet */
		
		/* Draft v3, section 2.1.1
		 * if the next hop (ie router) isn't the same that the one for
		 * the last rte, emit a next hop rte.  */
		register struct ripng_rte *n = fillp;
		
#ifdef FIXME
		if (!nexthop ||
		    sock2plen6(RT_ROUTER(tdp->td_rt)) != sock2plen6(nexthop) ||
		    sockaddrcmp2_in6(RT_ROUTER(tdp->td_rt), nexthop, sock2plen6(nexthop))) {
			if_addr *ifap_tmp;

			/* Sanity check: if nexthop is us and 
			   RT_ROUTER(tdp->td_rt) is us too, 
			   with different adresses, do nothing */
			if (!((ifap_tmp=if_withlcladdr(RT_ROUTER(tdp->td_rt), TRUE)) && is_us)) {
				is_us = (ifap_tmp != NULL);	
				
				nexthop = RT_ROUTER(tdp->td_rt);
				
				/* Build a next hop RTE */
				if (nexthop)
					COPY_ADDR6(n->prefix,sock2in6(nexthop))
				else {
					COPY_ADDR6(n->prefix, sock2in6(*tlp->target_src));
					nexthop = *(tlp->target_src);
				}
				n->preflen = 0;
				n->rte_tag = 0;
				n->metric = RIPNG_METRIC_NEXTHOP;
				
				/* Update the fill pointer */
				fillp++;
				n++;
			}
		}
#endif		
		n->rte_tag = 0;
		COPY_ADDR6(n->prefix, sock2in6(tdp->td_rt->rt_dest));
		n->preflen = sock2plen6(tdp->td_rt->rt_dest);
		
		if (BIT_TEST(tdp->td_flags, TDF_HOLDDOWN|TDF_POISON)) {
			n->metric = RIPNG_METRIC_UNREACHABLE;
		} else {
			if (BIT_TEST(ripng_flags, RIPNGF_TERMINATE)) {
				n->metric = RIPNG_METRIC_SHUTDOWN;
			} else {
				n->metric = tdp->td_metric;
			}
		}
		/* Update the fill pointer */
		fillp++;
	}
		
		if (cleanup) {
			if (TRACE_TP(tlp->target_task, TR_POLICY)) {
				trace_only_tp(tlp->target_task,
					      0,
					      ("\tPrefix %A/%d %s ended",
					       tdp->td_rt->rt_dest,
					       sock2plen6(tdp->td_rt->rt_dest),
					       BIT_TEST(tdp->td_flags, TDF_POISON) ? "poison" : "holddown"));
			}
			changes++;
			TD_CLEANUP(tlp, tdp, TRUE);
		}
	} TD_LIST_END(tdp, &tlp->target_td);
	
	if (fillp > &packet->rte[0]) {
		size = (byte *) fillp - (byte *) packet;
		ripng_send(tlp->target_task,
			   tlp->target_ifap,
			   send_flags,
			   dest,
			   packet,
			   size);
		count++;
	}
	if (TRACE_TP(tlp->target_task, TR_POLICY) || changes) {
		trace_only_tp(tlp->target_task,
			      0,
			      (NULL));
	}
	rt_close(tlp->target_task, (gw_entry *) 0, changes, NULL);
	
	return count;
}


/*
 * Process a valid response
 */
static void
ripng_recv_response 
__PF6(tp, task *,
      gwp, gw_entry *,
      src_addr, sockaddr_un *,
      ifap, if_addr *,
      n, register struct ripng_rte *,
      limit, void_t)
{
	int routes = 0;
	rt_parms rtparms;
	struct ifa_ps *ips = &ifap->ifa_ps[tp->task_rtproto];
	struct ripng_rte *endp = (struct ripng_rte *) limit - 1;
	sockaddr_un *gw_dest = gwp->gw_addr;

	memset(&rtparms, 0, sizeof(rtparms));
	
	rtparms.rtp_n_gw = 1;
	rtparms.rtp_gwp = gwp;
	
	rt_open(tp);

	for (; n <= endp; n++) {
		rt_entry       *rt;
		
		routes++;
		
		/* It might be a Route Table Entry (RTE) or a Next Hop RTE */
		if (n->metric == RIPNG_METRIC_NEXTHOP) {
			/* Check validity of next hop rte */
			if ((n->rte_tag) || (n->preflen)) {
				trace_log_tp(tp,
					     0,
					     LOG_NOTICE,
					     ("ripng_recv_response: bad nexthop rte for net from %#A, with route tag %d and prefix length %u",
					      src_addr,
					      n->rte_tag,
				      n->preflen));
			} else {
				if (inet6_get_addr_type(&n->prefix, n->preflen)==INET6_ADDR_LINKLOCAL) {
					/* Locate or create a gateway 
					 * structure for the gateway 
					 * where the route points to */ 
					gw_dest = sockbuild_in6(0, &(n->prefix), 128);
					if (if_withlcladdr(gw_dest, FALSE))
					  /* Router is me. Discard */
					  continue;
				} else {
					/* Invalid or non-link local nexthop */
					gw_dest = gwp->gw_addr;
				}
			}
			n++;	/* next rte */
			if (n > endp) {    /* bogus packet */ 
			  trace_log_tp(tp,
				       0,
				       LOG_NOTICE,
				       ("ripng_recv_response: bogus packet (nexthop without route from %A",
					src_addr));
			  continue;

			}
		}

		/* check rte */
		if ((n->preflen>128) ||
		    (n->metric > RIPNG_METRIC_UNREACHABLE) ||
		    (!((inet6_get_addr_type(&n->prefix, n->preflen)) & INET6_ADDR_UNICAST )))
		{
			trace_log_tp(tp,
				     0,
				     LOG_NOTICE,
				     ("ripng_recv_response: bad rte for net %A/%u from %#A, with metric %u",
				      sockbuild_in6(0, &(n->prefix), n->preflen),
				      n->preflen,
				      src_addr,
				      n->metric));
			
#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
			gwp->gw_bad_routes++;
#endif /* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
			
			continue;
		}

		rtparms.rtp_dest = rtparms.rtp_dest_mask =
		  sockbuild_in6(0, &(n->prefix), n->preflen);

		if (if_myaddr(ifap, rtparms.rtp_dest, rtparms.rtp_dest)) {
			/* Ignore route to interface */

#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
			gwp->gw_bad_routes++;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
			continue;
		}
		
		clean_suffix_in6(rtparms.rtp_dest); /* Route */

		rtparms.rtp_preference = ripng_preference;
		rtparms.rtp_metric = n->metric;
		
		/* Now add hop count to metric */
		rtparms.rtp_metric += ips->ips_metric_in;
		rtparms.rtp_state = RTS_INTERIOR | RTS_GATEWAY;

		/* Routes infos */
		rtparms.rtp_router = gw_dest;
		rtparms.rtp_tag = ntohs(n->rte_tag);
		rtparms.rtp_ifap = ifap; /* Source link */

		rt = rt_locate(rtparms.rtp_state,
			       rtparms.rtp_dest,
			       rtparms.rtp_dest,
			       RTPROTO_BIT(tp->task_rtproto));
		if (!rt) {
			rt_head *rth;

			/*
			 * No route installed. See if we are announcing
			 * another route
			 */
			rt = rt_locate(RTS_NETROUTE,
				       rtparms.rtp_dest,
				       rtparms.rtp_dest,
				       RTPROTO_BIT_ANY);
			if (rt &&
			    (rth = rt->rt_head) &&
			    rth->rth_n_announce &&
			    (rt == rth->rth_active || rt == rth->rth_holddown)) {
				/* We are announcing this route */
				register target *tlp;
				
				/*
				 * RIPng won't announce an active route if one
				 * in holddown so check the holddown route 
				 * first */
				rt = rth->rth_holddown;
				if (!rt)
					rt = rth->rth_active;
				
				TARGET_LIST(tlp, &ripng_targets) {
					if (BIT_TEST(tlp->target_flags, TARGETF_SUPPLY) &&
					    rtbit_isset(rt, tlp->target_rtbit)) {
						/*
						 * We are sending to this
						 * target
						 */
						td_entry       *tdp;
						
						TD_TSI_GET(tlp, rth, tdp);
						
						if (BIT_TEST(tdp->td_flags, TDF_HOLDDOWN | TDF_POISON) ||
						    tdp->td_metric < rtparms.rtp_metric) {
							/*
							 * We are announcing
							 * this route from
							 * another protocol
							 */

							break;
						}
					}
				} TARGET_LIST_END(tlp, &ripng_targets);

				if (tlp) {
					/*
					 * Announced via another protocol,
					 * ignore this route
					 */
					continue;
				}
			}

			/* New route */
			if (rtparms.rtp_metric < RIPNG_METRIC_UNREACHABLE &&
			    import(rtparms.rtp_dest,
				   rtparms.rtp_dest,
				   ripng_import_list,
				   ips->ips_import,
				   rtparms.rtp_gwp->gw_import,
				   &rtparms.rtp_preference,
				   ifap,
				   (void_t) 0)) {
	
				/* Add new route */
				rt = rt_add(&rtparms);
#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
				ripng_global_changes++;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
			} else {
				BIT_SET(rtparms.rtp_gwp->gw_flags, GWF_IMPORT);
			}
		} else if (rt->rt_gwp == gwp) {
			/* same route */
			if (rtparms.rtp_metric > RIPNG_METRIC_UNREACHABLE) {
				rt_delete(rt);
			} else if (rtparms.rtp_metric != rt->rt_metric
				   || !sockaddrcmp_in6(rtparms.rtp_router, RT_ROUTER(rt))
				   || rtparms.rtp_tag != rt->rt_tag) {
				(void) rt_change(rt,
						 rtparms.rtp_metric,
						 rtparms.rtp_metric2,
						 rtparms.rtp_tag,
						 rt->rt_preference,
						 rt->rt_preference2,
						 1, &rtparms.rtp_router);
				rt_refresh(rt);
#if	defined(PROTO_SNMP) && defined(MIB_RIPNG)
				ripng_global_changes++;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
			} else {
				rt_refresh(rt);
			}
		} else if ((rtparms.rtp_metric < rt->rt_metric
			    || (rt_age(rt) > (RIPNG_T_EXPIRE / 2)
				&& rt->rt_metric == rtparms.rtp_metric))
			   && import(rtparms.rtp_dest,
				     rtparms.rtp_dest,
				     ripng_import_list,
				     ips->ips_import,
				     rtparms.rtp_gwp->gw_import,
				     &rtparms.rtp_preference,
				     ifap,
				     (void_t) 0)) {
			/* Better metric or same metric and old route has */
			/* not been refreshed. */

			(void) rt_delete(rt);
			rt = rt_add(&rtparms);
#if	defined(PROTO_SNMP) && defined(MIB_RIPNG)
			ripng_global_changes++;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
		}
	}			/* for each route */

	rt_close(tp, rtparms.rtp_gwp, routes, NULL);
}


/*
 * Process a valid request
 */
static void
ripng_recv_request 
__PF7(tp, task *,
      src_addr, sockaddr_un *,
      ifap, if_addr *,
      tlp, target *,
      packet, struct ripng_packet *,
      n, register struct ripng_rte *,
      limit, void_t)
{
	struct ifa_ps *ips = &ifap->ifa_ps[tp->task_rtproto];
	struct ripng_rte *last = (struct ripng_rte *) limit - 1;
	
	packet->header.ripng_cmd = RIPNGCMD_RESPONSE; /* use the old message */
	
	/* draft v3 section 2.4.1: if no entries, no response */
	if (n == limit)
		return;
	
	/* draft v3 section 2.4.1: Check if routing table is requested */
	if (n == last
	    && IS_ANYADDR6(n->prefix)
	    && n->metric == RIPNG_METRIC_UNREACHABLE
	    && n->preflen == 0)
		(void) ripng_supply(tlp,
				    src_addr,
				    tlp->target_flags,
				    (flag_t) 0,
				    FALSE);
	else {
		/* Specific request */
		while (n <= last) {
			/* Process each request */
			
			n->metric = RIPNG_METRIC_UNREACHABLE;
			
			if (n->preflen<=128)
			{
				sockaddr_un *addrdest=sockbuild_in6(0, &(n->prefix), n->preflen);
				rt_entry *rt;
				clean_suffix_in6(addrdest);

				rt = rt_locate(RTS_INTERIOR,
							 addrdest,
							 addrdest,
							 RTPROTO_BIT_ANY);
				
				if (rt && !BIT_TEST(rt->rt_state, RTS_HIDDEN)) {
					n->metric = rt->rt_metric
						+ ips->ips_metric_out;
					if (n->metric > RIPNG_METRIC_UNREACHABLE) {
						n->metric = RIPNG_METRIC_UNREACHABLE;
					}
				}
			}
			n++;	/* process next entry */
		}

		ripng_send(tp,
			   (if_addr *) 0,
			   0,
			   src_addr,
			   packet,
			   (size_t) ((caddr_t) n - (caddr_t) packet));
	}

#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
	ripng_global_responses++;
#endif				/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
}


/*
 * Check out a newly received RIPng packet.
 */
static void
ripng_recv 
__PF1(tp, task *)
{
	size_t          size;
	int             n_packets = TASK_PACKET_LIMIT;
	
#define	REJECT(p, m)	{ reject_msg = (m); pri = (p); goto Reject; }

	/* Process each packets received */
	while (n_packets-- && !task_receive_packet(tp, &size)) {
		register if_addr *ifap = (if_addr *) 0;
		gw_entry *gwp = (gw_entry *) 0;
		register int OK = TRUE;
		struct ripng_packet *inpacket = task_get_recv_buffer(struct ripng_packet *);
		int pri = 0;
		const char *reject_msg = (char *) 0;
		void_t limit = (void_t) ((byte *) inpacket + size);

		/* Locate or create a gateway structure for this gateway 
		 * where this packet come from */ 
		gwp = gw_locate(&ripng_gw_list,
				tp->task_rtproto,
				tp,
				(as_t) 0,
				(as_t) 0,
				task_recv_srcaddr,
				GWF_NEEDHOLD);

		/*
		 * If we have a list of trusted gateways, verify that this
		 * gateway is trusted
		 */
		if (ripng_n_trusted && !BIT_TEST(gwp->gw_flags, GWF_TRUSTED)) {
			OK = FALSE;
		}

		if (TRACE_PACKET_RECV_TP(tp,
					 inpacket->header.ripng_cmd,
					 RIPNGCMD_MAX,
					 ripng_trace_masks)) {
			ripng_trace(tp,
				    0,
				    ifap,
				    gwp->gw_addr,
				    inpacket,
				    size,
				    TRACE_DETAIL_RECV_TP(tp,
							 inpacket->header.ripng_cmd,
							 RIPNGCMD_MAX,
							 ripng_trace_masks));
		}

		/* Verify that header is correct */
		if (inpacket->header.ripng_vers != RIPNG_VERSION_1) {
			pri = BIT_TEST(gwp->gw_flags, GWF_FORMAT) ? 0 : LOG_WARNING;
			BIT_SET(gwp->gw_flags, GWF_FORMAT);
			REJECT(0, "unknown RIPng version");
		}

		if (inpacket->header.ripng_zero) {

			pri = BIT_TEST(gwp->gw_flags, GWF_FORMAT) ? 0 : LOG_WARNING;
			BIT_SET(gwp->gw_flags, GWF_FORMAT);
			REJECT(pri, "reserved field not zero");
		}

		/* Subtract header length from size, and verify that size
		 * of all rte is correct */
		size -= sizeof(struct ripng_hdr);
		if (size % sizeof(struct ripng_rte))
			REJECT(0, "not a multiple of network entry size");

		/* Process packet */
		switch (inpacket->header.ripng_cmd) {
		case RIPNGCMD_REQUEST: 
		{	       
			target *tlp = (target *) 0, *tlp2;
			
			/* draft v3 section */
			
			/* This is a request from a router, make sure 
			 * we should and are allowed to reply */

			if (!OK)
				REJECT(0, "not on trustedgateways list");
			
			if (if_withlcladdr(gwp->gw_addr, FALSE))
				/*
				 * Ignore my own requests
				 */
				continue;

			ifap = if_getfromindex(received_if);
			if (!ifap)
				REJECT(0, "can not find interface for source address");

			if (BIT_TEST(ifap->ifa_ps[tp->task_rtproto].ips_state, IFPS_NOIN | IFPS_NOOUT))
				REJECT(0, "interface marked for no RIPng in/out");

			/* Find the Target for this host */
			TARGET_LIST(tlp2, &ripng_targets) {
				/*
				 * Look for target for this interface
				 */
				
				if (BIT_TEST(tlp2->target_flags, TARGETF_SUPPLY)) {
					if (tlp2->target_gwp == gwp) {
						/*
						 * Found a target for
						 * this gateway!
						 */
						
						tlp = tlp2;
						break;
					} else if (tlp2->target_ifap == ifap) {
						/*
						 * Found a target for
						 * this interface
						 */
						/*
						 * remember it, but keep
						 * looking in case there is
						 * one for the gateway
						 */
						tlp = tlp2;
					}
				}
			} TARGET_LIST_END(tlp2, &ripng_targets);
			
			/* Verify that we can respond */
			if (!tlp)
				REJECT(0, "not supplying RIPng");
			
			BIT_SET(gwp->gw_flags, GWF_QUERY | GWF_ACCEPT);
			
			ripng_recv_request(tp,
					   task_recv_srcaddr,
					   ifap,
					   tlp,
					   inpacket,
					   &inpacket->rte[0],
					   limit);
			
			continue;
		}

		case RIPNGCMD_RESPONSE:
			/* Simple checkings */
			if (!OK)
				REJECT(0, "not on trustedgateways list");

			if (sock2port6(task_recv_srcaddr) != ripng_port)
				REJECT(0, "not from the RIPng port");
			

			ifap = if_getfromindex(received_if);
			if (!ifap)
				REJECT(0, "can not find interface for source address");

			/* FIXME : Security hole if coming from a tunnel and
			   not authenticated */
			if (inet6_get_addr_type(&sock2in6(task_recv_srcaddr), 10) != INET6_ADDR_LINKLOCAL)
			  REJECT(0, "not from Link-Local adress");

			/*
			 * update interface timer on interface that
			 * packet came in on.
			 */
			if (!BIT_TEST(ifap->ifa_state, IFS_TUNNEL)) {
			  if_rtupdate(ifap);
			}
			  
			if (BIT_TEST(ifap->ifa_ps[tp->task_rtproto].ips_state, IFPS_NOIN))
				REJECT(0, "interface marked for no RIPng in");

			BIT_SET(gwp->gw_flags, GWF_ACCEPT);
			gwp->gw_time = time_sec;
			
			ripng_recv_response(tp,
					    gwp,
					    task_recv_srcaddr,
					    ifap,
					    &inpacket->rte[0],
					    limit);
			break;

		default:
			pri = BIT_TEST(gwp->gw_flags, GWF_FORMAT) ? 0 : LOG_WARNING;
			BIT_SET(gwp->gw_flags, GWF_FORMAT);
			REJECT(0, "invalid or not implemented command");
		}
		continue;	/* exit from function */

Reject:
#if	defined(PROTO_SNMP) && defined(MIB_RIPNG)
		if (gwp) {
			gwp->gw_bad_packets++;
		}
#endif	/* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
		tracef("ripng_recv: ignoring RIPng incoming packet ");
		if (inpacket->header.ripng_cmd < RIPNGCMD_MAX) {
			tracef("%s",
			       trace_state(ripng_cmd_bits, inpacket->header.ripng_cmd));
		} else {
			tracef("#%d", inpacket->header.ripng_cmd);
		}
		trace_log_tp(tp,
			     0,
			     pri,
			     (" packet from %#A to %#A - %s",
			      task_recv_srcaddr,
			      task_recv_dstaddr,
			      reject_msg));
		trace_only_tp(tp,
			      0,
			      (NULL));
		if (gwp) {
			BIT_SET(gwp->gw_flags, GWF_REJECT);
		}
	}
}

/* * */
/*
 * Deal with interface policy
 */
static void
ripng_control_reset 
__PF2(tp, task *,
      ifap, if_addr *)
{
	struct ifa_ps  *ips = &ifap->ifa_ps[tp->task_rtproto];

	BIT_RESET(ips->ips_state, IFPS_RESET);
	ips->ips_metric_in = ifap->ifa_metric + RIPNG_HOP;
	ips->ips_metric_out = (metric_t) 0;
}


static void
ripng_control_set 
__PF2(tp, task *,
      ifap, if_addr *)
{
	struct ifa_ps  *ips = &ifap->ifa_ps[tp->task_rtproto];
	config_entry  **list = config_resolv_ifa(ripng_int_policy,
						 ifap,
						 RIPNG_CONFIG_MAX);

	/* Reset */
	ripng_control_reset(tp, ifap);

	/* Set defaults */
	switch (BIT_TEST(ifap->ifa_state, IFS_POINTOPOINT | IFS_LOOPBACK )) {
	case IFS_LOOPBACK:
		/* By default we do not send or listen on the loopback */
		/* interface.  This does not prevent us from processing */
		/* diagnostic REQUEST packets. */
		BIT_SET(ips->ips_state, IFPS_NOIN | IFPS_NOOUT);
		break;

	case IFS_POINTOPOINT:
		/* By default we do not send RIP out a P2P interface. */
		BIT_SET(ips->ips_state, IFPS_NOOUT);
		break;

	default:
		/* On multicast and NBMA interfaces we default to sending 
		 * and receiving responses. */
		break;
	}

	/* Process configuration info */
	if (list) {
		int type = RIPNG_CONFIG_MAX;
		config_entry *cp;

		/* Fill in the parameters */
		while (--type) {
			if ((cp = list[type])) {
				switch (type) {
				case RIPNG_CONFIG_IN:
					if (GA2S(cp->config_data)) {
						BIT_RESET(ips->ips_state, IFPS_NOIN);
					} else {
						BIT_SET(ips->ips_state, IFPS_NOIN);
					}
					break;

				case RIPNG_CONFIG_OUT:
					if (GA2S(cp->config_data)) {
						BIT_RESET(ips->ips_state, IFPS_NOOUT);
					} else {
						BIT_SET(ips->ips_state, IFPS_NOOUT);
					}
					break;

				case RIPNG_CONFIG_METRICIN:
					BIT_SET(ips->ips_state, IFPS_METRICIN);
					ips->ips_metric_in = (metric_t) GA2S(cp->config_data);
					break;

				case RIPNG_CONFIG_METRICOUT:
					BIT_SET(ips->ips_state, IFPS_METRICOUT);
					ips->ips_metric_out = (metric_t) GA2S(cp->config_data);
					break;

				case RIPNG_CONFIG_FLAG:
					BIT_SET(ips->ips_state, (flag_t) GA2S(cp->config_data));
					break;
				}
			}
		}
		
		config_resolv_free(list, RIPNG_CONFIG_MAX);
	}
}


void
ripng_config_free 
__PF1(cp, config_entry *)
{
	switch (cp->config_type) {
	default:
		/* Not allocated */
		break;
	}
}


/*
 * Cleanup before re-init
 */
/* ARGSUSED */
static void
ripng_cleanup 
__PF1(tp, task *)
{
	adv_cleanup(RTPROTO_RIPNG,
		    &ripng_n_trusted,
		    &ripng_n_source,
		    ripng_gw_list,
		    &ripng_int_policy,
		    &ripng_import_list,
		    &ripng_export_list);

	if (ripng_multicast_ifap) {
		IFA_FREE(ripng_multicast_ifap);
		ripng_multicast_ifap = (if_addr *) 0;
	}

	if (tp) {
		trace_freeup(tp->task_trace);
	}
	trace_freeup(ripng_trace_options);
}


static void
ripng_exit 
__PF1(tp, task *)
{
	gw_entry       *gwp;
	if_addr        *ifap;

	/* Release the target list, bit assignments, and buffers */
	target_free_list(tp, &ripng_targets);

	/* Reset the policy */
	IF_ADDR(ifap) {
		ripng_control_reset(tp, ifap);
	} IF_ADDR_END(ifap);

	rt_open(tp);

	GW_LIST(ripng_gw_list, gwp) {
		rt_entry       *rt;
		
		RTQ_LIST(&gwp->gw_rtq, rt) {
			rt_delete(rt);
		} RTQ_LIST_END(&gwp->gw_rtq, rt);
	} GW_LIST(ripng_gw_list, gwp);

	rt_close(tp, (gw_entry *) 0, 0, NULL);

	ripng_cleanup(tp);

	if (ripng_addr_mc) {
		/* Initialize address constant */

		sockfree(ripng_addr_mc);
		ripng_addr_mc = (sockaddr_un *) 0;
	}
	
	task_delete(tp);
	ripng_timer_update = ripng_timer_flash = ripng_timer_age = (task_timer *) 0;
}

/*
 * Evaluate policy for changed routes.
 */
static int
ripng_policy __PF3(tp, task *, 
		   tlp, target *, 
		   change_list, rt_list *)
{
	if_addr *ifap = tlp->target_ifap;
	int changes = 0;
	int logged = 0;
	rt_head *rth;
	struct ifa_ps  *ips = &ifap->ifa_ps[tp->task_rtproto];

	RT_LIST(rth, change_list, rt_head) {
		register rt_entry *new_rt = rth->rth_active;
		adv_results     result;
		td_entry       *tdp;
		int             exportable = FALSE;
		int             changed = 0;
		int             move_bit = 0;
		int             holddown = 0;
		int             poison = 0;
		int             set_metric = 0;
		
		trace_tp(tp, TR_POLICY, 0,
			 ("ripng_policy: evaluating %A/%d (%s)",
			  rth->rth_dest, sock2plen6(rth->rth_dest),
			  (new_rt)? 
			    RT_IFAP(new_rt)->ifa_link->ifl_name : 
			    "???"));

		TD_TSI_GET(tlp, rth, tdp);
			 
		/* Can we announce the new route (if there is one)? */
		if (new_rt) {
			if (socktype(new_rt->rt_dest) != AF_INET6) {
				goto no_export;
			}
			if (BIT_TEST(new_rt->rt_state, RTS_NOADVISE|RTS_GROUP)){
				/* Absolutely not */
				
				goto no_export;
			}
					
			if (RT_IFAP(new_rt) == ifap &&
			    new_rt->rt_gwp->gw_proto == RTPROTO_DIRECT) {
				/*
				 * Do not send interface routes back to the
				 * same interface
				 */
				goto no_export;
			}
			
			/* Split horizon : do not send on same link */
			if ((new_rt->rt_gwp->gw_proto == tp->task_rtproto) 
			    && (ifap->ifa_link == RT_IFAP(new_rt)->ifa_link)) {
			      goto no_export;
			}
			
			if (RT_IFAP(new_rt)
			    && !BIT_TEST(RT_IFAP(new_rt)->ifa_state, IFS_UP)) {
				/* The interface is down */
				goto no_export;
			}
				
			/* Assign default metric */
			if (new_rt->rt_gwp->gw_proto == RTPROTO_AGGREGATE) {
				/* Originate aggregates with a metric of one */
				
				result.res_metric = RIPNG_HOP;
			} else if (!RT_IFAP(new_rt)
				   || BIT_TEST(RT_IFAP(new_rt)->ifa_state,
					       IFS_LOOPBACK)) {
				/*
				 * Routes via the loopback interface must
				 * have an explicit metric
				 */
				result.res_metric = RIPNG_METRIC_UNREACHABLE;
			} else if (new_rt->rt_gwp->gw_proto == RTPROTO_DIRECT) {
				/* Interface routes */
				
				if (BIT_TEST(RT_IFAP(new_rt)->ifa_state,
					     IFS_POINTOPOINT)) {
					/* Add a hop for the P2P link
					 */
					result.res_metric = RIPNG_HOP * 2;
				} else {
					/* Default to one hop
					 */
					result.res_metric = RIPNG_HOP;
				}
			} else {
				/* Use configured default metric */
				
				result.res_metric = ripng_default_metric;
			}

			
			if (!export(new_rt, tp->task_rtproto, ripng_export_list,
				    ips->ips_export,
				    tlp->target_gwp ?
				    tlp->target_gwp->gw_export :
				    (adv_entry *) 0,
				    &result)) {
				/* Policy prohibits announcement
				 */
				goto no_export;
			} else {
				/* Add the interface metric */
				
				result.res_metric += ips->ips_metric_out;
			}
			
			if (result.res_metric < RIPNG_METRIC_UNREACHABLE) {
				exportable = TRUE;
			}
		no_export:	;

                }
		/* Now that we have determined the exportablility of the new
		 * route we decide what changed need to be made.  The
		 * complexity is required to surpress routing loops both
		 * within RIP and between RIP and other protocols.
		 * There are two types of holddowns used, the first one is
		 * called HOLDDOWN and is used when a route goes away or is
		 * overridden by a route that is not suspected to be an echo
		 * of a route we are announcing.  The second is called POISON
		 * and is used when a route is overridden by a route
		 * suspected to be an echo of a route we are announcing.
		 */
		trace_tp(tp, TR_POLICY, 0,
			 ("ripng_policy: tdp %A exportable %d poison_reverse %d new_rt %A",
			(tdp)? tdp->td_rt->rt_dest:
				sockbuild_str("NULL"),
			exportable, poison,
			(new_rt)? new_rt->rt_dest:
				sockbuild_str("NULL")));

		if (!tdp) {
			/* New route */

			if (exportable) {
				/* and it is exportable */
				
				/* Allocate new entry and fill it in */
				TD_ALLOC(tdp);
				TD_TSI_SET(tlp, rth, tdp);
				tdp->td_rt = new_rt;
				rtbit_set(tdp->td_rt, tlp->target_rtbit);
				TD_ENQUE(tlp, tdp);
				set_metric++;
				changes++;
			}
		} else if (!new_rt) {
			/* No new route, just an old one */
			
			if (!BIT_TEST(tdp->td_flags, TDF_POISON | TDF_HOLDDOWN)) {
				if (BIT_TEST(tdp->td_rt->rt_state, RTS_DELETE | RTS_HIDDEN)) {
					/* Put into holddown */
					
					holddown++;
					changed++;
				} else {
					/* Poison the old route */
					
					poison++;
					changed++;
				}
			}
		} else if (new_rt == tdp->td_rt) {
			/*
			 * Something has changed with the route we are
			 * announcing
			 */

			if (BIT_TEST(tdp->td_flags, TDF_POISON | TDF_HOLDDOWN)) {
				if (exportable) {
					set_metric++;
					changed++;
				}
			} else {
				if (!exportable) {
					poison++;
					changed++;
				} else if (tdp->td_metric != result.res_metric) {
					set_metric++;
					changed++;
				}
			}
		} else if (!BIT_TEST(new_rt->rt_state, RTS_PENDING)) {
			/* The new route is not from a holddown protocol */
			
			if (exportable
			    && (BIT_TEST(new_rt->rt_gwp->gw_flags, GWF_NOHOLD)
				|| BIT_TEST(tdp->td_rt->rt_gwp->gw_flags, GWF_NOHOLD))) {
				/*
				 * The new route is a ``static'' route or the
				 * old route is a ``static'' route, 
				 *override immediately 
				 */
				
				set_metric++;
				changed++;
				move_bit++;
			} else if (BIT_TEST(tdp->td_flags, TDF_HOLDDOWN)
				   && !BIT_TEST(tdp->td_rt->rt_gwp->gw_flags, GWF_NEEDHOLD)
				   && !BIT_TEST(new_rt->rt_gwp->gw_flags, GWF_NEEDHOLD)) {
				/* No changes necessary */
				
				changed = move_bit = 0;
			} else if (!BIT_TEST(tdp->td_flags, TDF_POISON | TDF_HOLDDOWN)) {
				/* Previously active, has been deleted */
				
				holddown++;
				changed++;
			} else {
				if (exportable) {
					set_metric++;
					changed++;
				} else {
					if (!BIT_TEST(tdp->td_flags, TDF_POISON)) {
						poison++;
						changed++;
					}
				}
				move_bit++;
			}
		} else {
			/* New route is just better */

			if (BIT_TEST(tdp->td_rt->rt_state, RTS_DELETE | RTS_HIDDEN)) {
				if (!BIT_TEST(tdp->td_flags, TDF_HOLDDOWN)) {
					holddown++;
					changed++;
				}
			} else {
				if (!BIT_TEST(tdp->td_flags, TDF_POISON)) {
					poison++;
					changed++;
				}
				move_bit++;
			}
		}

		if ((changed || set_metric)
		    && TRACE_TP(tp, TR_POLICY)) {
			if (!logged) {
				logged++;
				trace_only_tp(tp,
					      TRC_NL_BEFORE,
					      ("ripng_policy: Policy for target %s (intf index %d) (%A%s)",
					       tlp->target_ifap->ifa_info.ifi_link->ifl_name,
					       tlp->target_ifap->ifa_info.ifi_link->ifl_index,
					       *tlp->target_dst,
					       "mc "));
			}
			tracef("\t%A/%d ",
			       rth->rth_dest,
			       sock2plen6(rth->rth_dest_mask));
		}
		if (set_metric) {
			target_set_metric(tdp, result.res_metric);
			trace_tp(tp,
				 TR_POLICY,
				 0,
				 ("metric %u",
				  tdp->td_metric));
		} else if (holddown) {
			target_set_holddown(tdp, RIPNG_HOLDCOUNT);
			trace_tp(tp,
				 TR_POLICY,
				 0,
				 ("starting holddown"));
		} else if (poison) {
			target_set_poison(tdp, RIPNG_HOLDCOUNT);
			trace_tp(tp,
				 TR_POLICY,
				 0,
				 ("starting poison"));
		} else if (TRACE_TP(tp, TR_POLICY)) {
			trace_clear();
		}
		if (changed) {
			/*
			 * Changed entries need to be at the head of the
			 * queue to
			 */
			/* make triggered updates quick */
			
			TD_DEQUE(tdp);
			TD_ENQUE(tlp, tdp);
			changes++;
		}
		if (move_bit) {
			/* Move lock to new route */
			
			rtbit_set(new_rt, tlp->target_rtbit);
			(void) rtbit_reset(tdp->td_rt, tlp->target_rtbit);
			tdp->td_rt = new_rt;
		}
	} RT_LIST_END(rth, change_list, rt_head);
	
	if (logged) {
		trace_tp(tp,
			 TR_POLICY,
			 0,
			 (NULL));
	}
	return changes;
}


/*
 * send RIPng packets
 */
/* ARGSUSED */
static void
ripng_job 
__PF2(tip, task_timer *,
      interval, time_t)
{
	target         *tlp;
	
	TARGET_LIST(tlp, &ripng_targets) {
		if (BIT_TEST(tlp->target_flags, TARGETF_SUPPLY)) {
			(void) ripng_supply(tlp,
					    ripng_addr_mc,
					    tlp->target_flags,
					    MSG_DONTROUTE,
					    FALSE);
		}
	} TARGET_LIST_END(tlp, &ripng_targets);

	/*
	 * Indicate that flash updates are possible as soon as the timer
	 * fires
	 */
	BIT_RESET(ripng_flags, RIPNGF_NOFLASH | RIPNGF_FLASHDUE);
	task_timer_set(ripng_timer_flash, RIPNG_T_FLASH, (time_t) 0);
}


/*
 * send a flash update packet
 */
/* ARGSUSED */
static void
ripng_do_flash 
__PF2(tip, task_timer *,
      interval, time_t)
{
	static int term_updates = 4;
	task *tp = tip->task_timer_task;
	
	if (BIT_TEST(ripng_flags, RIPNGF_FLASHDUE | RIPNGF_TERMINATE)) {
		int count = 0;
		target *tlp;

		trace_tp(tp,
			 TR_TASK,
			 0,
			 ("ripng_do_flash: Doing flash update for RIPng"));

		TARGET_LIST(tlp, &ripng_targets) {
			if (BIT_TEST(tlp->target_flags, TARGETF_SUPPLY)) {
				count += ripng_supply(tlp,
						      ripng_addr_mc,
						      tlp->target_flags,
						      MSG_DONTROUTE,
						      !BIT_TEST(ripng_flags, RIPNGF_TERMINATE));
			}
		} TARGET_LIST_END(tlp, &ripng_targets);
		
		trace_tp(tp,
			 TR_TASK,
			 0,
			 ("ripng_do_flash: Flash update done"));

		if (BIT_TEST(ripng_flags, RIPNGF_TERMINATE) && (!count || !--term_updates)) {
			/*
			 * Sent the requisite number of updates or nothing to
			 * send
			 */
			ripng_exit(tp);
			return;
		}
		/* Indicate no flash update is due */
		BIT_RESET(ripng_flags, RIPNGF_FLASHDUE);

		/* Schedule the next flash update */
		if (BIT_TEST(ripng_flags, RIPNGF_TERMINATE) ||
		    time_sec + RIPNG_T_MIN + RIPNG_T_MAX < ripng_timer_update->task_timer_next_time) {
			/*
			 * We can squeeze another flash update in before the
			 * full update
			 */

			task_timer_set(tip, RIPNG_T_FLASH, (time_t) 0);
		} else {
			/*
			 * The next flash update will be scheduled after the
			 * next full update
			 */

			task_timer_reset(tip);
			BIT_SET(ripng_flags, RIPNGF_NOFLASH);
		}
	} else {
		task_timer_reset(tip);
	}
}


/*
 * Do or schedule a flash update
 */
static void
ripng_need_flash
__PF1(tp, task *)
{
	task_timer *tip = ripng_timer_flash;

	if (!tip) {
		BIT_RESET(ripng_flags, RIPNGF_FLASHDUE);
		return;
	}

	/* Indicate we need a flash update */
	BIT_SET(ripng_flags, RIPNGF_FLASHDUE);

	/* And see if we can do it now */
	if (BIT_TEST(tip->task_timer_flags, TIMERF_INACTIVE)
	    && !BIT_TEST(ripng_flags, RIPNGF_NOFLASH)) {
		/* Do it now */
		ripng_do_flash(tip, (time_t) 0);
	}
}


/*
 * Process changes in the routing table.
 */
static void
ripng_flash 
__PF2(tp, task *,
      change_list, rt_list *)
{
	int changes = 0;
	target *tlp;

	/* Re-evaluate policy */
	rt_open(tp);

	TARGET_LIST(tlp, &ripng_targets) {
		if (BIT_TEST(tlp->target_flags, TARGETF_SUPPLY)) {
			changes += ripng_policy(tp, tlp, change_list);
		}
	} TARGET_LIST_END(tlp, &ripng_targets);

	/* Close the table */
	rt_close(tp, (gw_entry *) 0, 0, NULL);

	if (changes) {
		/* Schedule a flash update */

		ripng_need_flash(tp);
	}
}


/*
 * Re-evaluate routing table
 */
static void
ripng_newpolicy 
__PF2(tp, task *,
      change_list, rt_list *)
{
	/* Indicate reconfig done */
	BIT_RESET(ripng_flags, RIPNGF_RECONFIG);

	/* And evaluate policy */
	ripng_flash(tp, change_list);
}


/*
 * Age out RIP routes
 */
static void
ripng_age 
__PF2(tip, task_timer *,
      interval, time_t)
{
	time_t expire_to = time_sec - RIPNG_T_EXPIRE;
	time_t nexttime = time_sec + 1;

	if (expire_to > 0) {
		gw_entry *gwp;
		
		rt_open(tip->task_timer_task);

		GW_LIST(ripng_gw_list, gwp) {
			rt_entry       *rt;

			if (!gwp->gw_n_routes) {
				/* No routes for this gateway */

				if (!gwp->gw_import
				    && !gwp->gw_export
				    && !BIT_TEST(gwp->gw_flags, GWF_SOURCE | GWF_TRUSTED)) {
					/* No routes, delete this gateway */

					/* XXX */
				}
				continue;
			}
			/* Age any routes for this gateway */
			RTQ_LIST(&gwp->gw_rtq, rt) {
				if (rt->rt_time <= expire_to) {
					/* This route has expired */
					rt_delete(rt);
				} else {
					/* This is the next route to expire */
					if (rt->rt_time < nexttime) {
						nexttime = rt->rt_time;
					}
					break;
				}
			} RTQ_LIST_END(&gwp->gw_rtq, rt);
		} GW_LIST_END(ripng_gw_list, gwp);

		rt_close(tip->task_timer_task, (gw_entry *) 0, 0, NULL);
	}
	if (nexttime > time_sec) {
		/* No routes to expire */

		nexttime = time_sec;
	}
	task_timer_set(tip, (time_t) 0, nexttime + RIPNG_T_EXPIRE - time_sec);
}


/*
 * Initialize static variables
 */
void
ripng_var_init()
{
	ripng_flags = RIPNGF_ON;
	ripng_default_metric = RIPNG_METRIC_UNREACHABLE;
	ripng_preference = RTPREF_RIPNG;

	/* Set up interface bits to be printed */
	int_ps_bits[RTPROTO_RIPNG] = ripng_if_bits;

	/* Build the request routing table's packet */
	memset(&packet_request, 0, sizeof(packet_request));
	packet_request.header.ripng_cmd = RIPNGCMD_REQUEST;
	packet_request.header.ripng_vers = RIPNG_VERSION_1;
	packet_request.header.ripng_zero = 0;
	COPY_ADDR6(packet_request.rte.prefix, in6addr_any);
	packet_request.rte.preflen = 0;
	packet_request.rte.rte_tag = 0;
	packet_request.rte.metric = RIPNG_METRIC_UNREACHABLE;
}


static void
ripng_tsi_dump 
__PF4(fp, FILE *,
      rth, rt_head *,
      data, void_t,
      pfx, const char *)
{
	target         *tlp = (target *) data;
	td_entry       *tdp;

	TD_TSI_GET(tlp, rth, tdp);

	if (tdp) {
		if (BIT_TEST(tdp->td_flags, TDF_HOLDDOWN | TDF_POISON)) {
			(void) fprintf(fp,
				       "%sRIPng %A%s <%s> remaining %#T\n",
				       pfx,
				       *tlp->target_dst,
				       "mc ",
			       trace_bits(target_entry_bits, tdp->td_flags),
				       tdp->td_metric * RIPNG_T_UPDATE);
		} else {
			(void) fprintf(fp,
				       "%sRIPng %A%s <%s> metric %u\n",
				       pfx,
				       *tlp->target_dst,
				       "mc ",
			       trace_bits(target_entry_bits, tdp->td_flags),
				       tdp->td_metric);
		}
	}
	return;
}

/* 
 * Send some request packet 
 */
void
ripng_send_request __PF1(tp, task*)
{
	target *tlp;

	TARGET_LIST(tlp, &ripng_targets) {
		if ((BIT_TEST(ripng_flags, RIPNGF_RECONFIG)
		     || !BIT_TEST(tlp->target_flags, RIPNGTF_POLL))
		    && BIT_TEST(tlp->target_flags, TARGETF_MULTI)
		    && !BIT_TEST(tlp->target_ifap->ifa_ps[tp->task_rtproto].ips_state, IFPS_NOIN)) {
			if_addr *ifap = NULL, *ifap2;
			
			/* Search for the first Global adress */
			if (BIT_TEST(tlp->target_ifap->ifa_state, IFS_TUNNEL)) {
				ifap = tlp->target_ifap;
			} else {
				IF_ADDR(ifap2) {
					if (tlp->target_ifap->ifa_link->ifl_index == ifap2->ifa_link->ifl_index &&
					    inet6_get_addr_type(&sock2in6(ifap2->ifa_addr), 10) == INET6_ADDR_UNICAST) {
						ifap  = ifap2;
						break;
					}
				} IF_ADDR_END(ifap2);
			}
			
			if (ifap) {
				ripng_send(tp,
					   ifap,
					   MSG_DONTROUTE,
					   ripng_addr_mc,
					   (struct ripng_packet*)&packet_request,
					   (size_t) sizeof(struct ripng_hdr) + sizeof(struct ripng_rte));
				BIT_SET(tlp->target_flags, RIPNGTF_POLL);
			}
		}
	} TARGET_LIST_END(tlp, &ripng_targets);
}

/*
 * Update the target list
 */
static void
ripng_target_list 
__PF1(jp, task_job *)
{
	register task  *tp = jp->task_job_task;
	int targets;
	target *tlp;
	flag_t target_flags = TARGETF_ALLINTF;
	static int n_targets, n_source;
	register if_link *ifl;
	int nb_valid_intf = 0;


	/* Must we supply route ? */
	if (inet6_ipforwarding) {
		/* Check for valid interfaces (UP, MC and no-LOOPBACK) */
		IF_LINK(ifl) {
			if (BIT_TEST(ifl->ifl_state, IFS_UP | IFS_MULTICAST) == (IFS_UP | IFS_MULTICAST)
			    && !BIT_TEST(ifl->ifl_state, IFS_LOOPBACK))
				nb_valid_intf ++;
		} IF_LINK_END(ifl) ;
	}
	if (nb_valid_intf > 1)
		BIT_SET(ripng_flags, RIPNGF_MULTI);
	else
		BIT_RESET(ripng_flags, RIPNGF_MULTI);
	
	if (!ripng_timer_age) {
		/* Create route age timer */
		
		ripng_timer_age = task_timer_create(tp,
						    "Age",
						    (flag_t) 0,
						    (time_t) 0,
						    RIPNG_T_EXPIRE,
						    ripng_age,
						    (void_t) 0);
	}
	if (BIT_TEST(ripng_flags, RIPNGF_SOURCE | RIPNGF_MULTI)) {
		/* We are supplying updates */

		/* Gateways do not listen to redirects */
/*FIXME		redirect_disable(tp->task_rtproto);*/

		/* Make sure the timers are active */
		if (!ripng_timer_update) {
			/* Create the update timer */

			ripng_timer_update = 
				task_timer_create(tp,
						  "Update",
						  0,
						  (time_t) RIPNG_T_UPDATE,
						  (time_t) RIPNG_T_MAX,
						  ripng_job,
						  (void_t) 0);
		}
		if (!ripng_timer_flash) {
			/* Create flash update timer */

			ripng_timer_flash = 
				task_timer_create(tp,
						  "Flash",
						  (flag_t) 0,
						  (time_t) RIPNG_T_FLASH,
						  (time_t) RIPNG_T_MAX,
						  ripng_do_flash,
						  (void_t) 0);
		}
	} else {
		/* We are not supplying updates */

		/* Hosts do listen to redirects */
/*FIXME		redirect_enable(tp->task_rtproto);*/

		/* Make sure the timers do not exist */
		if (ripng_timer_update) {
			task_timer_delete(ripng_timer_update);
			ripng_timer_update = (task_timer *) 0;
		}
		if (ripng_timer_flash) {
			task_timer_delete(ripng_timer_flash);
			ripng_timer_flash = (task_timer *) 0;
		}
	}

        if (BIT_TEST(ripng_flags, RIPNGF_MULTI)) {
		BIT_SET(target_flags, TARGETF_MULTI);
	}

	if (BIT_TEST(ripng_flags, RIPNGF_SOURCE)) {
		BIT_SET(target_flags, TARGETF_SOURCE);
	}
	/* Build or update target list */
	targets = target_build(tp,
			       &ripng_targets,
			       ripng_gw_list,
			       &ripng_addr_mc,
			       target_flags,
			       ripng_tsi_dump);
	
	/* Enable reception of packets to the RIPng multicast group */
	TARGET_LIST(tlp, &ripng_targets) {
		ripng_mc_set(tp, tlp);
	} TARGET_LIST_END(tlp, &ripng_targets);

	/* Send a RIPng REQUEST for everyone's routing table */
	if (!BIT_TEST(task_state, TASKS_TEST)) {
		ripng_send_request(tp);
	}

	/* Evaluate policy for new targets */
	{
		int             changes = 0;
		int             have_list = 0;
		rt_list        *rthl = (rt_list *) 0;
		
		rt_open(tp);
		
		TARGET_LIST(tlp, &ripng_targets) {
			if (BIT_TEST(tlp->target_flags, TARGETF_MULTI)) {
				if (BIT_TEST(tlp->target_ifap->ifa_ps[tp->task_rtproto].ips_state, IFPS_NOIN)) {
					gw_entry       *gwp;

					GW_LIST(ripng_gw_list, gwp) {
						register rt_entry *rt;

						RTQ_LIST(&gwp->gw_rtq, rt) {
							if (RT_IFAP(rt) == tlp->target_ifap) {
								rt_delete(rt);
								changes++;
							}
						} RTQ_LIST_END(&gwp->gw_rtq, rt);
					} GW_LIST_END(ripng_gw_list, gwp);
				}
			}
			switch (BIT_TEST(tlp->target_flags, TARGETF_POLICY | TARGETF_SUPPLY)) {
			case TARGETF_SUPPLY:
				/* Need to run policy for this target */

				if (!have_list) {
					/* Get target list */
					rthl = rthlist_active(AF_INET6);
					have_list++;
				}
				if (rthl) {
					/* and run policy */
					changes += ripng_policy(tp, tlp, rthl);
				}
				/* Indicate policy has been run */
				BIT_SET(tlp->target_flags, TARGETF_POLICY);
				break;

			case TARGETF_POLICY:
				/* Indicate policy run on this target */

				BIT_RESET(tlp->target_flags, TARGETF_POLICY);
				break;

			default:
				break;
			}
		} TARGET_LIST_END(tlp, &ripng_targets);

		if (rthl) {
			RTLIST_RESET(rthl);
		}
		rt_close(tp, (gw_entry *) 0, 0, NULL);

		if (changes && !BIT_TEST(ripng_flags, RIPNGF_RECONFIG))
			ripng_need_flash(tp);
	}

	if (targets != n_targets
	    || ripng_n_source != n_source) {

		tracef("ripng_target_list: ");
		if (targets) {
			tracef("supplying updates to");
			if (targets - ripng_n_source) {
				tracef(" %d interface%s",
				       targets - ripng_n_source,
				    (targets - ripng_n_source) > 1 ? "s" : "");
			}
			if (ripng_n_source) {
				tracef(" %d gateways",
				       ripng_n_source);
			}
		} else {
			tracef("just listening");
		}
		n_targets = targets;
		n_source = ripng_n_source;
		trace_log_tp(tp,
			     TRC_NL_AFTER,
			     LOG_INFO,
			     (NULL));
	}
	ripng_target_list_job = (task_job *) 0;
	/* XXX - Send a full update if anything has changed */
}


/*
 * Reinit after parse
 */
/* ARGSUSED */
static void
ripng_reinit 
__PF1(tp, task *)
{
	int entries = 0;
	gw_entry *gwp;

	trace_set(tp->task_trace, ripng_trace_options);

	/* Open the routing table */
	rt_open(tp);

	GW_LIST(ripng_gw_list, gwp) {
		rt_entry *rt;

		RTQ_LIST(&gwp->gw_rtq, rt) {
			pref_t preference = ripng_preference;

			/* Calculate preference of this route */
			if (import(rt->rt_dest,
				   rt->rt_dest,
				   ripng_import_list,
			   RT_IFAP(rt)->ifa_ps[tp->task_rtproto].ips_import,
				   rt->rt_gwp->gw_import,
				   &preference,
				   RT_IFAP(rt),
				   (void_t) 0)) {
				if (rt->rt_preference != preference) {
					/*
					 * The preference has changed, change
					 * the route
					 */
					(void) rt_change(rt,
							 rt->rt_metric,
							 rt->rt_metric2,
							 rt->rt_tag,
							 preference,
							 rt->rt_preference2,
					       rt->rt_n_gw, rt->rt_routers);
				}
				entries++;
			} else {
				/* This route is now restricted */
				rt_delete(rt);
			}
		} RTQ_LIST_END(&gwp->gw_rtq, rt);
	} GW_LIST_END(ripng_gw_list, gwp);

	/* Close the routing table */
	rt_close(tp, (gw_entry *) 0, entries, NULL);

	/* Indicate a reconfig in process */
	BIT_SET(ripng_flags, RIPNGF_RECONFIG);
}

/*
 * Terminating - clean up
 */
static void
ripng_terminate 
__PF1(tp, task *)
{
	BIT_SET(ripng_flags, RIPNGF_TERMINATE);
	
	if (BIT_TEST(ripng_flags, RIPNGF_SOURCE | RIPNGF_MULTI)) {
		/* Disable receive */
		task_set_recv(tp, 0);
		task_set_socket(tp, tp->task_socket);
		task_timer_reset(ripng_timer_update);

		/* Start shutdown procedure */
		ripng_do_flash(ripng_timer_flash, (time_t) 0);

		/* Make sure we don't try this again */
		task_set_terminate(tp, 0);
	} else {
		/* Not supplying, go away now */
		ripng_exit(tp);
	}
}


/*
 * Dump info about RIPng
 */
static void
ripng_int_dump 
__PF2(fd, FILE *,
      list, config_entry *)
{
	register config_entry *cp;
	
	CONFIG_LIST(cp, list) {
		switch (cp->config_type) {
		case RIPNG_CONFIG_IN:
			(void) fprintf(fd, " %sripin",
				       GA2S(cp->config_data) ? "" : "no");
			break;
			
		case RIPNG_CONFIG_OUT:
			(void) fprintf(fd, " %sripout",
				       GA2S(cp->config_data) ? "" : "no");
			break;
			
		case RIPNG_CONFIG_METRICIN:
			(void) fprintf(fd, " metricin %u",
				       (metric_t) GA2S(cp->config_data));
			break;
			
		case RIPNG_CONFIG_METRICOUT:
			(void) fprintf(fd, " metricout %u",
				       (metric_t) GA2S(cp->config_data));
			break;
			
		case RIPNG_CONFIG_FLAG:
			(void) fprintf(fd, " <%s>",
				       trace_bits(ripng_if_bits, (flag_t) GA2S(cp->config_data)));
			break;
			
		default:
			assert(FALSE);
			break;
		}
	} CONFIG_LIST_END(cp, list);
}


static void
ripng_dump 
__PF2(tp, task *,
      fd, FILE *)
{
	(void) fprintf(fd, "\tFlags: %s\tDefault metric: %d\t\tDefault preference: %d\n",
		       trace_bits(ripng_flag_bits, ripng_flags),
		       ripng_default_metric,
		       ripng_preference);
	target_dump(fd, &ripng_targets, ripng_target_bits);
	if (ripng_gw_list) {
		(void) fprintf(fd, "\tActive gateways:\n");
		gw_dump(fd,
			"\t\t",
			ripng_gw_list,
			tp->task_rtproto);
		(void) fprintf(fd, "\n");
	}
	if (ripng_int_policy) {
		(void) fprintf(fd, "\tInterface policy:\n");
		control_interface_dump(fd, 2, ripng_int_policy, ripng_int_dump);
	}
	control_import_dump(fd, 1, RTPROTO_RIPNG, ripng_import_list, ripng_gw_list);
	control_export_dump(fd, 1, RTPROTO_RIPNG, ripng_export_list, ripng_gw_list);
	(void) fprintf(fd, "\n");
}


/*
 * Deal with an interface status change
 */
static void
ripng_ifachange 
__PF2(tp, task *,
      ifap, if_addr *)
{
	int changes = 0;
	gw_entry *gwp;

        if (socktype(ifap->ifa_addr) != AF_INET6) {
                return;
        }

	rt_open(tp);

	switch (ifap->ifa_change) {
	case IFC_NOCHANGE:
	case IFC_ADD:
		if (BIT_TEST(ifap->ifa_state, IFS_UP)) {
			ripng_control_set(tp, ifap);
		}
		break;

	case IFC_DELETE:
	case IFC_DELETE | IFC_UPDOWN:
Down:
		GW_LIST(ripng_gw_list, gwp) {
			rt_entry       *rt;

			RTQ_LIST(&gwp->gw_rtq, rt) {
				if (RT_IFAP(rt) == ifap) {
					rt_delete(rt);
					changes++;
				}
			} RTQ_LIST_END(&gwp->gw_rtq, rt);
		} GW_LIST_END(ripng_gw_list, gwp);

		ripng_control_reset(tp, ifap);

		if (ifap == ripng_multicast_ifap) {
			IFA_FREE(ripng_multicast_ifap);
			ripng_multicast_ifap = (if_addr *) 0;
		}

		break;

	default:
		/* Something has changed */

		if (BIT_TEST(ifap->ifa_change, IFC_UPDOWN)) {
			if (BIT_TEST(ifap->ifa_state, IFS_UP)) {
				ripng_control_set(tp, ifap);
			} else {
				goto Down;
			}
		}
		if (BIT_TEST(ifap->ifa_change, IFC_METRIC)) {
			struct ifa_ps *ips = &ifap->ifa_ps[tp->task_rtproto];

			/* The metric has changed, reset the POLL bit */
			/* on any targets using this interface so we'll */
			/* send another POLL */

			if (!BIT_TEST(ips->ips_state, IFPS_METRICIN)) {
				target         *tlp;

				ips->ips_metric_in = ifap->ifa_metric + RIPNG_HOP;
				TARGET_LIST(tlp, &ripng_targets) {
					if (tlp->target_ifap == ifap) {
						BIT_RESET(tlp->target_flags, RIPNGTF_POLL);
					}
				} TARGET_LIST_END(tlp, &ripng_targets);
			}
		}

		if (BIT_TEST(ifap->ifa_change, IFC_NETMASK)) {
			/* The netmask has changed, delete any routes that */
			/* point at gateways that are no longer reachable */

			target         *tlp;

			GW_LIST(ripng_gw_list, gwp) {
				rt_entry       *rt;

				RTQ_LIST(&gwp->gw_rtq, rt) {
					if (RT_IFAP(rt) == ifap
					    && (if_withdstaddr(RT_ROUTER(rt)) != ifap
						|| BIT_TEST(rt->rt_state, RTS_IFSUBNETMASK))) {
						/*
						 * Interface for this route
						 * has changed or we derived
						 * the subnet mask
						 */

						rt_delete(rt);
						changes++;
					}
				} RTQ_LIST_END(&gwp->gw_rtq, rt);
			} GW_LIST_END(ripng_gw_list, gwp);

			TARGET_LIST(tlp, &ripng_targets) {
				if (tlp->target_ifap == ifap
				    && BIT_COMPARE(tlp->target_flags, TARGETF_SUPPLY, TARGETF_SUPPLY)) {
					/*
					 * Some subnet masks may have been
					 * derrived, indicate that policy
					 * needs to be rerun
					 */

					BIT_RESET(tlp->target_flags, TARGETF_POLICY);
				}
			} TARGET_LIST_END(tlp, &ripng_targets);
		}

		/*
		 * A LOCALADDR change will take effect when the peers notice
		 * that the old address is no longer sending */
		/* An MTU change will take effece on output */
		/* A SEL change is not possible in IP */
		break;
	}

	rt_close(tp, (gw_entry *) 0, changes, NULL);

	/* Schedule a target list rebuild if necessary */
	if (!ripng_target_list_job) {
		ripng_target_list_job = task_job_create(tp,
							TASK_JOB_FG,
							"Target_Build",
							ripng_target_list,
							(void_t) 0);
	}
}


/*
 * initialize RIPng socket and RIPng task
 */

/* ARGSUSED */
void
ripng_init()
{
	static task    *ripng_task;

	if (BIT_TEST(ripng_flags, RIPNGF_ON)) {
		trace_inherit_global(ripng_trace_options, ripng_trace_types, (flag_t) 0);
		if (!ripng_task) {
			ripng_task = task_alloc("RIPNG",
						TASKPRI_PROTO,
						ripng_trace_options);

			if (!ripng_port) {
				ripng_port = task_get_port(ripng_trace_options,
							   "ripng", "udp",
							   htons(RIPNG_PORT));
			}
			ripng_task->task_addr = sockdup(inet6_addr_any);
			ripng_task->task_proto = IPPROTO_IPV6;
			sock2port6(ripng_task->task_addr) = ripng_port;
			ripng_task->task_rtproto = RTPROTO_RIPNG;
			task_set_recv(ripng_task, ripng_recv);
			task_set_cleanup(ripng_task, ripng_cleanup);
			task_set_reinit(ripng_task, ripng_reinit);
			task_set_dump(ripng_task, ripng_dump);
			task_set_terminate(ripng_task, ripng_terminate);
			task_set_ifachange(ripng_task, ripng_ifachange);
			task_set_flash(ripng_task, ripng_flash);
			task_set_newpolicy(ripng_task, ripng_newpolicy);

			ripng_var_init();

			if ((ripng_task->task_socket = task_get_socket(ripng_task, AF_INET6, SOCK_DGRAM, 0)) < 0) {
				task_quit(errno);
			}
			if (!task_create(ripng_task)) {
				task_quit(EINVAL);
			}

			if (task_set_option(ripng_task,
					    TASKOPTION_RECVBUF,
					    task_maxpacket) < 0) {
				task_quit(errno);
			}
			if (task_set_option(ripng_task,
					    TASKOPTION_NONBLOCKING,
					    TRUE) < 0) {
				task_quit(errno);
			}
			if (task_set_option(ripng_task,
					    TASKOPTION_MULTI_HOPS_IPV6,
					    255) < 0) {
				task_quit(errno);
			}
			if (task_addr_local(ripng_task, ripng_task->task_addr)) {
				trace_log_tp(ripng_task,
					     0,
					     LOG_ERR,
					     ("ripng_init: cant bind ripng port. Is an old copy of gated running ?"));
				task_quit(errno);
			}
			/* Allocate the buffers */
			task_alloc_send(ripng_task, RIPNG_PKTSIZE);
			task_alloc_recv(ripng_task, RIPNG_PKTSIZE);

#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
			ripng_init_mib(TRUE);
#endif /* defined(PROTO_SNMP) && defined(MIB_RIPNG) */

			if (!ripng_addr_mc) {
				/* Initialize address constant */

				ripng_addr_mc = sockdup(sockbuild_in6(0, &INET6_RIPNG_ALLROUTERS, 128));
				sock2port6(ripng_addr_mc)=htons(RIPNG_PORT);
			}
		}
	} else {
		ripng_cleanup((task *) 0);

#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
		ripng_init_mib(FALSE);
#endif /* defined(PROTO_SNMP) && defined(MIB_RIPNG) */
		if (ripng_task) {
			ripng_terminate(ripng_task);

			ripng_task = (task *) 0;
		}
	}
}
